// dosepr.C Version 7.43  Copyleft (c) 杉本典夫
// Time-stamp: <2025-04-08 16:23:28 tac>
//
// クロス集計表を入力し、反応率に関する各種の用量反応解析を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W dosepr dosepr.C
//               -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "dosepr  Version 7.43         Copyleft(Vo\\oV) 2025 杉本典夫",
  "書式:dosepr [<オプション> …]  [<クロス集計表ファイル> …]",
  "機能:縦を用量項目、横を反応項目としたクロス集計表に対して",
  "     反応率に関する用量反応解析をするのだ。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するのだ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -du={l|r}:用量の単位を指定するのだ。",
  "  -eps[=<収束誤差>]:最尤法の収束誤差を指定するのだ。(0<収束誤差)",
  "  -f=<ファイル>:オプションを指定したファイルから読み込むのだ。",
  "  -g[={-|+}]:Fiellerの式を使うかどうか指定するのだ。(-g:-g=+)",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するのだ。(-h:-h=man)",
  "  -log[={comp|exc|work}]:ロジット分析の種類を指定するのだ。(-log:-log=comp)",
  "    comp:反復補完法  exc:除外法  work:実用ロジット法",
  "  -ngp={drc|drl|drlp|pro}[,…]:Ngraph用出力データを指定するのだ。",
  "    drc:プロビット分析の用量反応率曲線  drl:傾向分析の用量反応率直線",
  "    drlp:ロジット分析の用量反応率曲線  pro:反応率  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データを出力するファイル名を指定する。",
  "    オプション省略時:Ngraph用データ出力せず",
  "    -ngpname:-ngpname=dosepr_ngp<2桁の連番>.dat",
  "  -res=[-][{def|all|ca|log|pa|pr[<反応率>]|r|tab[{f|p|pc|pr|pt}]}]",
  "       [,…]:解析結果の出力項目を指定するのだ。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    ca:傾向分析  log:ロジット分析  pa:プロビット分析",
  "    pr[<反応率>]:用量の逆推定(0≦反応率＜1)  r:寄与率",
  "    tab[{fr|p|pc|pr|pt}]:クロス集計表",
  "      fr:度数だけ出力  p:prと同じ(tabも同じ)  pc:列合計に対する％も出力",
  "      pr:行合計に対する％も出力  pt:全合計に対する％も出力",
  "  -rmax=<最大反復回数>:最大反復回数を指定するのだ。(0≦最大反復回数)",
  "  -xrange=<最小値>-<最小値>:X軸の最小値と最大値を指定します。"
};

const string Title = "=== 反応率に関する用量反応解析 ==="; // 表題
const string NGPNAMEdef = "dosepr_ngp"; // 省略時Ngraphファイル名
constexpr int RMAXdef = 200;       // 最大計算回数省略時値
constexpr double EPSdef = 0.00001; // 省略時収束誤差
constexpr double PRdef = 0.5;      // 省略時用量推定反応率

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

enum logtype                    // ロジット分析の種類
{
  lognon,                       // 実行せず
  logcomp,                      // 反復補完法
  logexc,                       // 除外
  logwork,                      // 実用ロジット
  logetc                        // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  dutype vdu;                   // 用量の単位
  double veps;                  // 最尤法の収束誤差
  bool vg;                      // Fiellerの式使用フラッグ
  logtype vlog;                 // ロジット分析の種類
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  double vpr;                   // 用量を推定する反応率
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  int vrmax;                    // 最大反復回数
  tabtype vtab;                 // クロス集計表の種類
  pairdbl vxrange;              // X軸の最小値と最大値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  bool g() const                // Fiellerの式使用フラッグをす
  { return vg; }
  logtype log() const           // ロジット分析の種類を返す
  { return vlog; }
  const string& ngpname(int fno); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pr() const             // 用量を推定する反応率を返す
  { return vpr; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  tabtype tab() const           // クロス集計表の種類を返す
  { return vtab; }
  double xmin() const           // X軸の最小値を返す
  { return vxrange.first; }
  double xmax() const           // X軸の最大値を返す
  { return vxrange.second; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdu(dulog), veps(EPSdef), vg(true), vlog(logcomp), vngpname(), vngpnamen(), vngplst(), vngpmap(), vpr(PRdef), vreslst(), vresmap(), vrmax(RMAXdef), vtab(tabfprr), vxrange(BLANKf, BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_citr ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

const string& option::ngpname(int fno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (fno < 10) ostr << vngpname << "0" << fno << ".dat";
  else ostr << vngpname << fno << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "du")      // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < FDeps) err = badopt(vpname, optstr, USAGE, FDeps, FDmax);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g")       // Fiellerの式使用指定
  {
    if (opt[1].empty() || opt[1] == "+") vg = true;
    else if (opt[1] == "-") vg = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "log")     // ロジット分析の種類指定
  {
    if (opt[1].empty() || opt[1] == "comp") vlog = logcomp;
    else if (opt[1] == "exc") vlog = logexc;
    else if (opt[1] == "work") vlog = logwork;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "pr")       // 用量を推定する反応率の指定
  {
    vpr = opt[1].empty() ? PRdef : strtodf(opt[1]);
    if (vpr <= 0.0 || 1.0 <= vpr) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "xrange")  // X軸の最小値と最大値の指定
  {
    if (getrange(vxrange.first, vxrange.second, opt[1]) == 0) // 範囲指定無
    { err = badopt(vpname, optstr, USAGE, -FDmax, FDmax); }
    else if (vxrange.first >= vxrange.second) // 範囲指定不適切
    { err = badopt(vpname, optstr, USAGE, -FDmax, FDmax); }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -resオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["drc"] = vngpmap["drl"] = vngpmap["drlp"] = vngpmap["pro"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["drc"] = vngpmap["drl"] = vngpmap["drlp"] = vngpmap["pro"] = false;
    for (const string& ngpstr : vngplst)
    {
      mapsb_citr ngppos = vngpmap.find(ngpstr);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngpstr + "オプション指定が不適切！"); }
      else
      { vngpmap[ngpstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["pa"] = vresmap["pr"] = true;
  vresmap["ca"] = vresmap["log"] = vresmap["r"] = vresmap["tab"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["pa"] = vresmap["pr"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def")    // 省略時出力
      { vresmap["def"] = vresmap["pa"] = vresmap["pr"] = true; }
      else if (resstr == "all")
      {
        vresmap["res"] = vresmap["def"] = vresmap["pa"] = vresmap["pr"] = vresmap["ca"] = vresmap["log"] = vresmap["r"] = vresmap["tab"] = true;
      }
      else if (resstr[0] == 'p' && resstr[1] == 'r' && resstr.size() > 2)
      {
        vpr = strtodf(resstr.substr(2)); // 反応率
        if (vpr < 0.0 || 1.0 <= vpr) return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！");
        vresmap["cc"] = true;
      }
      else if (resstr == "tabfr")
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpc")
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabp" || resstr == "tabpr")
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt")
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class npro                      // 用量別プロビットクラス
{
private:
  double vd;                    // 用量
  double vx;                    // 計算用用量
  int vtn;                      // 総例数
  int vrn;                      // 反応例数
  double vpc;                   // 補正反応率
  double vyo;                   // 観測プロビット
  double vy;                    // 期待プロビット
public:
  npro() : vd(0.0), vx(0.0), vtn(0), vrn(0), vpc(0.0), vyo(0.0), vy(0.0)
  // デフォールトコンストラクタ
  {}
  npro(double d0, double x0, int tn0, int rn0) : vd(d0), vx(x0), vtn(tn0), vrn(rn0), vpc(0.0), vyo(0.0), vy(0.0)
  // コンストラクタ
  {}
  npro(const npro& npro0) : vd(npro0.vd), vx(npro0.vx), vtn(npro0.vtn), vrn(npro0.vrn), vpc(npro0.vpc), vyo(npro0.vyo), vy(npro0.vy)
  // コピーコンストラクタ
  {}
  virtual ~npro()               // デストラクタ
  {}
  const npro& operator=(const npro& npro0); // コピー代入演算子
  double d() const                          // 用量を返す
  { return vd; }
  double d(double d0)           // 用量を設定する
  { return (vd = d0); }
  double x() const              // 計算用用量を返す
  { return vx; }
  double x(double x0)           // 計算用用量を設定する
  { return (vx = x0); }
  int tn() const                // 総例数を返す
  { return vtn; }
  int tn(int tn0)               // 総例数を設定する
  { return (vtn = tn0); }
  int rn() const                // 反応例数を返す
  { return vrn; }
  int rn(int rn0)               // 反応例数を設定する
  { return (vrn = rn0); }
  double p() const              // 反応率を返す
  { return (vtn > 0) ? (double)vrn / (double)vtn : 0.0; }
  double pc() const             // 補正反応率を返す
  { return vpc; }
  double yo() const             // 観測プロビットを返す
  { return vyo; }
  double yo(double yo0)         // 観測プロビットを設定する
  { return (vyo = yo0); }
  double y() const              // 期待プロビットを返す
  { return vy; }
  double y(double y0)           // 期待プロビットを設定する
  { return (vy = y0); }
  void sigma(int tn0, int rn0)  // データを累積する
  { vtn += tn0;  vrn += rn0; }
  double calpc(double p0);      // 補正反応率を計算する
};

const npro& npro::operator=(const npro& npro0) // コピー代入演算子
{
  if (&npro0 == this) return *this; // 自分自身への代入を防ぐ
  vd = npro0.vd;  vx = npro0.vx;  vtn = npro0.vtn;  vrn = npro0.vrn;
  vpc = npro0.vpc;  vyo= npro0.vyo;  vy = npro0.vy;
  return *this;
}

double npro::calpc(double p0)   // 補正反応率を計算する
{
  vpc = (vtn > 0) ? (double)vrn / (double)vtn : 0.0;
  if (p0 > 0.0) vpc = (vpc - p0) / (1.0 - p0);
  if (vpc < 0.0) vpc = 0.0;
  return vpc;                   // 補正反応率を返す
}

class varclass                  // 変数クラス
{
private:
  int vnd;                      // 用量数
  vector<npro> nprovec;         // 用量別プロビットベクトル
  statcor<double> dstatc;       // 用量プロビット用相関統計量
  pairdbl xrange;               // 用量の最小値と最大値
  double vsnw;                  // 重み合計
  double vmxw;                  // xの重み付き平均値
  double vmyw;                  // yの重み付き平均値
  double vsxxw;                 // xの重み付き平方和
  double vsyyw;                 // yの重み付き平方和
  double vsxyw;                 // xとyの重み付き積和
  double va;                    // 用量プロビット直線の切片
  double vb;                    // 用量プロビット直線の回帰係数
  int vnoutd;                   // 範囲外の用量数
  int vnr;                      // 繰り返し計算数
  restype vresvar;              // 処理結果の種類
  restype cal_pline(const option& opt); // プロビット直線の近似計算
  restype cal_logit(double& a, double& b, const option& opt, logtype logt);
  // 用量-ロジット分析
public:
  varclass(int nd0) : vnd(nd0), nprovec(nd0), dstatc(), xrange(FDmax, -FDmax), vsnw(0.0), vmxw(0.0), vsxxw(0.0), vsyyw(0.0), vsxyw(0.0), va(0.0), vb(0.0), vnoutd(0), vnr(0), vresvar(resnon) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  void cal(const option& opt, const ftable& ftab);
  // 用量別反応率とプロビットを計算する
  void caldpline(const option& opt);
  // Newton法による用量-プロビット直線の反復計算
  void calcaline(const option& opt);
  // Cochran-Armitageの用量-反応率直線の計算
  void callpline(const option& opt); // 用量-ロジット直線の反復計算
  int nd() const                     // 用量数を返す
  { return vnd; }
  double d(size_t idx) const    // 用量を返す
  { return (idx < nprovec.size()) ? nprovec[idx].d() : 0.0; }
  double x(size_t idx) const    // 計算用用量を返す
  { return (idx < nprovec.size()) ? nprovec[idx].x() : 0.0; }
  int tn(size_t idx) const      // 総例数を返す
  { return (idx < nprovec.size()) ? nprovec[idx].tn() : 0.0; }
  int rn(size_t idx) const      // 反応例数を返す
  { return (idx < nprovec.size()) ? nprovec[idx].rn() : 0.0; }
  double p(size_t idx) const    // 反応率を返す
  { return (idx < nprovec.size()) ? nprovec[idx].p() : 0.0; }
  double pc(size_t idx) const   // 補正反応率を返す
  { return (idx < nprovec.size()) ? nprovec[idx].pc() : 0.0; }
  double yo(size_t idx) const   // 観測プロビットを返す
  { return (idx < nprovec.size()) ? nprovec[idx].yo() : 0.0; }
  double y(size_t idx) const    // 期待プロビットを返す
  { return (idx < nprovec.size()) ? nprovec[idx].y() : 0.0; }
  double xmin() const           // 用量最小値を返す
  { return xrange.first; }
  double xmax() const           // 用量最大値を返す
  { return xrange.second; }
  double snw() const            // 重み合計を返す
  { return vsnw; }
  double mxw() const            // xの重み付き平均値を返す
  { return vmxw; }
  double myw() const            // yの重み付き平均値を返す
  { return vmyw; }
  double sxxw() const           // xの重み付き平方和を返す
  { return vsxxw; }
  double syyw() const           // yの重み付き平方和を返す
  { return vsyyw; }
  double sxyw() const           // xとyの重み付き積和を返す
  { return vsxyw; }
  double a() const              // 用量プロビット直線の定数を返す
  { return va; }
  double b() const              // 用量プロビット直線の回帰係数を返す
  { return vb; }
  int noutd() const             // 範囲外の用量数を返す
  { return vnoutd; }
  int nr() const                // 繰り返し計算数を返す
  { return vnr; }
  restype resvar() const        // 処理結果の種類を返す
  { return vresvar; }
};

void varclass::cal(const option& opt, const ftable& ftab)
// 用量別反応率とプロビットを計算する
{
  double p0 = 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    size_t j = i - 1;
    double dose = strtodf(ftab.rcode(i));
    nprovec[j].d(dose);
    double x = dose;
    if (opt.du() == dulog) x = (dose > 0.0) ? log10(dose) : BLANKf;
    if (x != BLANKf && x < xrange.first) xrange.first = x;
    if (x != BLANKf && x > xrange.second) xrange.second = x;
    nprovec[j].x(x);
    nprovec[j].tn(ftab.frtotal(i));  nprovec[j].rn(ftab.f(i, 2));
    if (i == 1 && opt.du() == dulog && dose == 0.0) // 用量0のデータがある時
    {
      p0 = nprovec[j].p();      // 反応率補正用
      nprovec[j].calpc(p0);     // 補正反応率計算(0になる)
      nprovec[j].yo(BLANKf);    // 用量0の観測プロビットは欠損にする
      nprovec[j].y(BLANKf);     // 用量0の期待プロビットも欠損にする
    }
    else
    {
      double yo = probit(nprovec[j].calpc(p0)); // 観測プロビット計算
      nprovec[j].yo(yo);
      if (x != BLANKf && yo != BLANKf) dstatc.sigma(x, yo);
    }
  }
  if (dstatc.n() < 2)           // 期待プロビットが計算できない
  {
    vresvar = reserr;
    return;
  }
  double b = (dstatc.sxx() > 0.0) ? dstatc.sxy() / dstatc.sxx() : 0.0;
  double a = dstatc.my() - b * dstatc.mx();
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0) istart = 1;
  for (int i = istart; i < vnd; ++i) // 期待プロビットをセットする
  { nprovec[i].y(a + b * nprovec[i].x()); }
  vresvar = resok;              // プロビット計算完了
}

void varclass::caldpline(const option& opt)
// Newton法による用量プロビット直線の反復計算
{
  vnr = 0;  vresvar = resnon;
  do {
    ++vnr;
    restype respline = cal_pline(opt);
    if (respline == resok)
    {
      vb = (sgn(vsxxw) > 0) ? vsxyw / vsxxw : 0.0;
      va = vmyw - vb * vmxw;
      double emax = 0.0;
      for (int i = 0; i < vnd; ++i)
      {
        double x = nprovec[i].x();
        if (x != BLANKf)
        {
          double y = va + vb * x;
          double e = fabs(y - nprovec[i].y());
          if (e > emax) emax = e;
          nprovec[i].y(y);
        }
      }
      if (emax < opt.eps()) vresvar = resok; // 収束
      else if (opt.rmax() > 0 && vnr >= opt.rmax()) vresvar = resrmax; // 反復回数最大値
    }
    else
    { vresvar = respline; }     // エラー発生
  } while (vresvar == resnon);
}

void varclass::calcaline(const option& opt)
// Cochran-Armitageの用量反応率直線の計算
{
  vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0) istart = 1; // 用量0のデータがある時
  if (vnd - istart < 2) // 計算できず！
  {
    vresvar = reserr;
    return;
  }
  for (int i = istart; i < vnd; ++i)
  {
    double x = nprovec[i].x(), pc = nprovec[i].pc();
    double nw = (double)(nprovec[i].tn());
    vsnw += nw;  vmxw += nw * x;  vmyw += nw * pc;
    vsxxw += nw * x * x;  vsyyw += nw * pc * pc;
    vsxyw += nw * x * pc;
  }
  if (vsnw < FDeps)
  {
    vsnw = 0.0;
    vresvar = reserr;           // 用量反応直線計算できず
    return;
  }
  vmxw /= vsnw;  vmyw /= vsnw;
  double w = (sgn(vmyw) == 0 || sgndiff(1.0 , vmyw) == 0) ? 1.0 : 1.0 / vmyw / (1.0 - vmyw);
  vsnw *= w;
  vsxxw = w * vsxxw - vsnw * vmxw * vmxw;
  if (vsxxw < FDeps) vsxxw = 0.0;
  vsyyw = w * vsyyw - vsnw * vmyw * vmyw;
  if (vsyyw < FDeps) vsyyw = 0.0;
  vsxyw = w * vsxyw - vsnw * vmxw * vmyw;
  if (fabs(vsxyw) < FDeps) vsxyw = 0.0;
  vb = (vsxxw > 0.0) ? vsxyw / vsxxw : 0.0;
  va = vmyw - vb * vmxw;
  vresvar = resok;              // 用量反応直線計算完了
}

void varclass::callpline(const option& opt) // 用量ロジット直線の反復計算
{
  vnr = 0;
  restype reslogit = cal_logit(va, vb, opt, (opt.log() == logcomp) ? logwork : opt.log());
  if (reslogit == reserr)       // エラー発生
  {
    vresvar = reslogit;
    return;
  }
  const double Eps = vsyyw * opt.eps();
  vresvar = resok;
  if (vnoutd > 0 && opt.log() == logcomp) // 反復補完法
  {
    double syywpre = vsyyw;
    vresvar = resnon;
    while (vresvar == resnon)
    {
      ++vnr;
      reslogit = cal_logit(va, vb, opt, logcomp);
      if (reslogit == reserr) vresvar = reslogit;
      else if (fabs(vsyyw - syywpre) < Eps) vresvar = resok;
      else if (opt.rmax() > 0 && vnr >= opt.rmax()) vresvar = resrmax;
      else syywpre = vsyyw;
    }
    double at = va, bt = vb;    // 収束後のaとbで平方和を求める
    reslogit = cal_logit(at, bt, opt, logcomp);
    if (reslogit == reserr) vresvar = reslogit; // 用量ロジット直線計算できず
  }
}

restype varclass::cal_pline(const option& opt) // プロビット直線の近似計算
{
  constexpr double Sqrpi = 1.0 / sqrt(2.0 * M_PI);
  constexpr double NEDmax = 10.0;
  vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
  double p0 = 0.0;
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0)
  {
    p0 = nprovec[0].p();
    istart = 1;
  }
  if (vnd - istart < 2) return reserr; // 計算できず！
  double pc = (p0 < 1.0) ? p0 / (1.0 - p0) : 0.0;
  for (int i = istart; i < vnd; ++i)
  {
    double x = nprovec[i].x(), y = nprovec[i].y();
    if (x != BLANKf && y != BLANKf)
    {
      double ned = (y - 5.0) * (y - 5.0);
      if (ned > NEDmax) ned = NEDmax;
      double z = Sqrpi * exp(-ned * 0.5);
      ned = y - 5.0;
      if (ned > NEDmax) ned = NEDmax;
      else if (ned < -NEDmax) ned = -NEDmax;
      double p = nedp(ned);
      y += (nprovec[i].pc() - p) / z;
      double nw = nprovec[i].tn() * z * z / (p + pc) / (1.0 - p);
      vsnw += nw;  vmxw += nw * x;  vsxxw += nw * x * x;
      vmyw += nw * y;  vsyyw += nw * y * y;  vsxyw += nw * x * y;
    }
  }
  if (vsnw < FDeps)
  {
    vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
    return reserr;              // プロビット直線計算できず
  }
  vsxxw = vsxxw - vmxw * vmxw / vsnw;
  if (vsxxw < FDeps) vsxxw = 0.0;
  vsyyw = vsyyw - vmyw * vmyw / vsnw;
  if (vsyyw < FDeps) vsyyw = 0.0;
  vsxyw = vsxyw - vmxw * vmyw / vsnw;
  if (fabs(vsxyw) < FDeps) vsxyw = 0.0;
  vmxw = vmxw / vsnw;  vmyw = vmyw / vsnw;
  return resok;                 // プロビット直線計算完了
}

restype varclass::cal_logit(double& a, double& b, const option& opt, logtype logt)
// 用量反応ロジット分析
{
  vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
  vnoutd = 0;                   // 範囲外の用量数
  double y = 0.0, nw = 0.0;
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0) istart = 1; // 用量0のデータがある時
  if (vnd - istart < 2) return reserr; // 計算できず！
  for (int i = istart; i < vnd; ++i)
  {
    double x = nprovec[i].x(), pc = nprovec[i].pc();
    int n = nprovec[i].tn();
    if (0.0 < pc && pc < 1.0)
    {
      y = log(pc / (1.0 - pc));
      nw = (double)n * pc * (1.0 - pc);
    }
    else
    {
      if (logt == logcomp)      // 反復補完法
      {
        double p = 1.0 / (1.0 + exp(-a - b * x));
        y = log(p / (1.0 - p));  nw = (double)n * p * (1.0 - p);
      }
      else if (logt == logwork) // 実用ロジット法
      {
        double cf = 0.5 / (double)n;
        y = log((pc + cf) / (1.0 - pc + cf));
        nw = (double)n * (pc + cf) * (1.0 - pc + cf) / (1.0 + 2.0 * cf);
      }
      else                      // それ以外は除外法
      { y = 0.0;  nw = 0.0; }
      ++vnoutd;
    }
    vsnw += nw;  vmxw += nw * x;  vsxxw += nw * x * x;
    vmyw += nw * y;  vsyyw += nw * y * y;  vsxyw += nw * x * y;
  }
  if (vsnw < FDeps)
  {
    vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
    return reserr;              // ロジット分析不可能
  }
  vsxxw = vsxxw - vmxw * vmxw / vsnw;
  if (vsxxw < FDeps) vsxxw = 0.0;
  vsyyw = vsyyw - vmyw * vmyw / vsnw;
  if (vsyyw < FDeps) vsyyw = 0.0;
  vsxyw = vsxyw - vmxw * vmyw / vsnw;
  if (fabs(vsxyw) < FDeps) vsxyw = 0.0;
  vmxw = vmxw / vsnw;  vmyw = vmyw / vsnw;
  b = vsxyw / vsxxw;
  a = vmyw - b * vmxw;
  return resok;                 // ロジット分析完了
}

class doseprclass               // doseprメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_dosepr(ostream& fout, option& opt, const ftable& ftab, varclass& var) const;
  // 各種の用量反応解析の結果を出力する
  void put_nstat(ostream& fout, option& opt, const varclass& var) const;
  // 用量別反応率・プロビット等を出力する
  void put_probit(ostream& fout, option& opt, const varclass& var) const;
  // プロビット分析の結果を出力する
  void put_estdp(ostream& fout, option& opt, const varclass& var) const;
  // プロビット分析による用量の逆推定値を出力する
  void put_doseprtab(ostream& fout, const string& head, double syyw, double sb, int dft) const;
  // 用量反応解析の分散分析表を出力する
  void put_catest(ostream& fout, option& opt, const varclass& var) const;
  // Cochran-Armitageの傾向分析の結果を出力する
  void put_estca(ostream& fout, option& opt, const varclass& var) const;
  // 傾向分析による用量の逆推定値を出力する
  void put_logit(ostream& fout, option& opt, const varclass& var) const;
  // ロジット分析の結果を出力する
  void put_estlog(ostream& fout, option& opt, const varclass& var) const;
  // ロジット分析による用量の逆推定値を出力する
  void plot_ngp(ofstream& fout, option& opt, idd& id, varclass& var) const;
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngppro(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に反応率を出力する
  void plot_ngpdrc(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用にプロビット分析の用量-反応率曲線を出力する
  void plot_ngpdrl(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に傾向分析の用量-反応率直線を出力する
  void plot_ngpdrlp(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用にロジット分析の用量-反応率曲線を出力する
public:
  doseprclass() : fname()       // コンストラクタ
  {}
  ~doseprclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // クロス集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst, int fno) const;
  // 結果を出力する
};

void doseprclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "用量項目(縦):" << id.item(i) << "\n";
      break;
    default : fout << "反応項目(横):" << id.item(i) << "\n";
    }
  }
}

void doseprclass::put_dosepr(ostream& fout, option& opt, const ftable& ftab, varclass& var) const
// 各種の用量反応解析の結果を出力する
{
  if (opt.res("tab")) putftable(fout, ftab, opt.tab());
  put_nstat(fout, opt, var);
  if (ftab.nr() < 2 || ftab.nc() != 2)
  {
    fout << "※用量反応解析不可能！\n";
    return;
  }
  if (opt.res("pa"))            // プロビット分析出力
  {
    var.caldpline(opt);
    put_probit(fout, opt, var);
  }
  if (opt.res("ca"))            // Cochran-Armitageの傾向分析出力
  {
    var.calcaline(opt);
    put_catest(fout, opt, var);
  }
  if (opt.res("log"))           // ロジット分析出力
  {
    var.callpline(opt);
    put_logit(fout, opt, var);
  }
}

void doseprclass::put_nstat(ostream& fout, option& opt, const varclass& var) const
// 用量別反応率・プロビット等を出力する
{
  const int LENline = (opt.du() == dureal) ? 68 : 78;
  string tstr("観測      期待プロ");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << right << setw(LENline - 1 + npad) << tstr;
  fout <<  "\n";
  fout << "実数用量";
  if (opt.du() == dulog) fout << "  対数用量";
  fout << "     反応率(反応数/総数)  補正反応率  プロビット   ビット(y)\n";
  putline(fout, LENline);
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  int istart = 0;
  if (opt.du() == dulog && var.d(0) == 0.0) // 用量0のデータがある時
  {
    fout << setw(8) << setprecision(4) << var.d(0) << setprecision(prec0);
    fout << setw(10) << "-";
    fout << setw(11) << var.p(0);
    fout << "(" << setw(5) << var.rn(0);
    fout << "/";
    fout << left << setw(5) << var.tn(0) << right;
    fout << ")" << setw(12) << 0;
    fout << "\n";
    istart = 1;
  }
  tstr = "↓";
  npad = tstr.size() - strclen(tstr, CHARwide);
  string tup("↑");
  int npadu = tup.size() - strclen(tup, CHARwide);
  for (int i = istart; i < var.nd(); ++i)
  {
    fout << setw(8) << setprecision(4) << var.d(i) << setprecision(prec0);
    if (opt.du() == dulog)
    {
      fout << setw(10) << setprecision(4) << var.x(i) << setprecision(prec0);
    }
    fout << setw(11) << var.p(i);
    fout << "(" << setw(5) << var.rn(i);
    fout << "/" << left << setw(5) << var.tn(i) << right;
    fout << ")" << setw(12) << var.pc(i);
    if (var.pc(i) <= 0.0) fout << setw(12 + npad) << tstr;
    else if (var.pc(i) >= 1.0) fout << setw(12 + npadu) << tup;
    else fout << setw(12) << var.yo(i);
    fout << setw(12) << var.y(i) << "\n";
  }
  putline(fout, LENline);
}

void doseprclass::put_probit(ostream& fout, option& opt, const varclass& var) const
// プロビット分析の結果を出力する
{
  fout << "\n・";
  if (opt.du() == dulog) fout << "対数";
  fout << "用量プロビット直線:y=" << var.a() << showpos << var.b() << noshowpos;
  fout << "x  反復回数=" << var.nr();
  if (var.resvar() == resrmax) fout << "(最大回数)";
  fout << "\n";
  if (var.resvar() == reserr)
  {
    fout << "※計算エラー発生！" << endl;
    return;
  }
  double sb = var.b() * var.sxyw();
  if (opt.res("r"))
  {
    double r2 = sb / var.syyw();
    fout << "  寄与率:r2乗=" << r2 << "\n";
  }
  if (opt.res("pr")) put_estdp(fout, opt, var);
  int dft = var.nd() - 1;
  if (opt.du() == dulog && var.d(0) == 0.0) --dft;
  put_doseprtab(fout, "プロビット分析", var.syyw(), sb, dft);
}

void doseprclass::put_estdp(ostream& fout, option& opt, const varclass& var) const
// プロビット分析による用量の逆推定値を出力する
{
  if (sgn(var.b()) == 0) return; // 逆推定できず
  double pr = opt.pr() * 100.0, y = probit(opt.pr());
  double m = (y - var.a()) / var.b();
  fout << "  反応率=" << pr << "%(y=" << y << ")の時の用量:D" << pr << "=";
  if (opt.du() == dureal) fout << "x=" << m << "\n";
  else fout << pow(10.0, m) << "(x=log(D" << pr << ")=" << m << ")\n";
  double vr = 1.0, t = ptoz(1.0 - opt.cc());
  double g = t * t * vr / var.b() / var.b() / var.sxxw();
  if (opt.g() == false || g >= 1.0) g = 0.0; // g>1の時は計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / var.snw() + (m - var.mxw()) * (m - var.mxw()) / var.sxxw());
  double ci = t / var.b() * sqrt(v);
  double ml = (m - g * var.mxw() - ci) / gr, mu = (m - g * var.mxw() + ci) / gr;
  fout << "  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal)
  {
    fout << "D" << pr << "下限=" << ml;
    fout << "  D" << pr << "上限=" << mu << "\n";
  }
  else
  {
    fout << "下限=" << pow(10.0, ml) << "(x=" << ml;
    fout << ")  上限=" << pow(10.0, mu) << "(x=" << mu << ")\n";
  }
}

void doseprclass::put_doseprtab(ostream& fout, const string& head, double syyw, double sb, int dft) const
// 用量反応解析の分散分析表を出力する
{
  int w = strclen(head, CHARwide);
  int npad = head.size() - w;
  fout << "\n" << setw((68 + w) / 2 + npad) << head;
  fout << "\n要因                平方和    自由度        χ2乗値   有意確率p値\n";
  constexpr int LENline = 68;
  putline(fout, LENline);
  int dfb = 1;
  int dfl = dft - dfb;
  double sl = syyw - sb;
  if (dfl < 1 || sl < FDeps) sl = 0.0;
  ldbl pr = ctop(sb, dfb);
  fout << "直線性" << setw(20) << sb << setw(10) << dfb;
  fout << setw(14) << sb << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(sl, dfl);
  fout << "異質性(ズレ)" << setw(14) << sl << setw(10) << dfl;
  fout << setw(14) << sl << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "全体" << setw(22) << syyw << setw(10) << dft << endl;
}

void doseprclass::put_catest(ostream& fout, option& opt, const varclass& var) const
// Cochran-Armitageの傾向分析の結果を出力する
{
  int dft = var.nd() - 1;
  if (opt.du() == dulog && var.d(0) == 0.0) --dft; // 用量0のデータがある時
  double sb = var.sxyw() * var.sxyw() / var.sxxw();
  fout<< "\n・";
  if (opt.du() == dulog) fout << "対数";
  fout << "用量反応率直線:p=" << var.a() << showpos << var.b() << noshowpos << "x\n";
  if (var.resvar() == reserr)
  {
    fout << "※計算エラー発生！" << endl;
    return;
  }
  if (opt.res("r"))
  {
    double r2 = (var.syyw() > 0.0) ? sb / var.syyw() : 0.0;
    fout << "  寄与率:r2乗=" << r2 << "\n";
  }
  if (opt.res("pr")) put_estca(fout, opt, var);
  put_doseprtab(fout, "Cochran-Armitageの傾向分析", var.syyw(), sb, dft);
}

void doseprclass::put_estca(ostream& fout, option& opt, const varclass& var) const
// 傾向分析による用量の逆推定値を出力する
{
  if (sgn(var.b()) == 0) return; // 逆推定できず
  double pr = opt.pr() * 100.0, p = opt.pr();
  double m = (p - var.a()) / var.b();
  fout << "  反応率=" << pr << "%(p=" << p << ")の時の用量:D" << pr << "=";
  if (opt.du() == dureal) fout << "x=" << m << "\n";
  else fout << pow(10.0, m) << "(x=log(D" << pr << ")=" << m << ")\n";
  double vr = 1.0, t = ptoz(1.0 - opt.cc());
  double g = t * t * vr / var.b() / var.b() / var.sxxw();
  if (g >= 1.0) g = 0.0;        // g>1の時は計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / var.snw() + (m - var.mxw()) * (m - var.mxw()) / var.sxxw());
  double ci = t / var.b() * sqrt(v);
  double ml = (m - g * var.mxw() - ci) / gr, mu = (m - g * var.mxw() + ci) / gr;
  fout << "  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal)
  {
    fout << "D" << pr << "下限=" << ml;
    fout << "  D" << pr << "上限=" << mu << "\n";
  }
  else
  {
    fout << "下限=" << pow(10.0, ml) << "(x=" << ml;
    fout << ")  上限=" << pow(10.0, mu) << "(x=" << mu << ")\n";
  }
}

void doseprclass::put_logit(ostream& fout, option& opt, const varclass& var) const
// ロジット分析の結果を出力する
{
  fout<< "\n・";
  if (opt.du() == dulog) fout << "対数";
  fout << "用量ロジット直線:lp=ln(p/(1-p))=" << var.a() << showpos << var.b() << noshowpos << "x\n  ";
  if (opt.du() == dulog) fout << "対数";
  fout << "用量反応率曲線:p=1/{1+exp(" << -var.a() << showpos << -var.b() << noshowpos << "x)}\n";
  fout << "  p0=1/{1+exp(" << -var.a() << ")}=" << 1.0 / (1.0 + exp(-var.a())) << "\n";
  fout << "  範囲外用量数=" << var.noutd() << "  範囲外用量計算法:";
  switch (opt.log())
  {
  case logwork : fout << "実用ロジット法";
    break;
  case logexc : fout << "除外法";
    break;
  default : fout << "反復補完法  補完回数=" << var.nr();
    if (var.resvar() == resrmax) fout << "(最大回数)";
  }
  fout << "\n";
  if (var.resvar() == reserr)
  {
    fout << "※計算エラー発生！" << endl;
    return;
  }
  int dft = var.nd() - 1;
  if (opt.du() == dulog && var.d(0) == 0.0) --dft; // 用量0のデータがある時
  if (opt.log() == logexc) dft -= var.noutd();     // 範囲外用量数を引く
  double sb = var.sxyw() * var.sxyw() / var.sxxw();
  if (opt.res("r"))
  {
    double r2 = (var.syyw() > 0.0) ? sb / var.syyw() : 0.0;
    fout << "  寄与率:r2乗=" << r2 << "\n";
  }
  if (opt.res("pr")) put_estlog(fout, opt, var);
  put_doseprtab(fout, "ロジット分析", var.syyw(), sb, dft);
}

void doseprclass::put_estlog(ostream& fout, option& opt, const varclass& var) const
// ロジット分析による用量の逆推定値を出力する
{
  if (sgn(var.b()) == 0) return; // 逆推定できず
  double pr = opt.pr() * 100.0, p = opt.pr();
  double lp = log(p / (1.0 - p));
  double m = (lp - var.a()) / var.b();
  fout << "  反応率=" << pr << "%(p=" << p << ")の時の用量:D" << pr << "=";
  if (opt.du() == dureal) fout << "x=" << m << "\n";
  else fout << pow(10.0, m) << "(x=log(D" << pr << ")=" << m << ")\n";
  double vr = 1.0, t = ptoz(1.0 - opt.cc());
  double g = t * t * vr / var.b() / var.b() / var.sxxw();
  if (g >= 1.0) g = 0.0;        // g>1の時は計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / var.snw() + (m - var.mxw()) * (m - var.mxw()) / var.sxxw());
  double ci = t / var.b() * sqrt(v);
  double ml = (m - g * var.mxw() - ci) / gr, mu = (m - g * var.mxw() + ci) / gr;
  fout << "  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal)
  {
    fout << "D" << pr << "下限=" << ml;
    fout << "  D" << pr << "上限=" << mu << "\n";
  }
  else
  {
    fout << "下限=" << pow(10.0, ml) << "(x=" << ml;
    fout << ")  上限=" << pow(10.0, mu) << "(x=" << mu << ")\n";
  }
}

void doseprclass::plot_ngp(ofstream& fout, option& opt, idd& id, varclass& var) const
// Ngraph用データ出力
{
  var.caldpline(opt);
  bool delml = false;
  if (opt.ngp("pro"))           // 反応率
  {
    fout << "-s1 -dmark,00 -o300 -zy0,1,0.2";
    if (opt.du() == dulog) fout << " -exlog";
    fout << "\n";
    plot_ngpid(fout, id);
    plot_ngppro(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("drc"))           // プロビット分析の用量反応曲線
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngpdrc(fout, opt, var);
  }
  if (opt.ngp("drl"))           // 傾向分析の用量反応直線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true;
    }
    var.calcaline(opt);
    plot_ngpid(fout, id);
    plot_ngpdrl(fout, opt, var);
  }
  if (opt.ngp("drlp"))          // ロジット分析の用量反応曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    var.callpline(opt);
    plot_ngpid(fout, id);
    plot_ngpdrlp(fout, opt, var);
  }
}

void doseprclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "%用量項目(縦):" << id.item(i) << "\n";
      break;
    default : fout << "%反応項目(横):" << id.item(i) << "\n";
    }
  }
}

void doseprclass::plot_ngppro(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に反応率を出力する
{
  fout << "%反応率\n";
  for (int i = 0; i < var.nd(); ++i)
  {
    if (opt.du() == dulog && var.d(i) <= 0.0) fout << "%";
    fout << var.d(i) << " " << var.p(i) << "\n";
  }
}

void doseprclass::plot_ngpdrc(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用にプロビット分析の用量反応率曲線を出力する
{
  fout << "%プロビット分析の用量反応率曲線  反復回数=" << var.nr();
  if (var.resvar() == resrmax) fout << "(最大回数)";
  fout << "\n";
  if (var.resvar() == reserr)
  {
    fout << "%計算エラー発生" << endl;
    return;
  }
  constexpr uint RDdef = 50;    // 用量の分割数
  double xmin = var.xmin(), xmax = var.xmax();
  double xmg = 0.1;             // 描画範囲のマージン
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { xmin = opt.xmin();  xmax = opt.xmax();  xmg = 0.0; }
  double range = xmax - xmin;
  double xstart = xmin - range * xmg, xend = xmax + range * xmg;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double y = var.a() + var.b() * x;
    double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
    double pro = nedp(y - 5.0);
    fout << dose << " " << pro << "\n";
  }
}

void doseprclass::plot_ngpdrl(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に傾向分析の用量反応率直線を出力する
{
  fout << "%傾向分析の用量反応率直線\n";
  if (var.resvar() == reserr)
  {
    fout << "%計算エラー発生" << endl;
    return;
  }
  double xmin = var.xmin(), xmax = var.xmax();
  double xmg = 0.1;             // 描画範囲のマージン
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { xmin = opt.xmin();  xmax = opt.xmax();  xmg = 0.0; }
  double range = xmax - xmin;
  double x = xmin - range * xmg;
  double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
  double pro = var.a() + var.b() * x;
  fout << dose << " " << pro << "\n";
  x = xmax + range * xmg;
  dose = (opt.du() == dulog) ? pow(10.0, x) : x;
  pro = var.a() + var.b() * x;
  fout << dose << " " << pro << endl;
}

void doseprclass::plot_ngpdrlp(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用にロジット分析の用量反応率曲線を出力する
{
  fout << "%ロジット分析の用量反応率曲線\n";
  fout << "%範囲外用量数=" << var.noutd() << "  範囲外用量計算法:";
  switch (opt.log())
  {
  case logwork : fout << "実用ロジット法";
    break;
  case logexc : fout << "除外法";
    break;
  default : fout << "反復補完法  補完回数=" << var.nr();
    if (var.resvar() == resrmax) fout << "(最大回数)";
  }
  fout << "\n";
  if (var.resvar() == reserr)
  {
    fout << "%計算エラー発生" << endl;
    return;
  }
  constexpr uint RDdef = 50;    // 用量の分割数
  double xmin = var.xmin(), xmax = var.xmax();
  double xmg = 0.1;             // 描画範囲のマージン
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { xmin = opt.xmin();  xmax = opt.xmax();  xmg = 0.0; }
  double range = xmax - xmin;
  double xstart = xmin - range * xmg, xend = xmax + range * xmg;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
    double pro = 1.0 / (1.0 + exp(-var.a() - var.b() * x));
    fout << dose << " " << pro << "\n";
  }
}

int doseprclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// クロス集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check() != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void doseprclass::put(option& opt, listidd& idlst, listftab& ftablst, int fno) const
// 結果を出力する
{
  listidd_itr iitr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    ftabi.trm();                // 行・列の合計が0のものを削除する
    varclass var(ftabi.nr()); // 変数オプジェクト
    var.cal(opt, ftabi);
    if (opt.res("res"))
    {
      put_resid(cout, *iitr);
      put_dosepr(cout, opt, ftabi, var);
      cout << "\n" << endl;
    }
    if (var.resvar() != reserr && opt.ngpname(fno) != "")
    {
      ofstream foutngp;
      if (ofopen(foutngp, (opt.ngpname(fno)).c_str()) == false) return; // ファイルオープン失敗
      var.cal(opt, ftabi); // 期待プロビットを初期値に戻すため
      if (var.resvar() != reserr) plot_ngp(foutngp, opt, *iitr, var);
      foutngp.close();
    }
    ++iitr;  ++fno;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  doseprclass dosepr;           // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nftab = 0, fno = 0;     // 読み込んだ度数表の数とファイル番号
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          nftab = dosepr.get(nullptr, opt, idlst, ftablst);
          dosepr.put(opt, idlst, ftablst, fno);
          fno += nftab;
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        nftab = dosepr.get(argv[i], opt, idlst, ftablst);  fok = false;
        dosepr.put(opt, idlst, ftablst, fno);
        fno += nftab;
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
