// doser.C Version 7.43  Copyleft (c) 杉本典夫
// Time-stamp: <2025-03-05 08:50:13 tac>
//
//  最初の項目を用量項目、次の項目を反応項目として用量反応解析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W doser doser.C
//               -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "doser  Version 7.43          Copyleft(Vo\\oV) 2025 杉本典夫",
  "書式:doser [<オプション> …]  [<データファイル> …]",
  "機能:最初の項目を用量、次の項目を反応として用量反応解析をするのだ。",
  "<オプション>",
  "  -c=<係数1>,<係数2>[,<係数3>,…]:最大対比法用係数を指定するのだ。",
  "    係数は合計が0にならんといかんのだ。(オプション省略時は最大対比法非実行)",
  "  -cc[=<信頼係数>]:信頼係数を指定するのだ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -dnsname[=<出力ファイル名>]:DANSデータを出力するファイル名を指定するのだ。",
  "    オプション省略時:DANSデータ出力せず  -dnsname:-dnsname=doser_dns.d",
  "  -du={l|r}:用量の単位を指定するのだ。",
  "    l:用量を対数変換して計算(オプション省略時)  r:実数のまま計算",
  "  -eps[=<収束誤差>]:反復計算の収束誤差を指定するのだ。(0<収束誤差)",
  "  -f=<ファイル>:オプションを指定したファイルから読み込むのだ。",
  "  -g[={-|+}]:Fiellerの式を使うかどうか指定するのだ。(-g:-g=+)",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するのだ。(-h:-h=man)",
  "  -h0=<初期値1>[,<初期値2>,…,<初期値p>]:パラメータの初期値を指定するのだ。",
  "  -meth={newton|non|simplex|etc}:非線形最小2乗法の計算法を指定するのだ。",
  "  -ngp={drc|drl|drly|mlyse|myse}[,…]:Ngraph用出力データを指定するのだ。",
  "    drc:用量反応曲線  drl:用量反応直線  drly:用量反応ロジット直線",
  "    mlyse:反応ロジットの用量別平均値±標準誤差  myse:反応の用量別平均値±標準誤差",
  "    オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データの出力ファイル名を指定するのだ。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=doser_ngp.dat",
  "  -res=[-][{def|all|drc|drl|drly|statly|staty|y[{Mh|my|<反応値>}]}]",
  "      [,…]:解析結果の出力項目を指定するのだ。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    drc:用量反応曲線  drl:用量反応直線  drly:用量反応ロジット直線",
  "    statly:反応ロジットの基礎統計量  staty:反応実測値の基礎統計量",
  "    y[{Mh|my|<反応値>}]:指定された反応値の用量逆推定値",
  "      Mh:ym+M/2(D50)  my:反応の総平均値  <反応値>:反応値指定",
  "  -rmax=<最大回数>:反復計算の最大回数を指定するのだ。(0≦最大回数)",
  "  -xrange=<最小値>-<最小値>:X軸(用量)の最小値と最大値を指定するのだ。",
  "  -ym=<a|d|最小値>:ロジスティック関数の最小値を指定するのだ。(-ym:-ym=a)",
  "    a:データから自動設定(オプション省略時)  d:微分回帰分析によって設定"
};

const string Title = "=== 用量反応解析 ==="; // 表題
const string DNSNAMEdef = "doser_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "doser_ngp.dat"; // 省略時Ngraphファイル名
constexpr double EPSdef = 1.0e-4; // 収束誤差省略時値
constexpr int RMAXdef = 200;      // 最大計算回数省略時値
constexpr int NPdef = 2;          // 用量反応関数のパラメータ数
constexpr double EPSly = 0.001;   // 範囲外ロジットの計算用微小値

using mapds = map<double, statv<double>>; // 用量別統計量マップの別名定義
using mapds_itr = mapds::iterator;        // 反復子の別名定義
using mapds_citr = mapds::const_iterator; // const反復子の別名定義

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

enum methtype                   // 非線形最小2乗法の種類
{
  methnon,                      // 指定なし
  methnewton,                   // Newton-Raphson法
  methsimplex,                  // Simplex法
  methetc                       // その他
};

double absmin(const vec<double>& hv) // 絶対値最小のパラメターの値を返す
{
  double hmin = FDmax;
  for (size_t i = 0;  i < hv.dim();  ++i)
  {
    double hi = fabs(hv(i));
    if (FDmin < hi && hi < hmin) hmin = hi;
  }
  return (hmin == FDmax) ? 1.0 : hmin;
}

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  lstr clst;                    // 最大対比法用係数リスト
  vecdbl vcvec;                 // 最大対比法用係数ベクトル
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  dutype vdu;                   // 用量の単位
  double veps;                  // 反復計算の収束誤差
  bool vg;                      // Fiellerの式使用フラッグ
  lstr h0lst;                   // パラメータ初期値リスト
  vecdbl vh0vec;                // パラメータ初期値ベクトル
  methtype vmeth;               // 非線形最小2乗法の種類
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  int vrmax;                    // 反復計算の最大回数
  pairdbl vxrange;              // X軸(用量)の最小値と最大値
  double vy;                    // 用量を推定する反応値
  double vym;                   // ロジスティック関数の最小値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  size_t csize() const          // 最大対比法用係数の数を返す
  { return vcvec.size(); }
  double c(size_t idx) const    // 最大対比法用係数を返す
  { return (idx < vcvec.size()) ? vcvec[idx] : 0.0; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  double eps() const            // 反復計算の収束誤差を返す
  { return veps; }
  size_t nh0() const            // パラメータの初期値数を返す
  { return vh0vec.size(); }
  double h0(size_t idx) const   // パラメータの初期値を返す
  { return (idx < vh0vec.size()) ? vh0vec[idx] : 0.0; }
  methtype meth() const         // 非線形最小2乗法の計算法を返す
  { return vmeth; }
  const string methname() const; // 計算法の名称を返す
  bool g() const                // Fiellerの式使用フラッグをす
  { return vg; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 反復計算の最大回数を返す
  { return vrmax; }
  double xmin() const           // X軸(用量)の最小値を返す
  { return vxrange.first; }
  double xmax() const           // X軸(用量)の最大値を返す
  { return vxrange.second; }
  double y() const              // 用量を推定する反応値を返す
  { return vy; }
  double ym() const             // ロジスティック関数の最小値を返す
  { return vym; }
  double ym(double ym0)         // ロジスティック関数の最小値を設定する
  { return (vym = ym0); }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), clst(), vcvec(), vcc(CCdef), vdnsname(), vdu(dulog), veps(EPSdef), h0lst(), vh0vec(), vmeth(methetc), vg(true), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vrmax(RMAXdef), vxrange(BLANKf, BLANKf), vy(BLANKf), vym(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_citr ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 最大対比法用係数の指定
  {
    if (strsplit(clst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "cc")      // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "du")      // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
   else if (opt[0] == "eps")     // 反復計算の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < FDeps) err = badopt(vpname, optstr, USAGE, FDeps, FDmax);
  }
 else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g")       // Fiellerの式使用指定
  {
    if (opt[1].empty() || opt[1] == "+") vg = true;
    else if (opt[1] == "-") vg = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "h0")      // パラメータの初期値の指定
  {
    if (strsplit(h0lst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "meth")    // 非線形最小2乗法の計算法の指定
  {
    if (opt[1].empty() || opt[1] == "etc") vmeth = methetc;
    else if (opt[1] == "newton") vmeth = methnewton;
    else if (opt[1] == "non") vmeth = methnon;
    else if (opt[1] == "simplex") vmeth = methsimplex;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 反復計算の最大回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "xrange")  // X軸(用量)の最小値と最大値の指定
  {
    if (getrange(vxrange.first, vxrange.second, opt[1]) == 0) // 範囲指定無
    { err = badopt(vpname, optstr, USAGE, -FDmax, FDmax); }
    else if (vxrange.first >= vxrange.second) // 範囲指定不適切
    { err = badopt(vpname, optstr, USAGE, -FDmax, FDmax); }
  }
  else if (opt[0] == "ym")      // ロジスティック関数の最小値指定
  {
    double fdata = 0.0;
    if (opt[1].empty() || opt[1] == "a") vym = BLANKf; // 自動設定(省略時と同じ)
    else if (opt[1] == "d") vym = -DBL_MAX; // 微分回帰分析によって設定
    else vym = fdata = strtodf(opt[1]);     // 最小値指定
    if (fdata == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)   // オプションのチェックとセット
{
  if (ni != 2) return disperr(vpname, "項目数は2つなのだ！");
  if (set_ngp() < 0) return -2; // -resオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (clst.size() > 0)          // 最大対比法の係数指定あり
  {
    vcvec.resize(clst.size());
    vcvec.assign(clst.size(), 0.0); // 初期値を0にする
    size_t i = 0;
    double sum = 0.0;
    for (const string& cstr : clst)
    {
      double fdata = strtodf(cstr);
      if (fdata == BLANKf) return disperr(vpname, "最大対比法用係数指定が間違っとるのだ→" + cstr);
      vcvec[i] = fdata;
      sum += fdata;  ++i;
    }
    if (sgn(sum) != 0) return disperr(vpname, "最大対比法用係数の合計が0にならないのだ！");
  }
  if (h0lst.size() < 1) return 0; // 初期値指定なし
  if ((int)(h0lst.size()) > NPdef) return disperr(vpname, "初期値の値が多すぎるのだ！");
  vh0vec.resize(h0lst.size());
  vh0vec.assign(h0lst.size(), 0.0); // 初期値を0にする
  size_t i = 0;
  for (const string& hstr : h0lst)
  {
    vh0vec[i] = strtodf(hstr);
    if (vh0vec[i] == BLANKf) return disperr(vpname, "初期値指定=" + hstr + " が不適切なのだ！");
    ++i;
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["drc"] = vngpmap["drl"] = vngpmap["drly"] = vngpmap["mlyse"] = vngpmap["myse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["drc"] = vngpmap["drl"] = vngpmap["drly"] = vngpmap["mlyse"] = vngpmap["myse"] = false;
    for (const string& ngpstr : vngplst)
    {
      mapsb_citr ngppos = vngpmap.find(ngpstr);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngpstr + "オプション指定が不適切！"); }
      else
      { vngpmap[ngpstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["drl"] = vresmap["staty"] = vresmap["y"] = true;
  vresmap["drc"] = vresmap["drly"] = vresmap["statly"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["drl"] = vresmap["staty"] = vresmap["y"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      {
        vresmap["def"] = vresmap["drl"] = vresmap["staty"] = vresmap["y"] = true;
      }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["drc"] = vresmap["drl"] = vresmap["drly"] = vresmap["statly"] = vresmap["staty"] = vresmap["y"] = true;
      }
      else if (resstr == "yMh")
      { vresmap["y"] = true;  vy = BLANKf; }
      else if (resstr == "ymy")
      { vresmap["y"] = true;  vy = -DBL_MAX; }
      else if (resstr[0] == 'y' && resstr.size() > 1)
      {
        vy = strtodf(resstr.substr(1)); // 反応値
        if (vy == BLANKf) return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！");
        vresmap["y"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

const string option::methname() const // 計算法の名称を返す
{
  switch (vmeth)
  {
  case methnewton : return "Newton法";
  case methnon : return "なし";
  case methsimplex : return "滑降Simplex法";
  default : return "滑降Simplex+Newton法";
  }
}

class varclass                  // 変数クラス
{
private:
  mapds vdysmap;                // 用量別反応統計量マップ
  pairdbl drange;               // 用量の最小値と最大値
  pairdbl yrange;               // 反応の最小値と最大値
  statcor<double> dystatc;      // 用量と反応の相関用基礎統計量
  restype vresdrl;              // 用量反応直線の処理結果フラッグ
  int vndrl;                    // 用量反応直線の用量数
  int vdfb;                     // 用量反応直線の回帰自由度
  double vsb;                   // 用量反応直線の回帰平方和
  int vdfl;                     // 用量反応直線のズレ自由度
  double vsl;                   // 用量反応直線のズレ平方和
  int vdfd;                     // 用量反応直線の用量自由度
  double vsd;                   // 用量反応直線の用量平方和
  int vdfr;                     // 用量反応直線の残差自由度
  double vsr;                   // 用量反応直線の残差平方和
  int vdft;                     // 用量反応直線の全自由度
  double vst;                   // 用量反応直線の全平方和
  double va;                    // 用量反応直線の切片
  double vb;                    // 用量反応直線の回帰係数
  double vseb;                  // 用量反応直線の回帰係数標準誤差
  mapds vdlysmap;               // 用量別反応ロジット統計量マップ
  statcor<double> dlystatc;     // 用量と反応ロジットの相関用基礎統計量
  restype vresdrly;             // 用量反応ロジット直線の処理結果フラッグ
  int vnoutdly;                 // 用量反応ロジット直線の範囲外データ数
  int vndrly;                   // 用量反応ロジット直線の用量数
  int vdfbly;                   // 用量反応ロジット直線の回帰自由度
  double vsbly;                 // 用量反応ロジット直線の回帰平方和
  int vdflly;                   // 用量反応ロジット直線のズレ自由度
  double vslly;                 // 用量反応ロジット直線のズレ平方和
  int vdfdly;                   // 用量反応ロジット直線の用量自由度
  double vsdly;                 // 用量反応ロジット直線の用量平方和
  int vdfrly;                   // 用量反応ロジット直線の残差自由度
  double vsrly;                 // 用量反応ロジット直線の残差平方和
  int vdftly;                   // 用量反応ロジット直線の全自由度
  double vstly;                 // 用量反応ロジット直線の全平方和
  double vym;                   // 用量反応ロジット直線の最小値
  double vy0;                   // 用量反応ロジット直線のy切片
  double vy00;                  // 用量反応ロジット直線のy切片初期値
  double vc;                    // 用量反応ロジット直線の回帰係数
  double vc0;                   // 用量反応ロジット直線の回帰係数初期値
  double vm;                    // 用量反応ロジット直線の変動最大値
  double vsebly;                // 用量反応直線の回帰係数標準誤差
  restype vresdrc;              // 用量反応曲線の処理結果フラッグ
  int vnrdrc;                   // 用量反応曲線の繰り返し計算数
  int vnpdrc;                   // 用量反応曲線のパラメータ数
  vec<double> vh0v;             // 用量反応曲線のパラメータ初期値ベクトル
  vec<double> vhv;              // 用量反応曲線のパラメータ推測値ベクトル
  double vqerr;                 // 用量反応曲線の回帰誤差
  int vdfbdrc;                  // 用量反応曲線の回帰自由度
  double vsbdrc;                // 用量反応曲線の回帰平方和
  int vdfldrc;                  // 用量反応曲線のズレ自由度
  double vsldrc;                // 用量反応曲線のズレ平方和
  int vdfddrc;                  // 用量反応曲線の用量自由度
  double vsddrc;                // 用量反応曲線の用量平方和
  int vdfrdrc;                  // 用量反応曲線の残差自由度
  double vsrdrc;                // 用量反応曲線の残差平方和
  int vdftdrc;                  // 用量反応曲線の全自由度
  double vstdrc;                // 用量反応曲線の全平方和
  restype vresvar;              // 全体の処理結果フラッグ
  restype cal_drl(option& opt); // 用量反応直線を計算する
  restype cal_drly(option& opt, ldata datalst); // 用量反応ロジット直線を計算する
  int init_drly(option& opt);   // パラメータの初期値を設定する
  int cal_smat(mat<double>& smat, option& opt) const;
  // 2次回帰分析用に積和行列を計算する
  restype cal_drly0(double& y0, double ym, double c, double m, option& opt) const;
  // 用量反応ロジット直線のy0をロジット分析によって求める
  restype cal_dslymap(option& opt, ldata datalst);
  // 最終的なロジット分析を行って用量別反応ロジットマップをセットする
  restype cal_drc(option& opt, ldata datalst); // 用量反応曲線を計算する
  restype cal_simplex(int& nr, vec<double>& hv, option& opt, ldata& datalst) const;
  // Simplex法による非線形最小2乗計算
  void init_simplex(mat<double>& simplexmat, const vec<double>& hv) const;
  // シンプレックスの初期設定
  restype check_simplex(mat<double>& simplexmat, vec<double>& hlv, vec<double>& hmv, vec<double>& hhv, double& fl, double& fs, double& fh, int& lidx, int& hidx, int nr, option& opt, ldata& datalst) const;
  // fiの計算と終了基準のチェック
  restype check_hv(const vec<double>& hv) const; // パラメータの発散チェック
  restype cal_newton(int& nr, vec<double>& hv, option& opt, ldata& datalst) const;
  // Newton-Raphson法
  void cal_gh(vec<double>& gradv, mat<double>& hessmat, const vec<double>& hv, option& opt, ldata& datalst) const;
  // 傾斜ベクトルとHessの行列の計算
  restype check_q(double qnew, double qold, const vec<double>& hv, int nr, option& opt) const;
  // Q(h)の計算と終了基準のチェック
  double fq(const vec<double>& hv, option& opt, ldata& datalst) const;
  // 誤差関数
  double fqd(int hidx, const vec<double>& hv, option& opt, ldata& datalst) const;
  // 誤差の偏導関数
  double fqd2(int hiidx, int hjidx, const vec<double>& hv, option& opt, ldata& datalst) const;
  // 誤差の2階偏導関数
  double func(double x, const vec<double>& hv) const;
  // 用量反応関数
  double funcd(double x, int hidx, const vec<double>& hv) const;
  // モデル偏導関数
  double funcd2(double x, int hiidx, int hjidx, const vec<double>& hv) const;
  // モデル2階偏導関数
  void cal_ssdrc(option& opt, ldata& datalst);
  // 用量反応曲線の各種の平方和を計算する
public:
  varclass() : vdysmap(), drange(FDmax, -FDmax), yrange(FDmax, -FDmax), dystatc(), vresdrl(resnon), vndrl(0), vdfb(0), vsb(0.0), vdfl(0), vsl(0.0), vdfd(0), vsd(0.0), vdfr(0), vsr(0.0), vdft(0), vst(0.0), va(0.0), vb(0.0), vseb(0.0), vdlysmap(), dlystatc(), vresdrly(resnon), vnoutdly(0), vndrly(0), vym(0.0), vy0(0.0), vy00(0.0), vc(0.0), vc0(0.0), vm(0.0), vresdrc(resnon), vnrdrc(0), vnpdrc(NPdef), vh0v(NPdef), vhv(NPdef), vqerr(0.0), vdfbdrc(0), vsbdrc(0.0), vdfldrc(0), vsldrc(0.0), vdfddrc(0), vsddrc(0.0), vdfrdrc(0), vsrdrc(0.0), vdftdrc(0), vstdrc(0.0), vresvar(resnon)
  // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  void cal(option& opt, ldata datalst); // データを読み込み各種の用量反応解析を行う
  const mapds& dysmap() const   // 用量別統計量マップを返す
  { return vdysmap; }
  double dmin() const           // 用量最小値を返す
  { return drange.first; }
  double dmax() const           // 用量最大値を返す
  { return drange.second; }
  double ymin() const           // 反応最小値を返す
  { return yrange.first; }
  double ymax() const           // 反応最大値を返す
  { return yrange.second; }
  int n() const                 // 用量と反応の例数を返す
  { return dystatc.n(); }
  double mx() const             // 用量と反応の用量平均値を返す
  { return dystatc.mx(); }
  double my() const             // 用量と反応の反応平均値を返す
  { return dystatc.my(); }
  double sxx() const            // 用量と反応の用量平方和を返す
  { return dystatc.sxx(); }
  double syy() const            // 用量と反応の反応平方和を返す
  { return dystatc.syy(); }
  double sdy() const            // 用量と反応の反応標準偏差を返す
  { return dystatc.sdy(); }
  double sey() const            // 用量と反応の反応標準誤差を返す
  { return dystatc.sey(); }
  double sxy() const            // 用量と反応の用量と反応積和を返す
  { return dystatc.sxy(); }
  restype resdrl() const        // 用量反応直線の処理結果フラッグを返す
  { return vresdrl; }
  int ndrl() const              // 用量反応直線の用量数を返す
  { return vndrl; }
  int dfb() const               // 用量反応直線の回帰自由度を返す
  { return vdfb; }
  double sb() const             // 用量反応直線の回帰平方和を返す
  { return vsb; }
   int dfl() const              // 用量反応直線のズレ自由度を返す
  { return vdfl; }
  double sl() const             // 用量反応直線のズレ平方和を返す
  { return vsl; }
  int dfd() const               // 用量反応直線の用量自由度を返す
  { return vdfd; }
  double sd() const             // 用量反応直線の用量平方和を返す
  { return vsd; }
  int dfr() const               // 用量反応直線の残差自由度を返す
  { return vdfr; }
  double sr() const             // 用量反応直線の残差平方和を返す
  { return vsr; }
  int dft() const               // 用量反応直線の全自由度を返す
  { return vdft; }
  double st() const             // 用量反応直線の全平方和を返す
  { return vst; }
  double vr() const             // 用量反応直線の残差分散を返す
  { return (vdfr > 0) ? vsr / (double)vdfr : 0.0; }
  double a() const              // 用量反応直線の切片を返す
  { return va; }
  double b() const              // 用量反応直線の回帰係数を返す
  { return vb; }
  double seb() const            // 用量反応直線の回帰係数標準誤差を返す
  { return vseb; }
  const mapds& dlysmap() const  // 用量別反応ロジット統計量マップを返す
  { return vdlysmap; }
  int lyn() const               // 用量と反応ロジットの例数を返す
  { return dlystatc.n(); }
  double lymx() const           // 用量と反応ロジットの用量平均値を返す
  { return dlystatc.mx(); }
  double lymy() const           // 用量と反応ロジットの反応ロジット平均値を返す
  { return dlystatc.my(); }
  double lysxx() const          // 用量と反応ロジットの用量平方和を返す
  { return dlystatc.sxx(); }
  double lysyy() const          // 用量と反応ロジットの反応ロジット平方和を返す
  { return dlystatc.syy(); }
  double lysdy() const          // 用量と反応ロジットの反応ロジット標準偏差を返す
  { return dlystatc.sdy(); }
  double lysey() const          // 用量と反応ロジットの反応ロジット標準誤差を返す
  { return dlystatc.sey(); }
  double lysxy() const          // 用量と反応ロジットの用量と反応ロジット積和を返す
  { return dlystatc.sxy(); }
  restype resdrly() const       // 用量反応ロジット直線の処理結果フラッグを返す
  { return vresdrly; }
  int noutdly() const           // 用量反応ロジット直線の範囲外データ数を返す
  { return vnoutdly; }
  int ndrly() const             // 用量反応ロジット直線の用量数を返す
  { return vndrly; }
  int dfbly() const             // 用量反応ロジット直線の回帰自由度を返す
  { return vdfbly; }
  double sbly() const           // 用量反応ロジット直線の回帰平方和を返す
  { return vsbly; }
   int dflly() const            // 用量反応ロジット直線のズレ自由度を返す
  { return vdflly; }
  double slly() const           // 用量反応ロジット直線のズレ平方和を返す
  { return vslly; }
  int dfdly() const             // 用量反応ロジット直線の用量自由度を返す
  { return vdfdly; }
  double sdly() const           // 用量反応ロジット直線の用量平方和を返す
  { return vsdly; }
  int dfrly() const             // 用量反応ロジット直線の残差自由度を返す
  { return vdfrly; }
  double srly() const           // 用量反応ロジット直線の残差平方和を返す
  { return vsrly; }
  int dftly() const             // 用量反応ロジット直線の全自由度を返す
  { return vdftly; }
  double stly() const           // 用量反応ロジット直線の全平方和を返す
  { return vstly; }
  double vrly() const           // 用量反応ロジット直線の残差分散を返す
  { return (vdfrly > 0) ? vsrly / (double)vdfrly : 0.0; }
  double ym() const             // 用量反応ロジット直線の最小値を返す
  { return vym; }
  double y0() const             // 用量反応ロジット直線のy切片を返す
  { return vy0; }
  double y00() const            // 用量反応ロジット直線のy切片初期値を返す
  { return vy00; }
  double c() const              // 用量反応ロジット直線の回帰係数を返す
  { return vc; }
  double c0() const             // 用量反応ロジット直線の回帰係数初期値を返す
  { return vc0; }
  double m() const              // 用量反応ロジット直線の変動最大値を返す
  { return vm; }
  restype resdrc() const        // 用量反応曲線の処理結果フラッグを返す
  { return vresdrc; }
  int nrdrc() const             // 用量反応曲線の繰り返し計算数を返す
  { return vnrdrc; }
  int npdrc() const             // 用量反応曲線のパラメータ数を返す(将来的な変更用)
  { return vnpdrc; }
  double ymdrc() const          // 用量反応曲線の最小値を返す(固定)
  { return vym; }
  double mdrc() const           // 用量反応曲線の変動最大値を返す(固定)
  { return vm; }
  double y00drc() const         // 用量反応曲線のy切片初期値を返す
  { return vh0v(0); }
  double y0drc() const          // 用量反応曲線のy切片を返す
  { return vhv(0); }
  double c0drc() const          // 用量反応曲線の回帰係数初期値を返す
  { return vh0v(1); }
  double cdrc() const           // 用量反応曲線の回帰係数を返す
  { return vhv(1); }
  double qerr() const           // 用量反応曲線の回帰誤差を返す
  { return vqerr; }
  int ndrc() const              // 用量反応曲線の用量数を返す
  { return vndrl; }
  int dfbdrc() const            // 用量反応曲線の回帰自由度を返す
  { return vdfbdrc; }
  double sbdrc() const          // 用量反応曲線の回帰平方和を返す
  { return vsbdrc; }
   int dfldrc() const           // 用量反応曲線のズレ自由度を返す
  { return vdfldrc; }
  double sldrc() const          // 用量反応曲線のズレ平方和を返す
  { return vsldrc; }
  int dfddrc() const            // 用量反応曲線の用量自由度を返す
  { return vdfddrc; }
  double sddrc() const          // 用量反応曲線の用量平方和を返す
  { return vsddrc; }
  int dfrdrc() const            // 用量反応曲線の残差自由度を返す
  { return vdfrdrc; }
  double srdrc() const          // 用量反応曲線の残差平方和を返す
  { return vsrdrc; }
  int dftdrc() const            // 用量反応曲線の全自由度を返す
  { return vdftdrc; }
  double stdrc() const          // 用量反応曲線の全平方和を返す
  { return vstdrc; }
  double vrdrc() const          // 用量反応曲線の残差分散を返す
  { return (vdfrdrc > 0) ? vsrdrc / (double)vdfrdrc : 0.0; }
  double fdy(double x) const    // 用量反応関数(vhvを使ったもの)
  { return func(x, vhv); }
  restype resvar() const        // 全体の処理結果フラッグを返す
  { return vresvar; }
};

void varclass::cal(option& opt, ldata datalst)
// データを読み込み各種の用量反応解析を行う
{
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (dose != BLANKf && y != BLANKf) // dose=0も集計する
    {
      mapds_itr dysitr = vdysmap.find(dose);
      if (dysitr == vdysmap.end()) // 新規用量
      {
        statv<double> statvi(y);
        vdysmap.insert(make_pair(dose, statvi));
      }
      else                      // すでにある用量
      { dysitr->second.sigma(y);  }
      if (opt.du() == dureal || dose > 0.0)
      {
        if (dose < drange.first) drange.first = dose;
        if (dose > drange.second) drange.second = dose;
        if (y < yrange.first) yrange.first = y;
        if (y > yrange.second) yrange.second = y;
        double mdose = (opt.du() == dulog) ? log10(dose) : dose; // メタ用量
        dystatc.sigma(mdose, y);
      }
    }
  }
  if (dystatc.n() < 1)          // 用量反応データなし
  {
    vresvar = reserr;
    return;
  }
  vresdrl = cal_drl(opt);
  vresdrly = cal_drly(opt, datalst);
  vresdrc = cal_drc(opt, datalst);
  if (vresdrl == reserr && vresdrly == reserr && vresdrc == reserr) vresvar = reserr;
  else vresvar = resok;
}

restype varclass::cal_drl(option& opt)
// 用量反応直線を計算する
{
  vsd = 0.0;  vndrl = 0;  vym = BLANKf;
  for (auto& dystat : vdysmap)
  {
    double dose = dystat.first, y = dystat.second.mean();
    if (sgn(dose) == 0 && opt.du() == dulog && opt.ym() == BLANKf) vym = y;
    if (opt.du() == dureal || dose > 0.0)
    {
      ++vndrl;
      vsd += dystat.second.n() * y * y;
    }
  }
  if (vndrl < NPdef) return reserr; // 用量数不足で計算できない
  vsd -= dystatc.n() * dystatc.my() * dystatc.my();
  if (vsd < FDeps) vsd = 0.0;
  vsb = (dystatc.sxx() > 0.0) ? dystatc.sxy() * dystatc.sxy() / dystatc.sxx() : 0.0;
  if (vsb < FDeps) vsb = 0.0;
  vsr = dystatc.syy() - vsd;
  if (vsr < FDeps) vsr = 0.0;
  vdft = dystatc.n() - 1;  vdfd = vndrl - 1;  vdfb = 1;
  vdfl = vdfd - vdfb;  vdfr = vdft - vdfd;
  vsl = vsd - vsb;
  if (vdfl == 0 || vsl < FDeps) vsl = 0.0;
  vst = dystatc.syy();
  vb = (dystatc.sxx() > 0.0) ? dystatc.sxy() / dystatc.sxx() : 0.0;
  va = dystatc.my() - vb * dystatc.mx();
  if (vdfr < 1)                 // 用量データが1例のみの時
  {
    vdfr = vdft -  vdfb;
    vsr = vsd - vsb;
  }
  vseb = (vdfr > 0 && dystatc.sxx() > 0.0) ? sqrt(vsr / (double)vdfr / dystatc.sxx()) : 0.0;
  return resok;                 // 用量反応直線を計算できた
}

restype varclass::cal_drly(option& opt, ldata datalst)
// 用量反応ロジット直線を計算する
{
  vndrly = init_drly(opt);
  if (vndrly < NPdef) return reserr; // 用量数不足で計算できない
  restype resdsly = cal_dslymap(opt, datalst);
  if (resdsly == reserr) return resdsly; // ロジット分析ができない
  vsdly = 0.0;
  for (auto& dlystat : vdlysmap)
  {
    double ly = dlystat.second.mean();
    vsdly += (double)(dlystat.second.n()) * ly * ly;
  }
  vsdly -= (double)(dlystatc.n()) * dlystatc.my() * dlystatc.my();
  if (vsdly < FDeps) vsdly = 0.0;
  vsbly = (dlystatc.sxx() > 0.0) ? dlystatc.sxy() * dlystatc.sxy() / dlystatc.sxx() : 0.0;
  if (vsbly < FDeps) vsbly = 0.0;
  vsrly = dlystatc.syy() - vsdly;
  if (vsrly < FDeps) vsrly = 0.0;
  vdftly = dlystatc.n() - 1;  vdfdly = vndrly - 1;  vdfbly = 1;
  vdflly = vdfdly - vdfbly;  vdfrly = vdftly - vdfdly;
  vslly = vsdly - vsbly;
  vstly = dlystatc.syy();
  if (vdflly == 0 || vslly < FDeps) vslly = 0.0;
  if (vdfrly < 1)               // 用量データが1例のみの時
  {
    vdfrly = vdftly -  vdfbly;
    vsrly = vsdly - vsbly;
  }
  vsebly = (vdfrly > 0 && dlystatc.sxx() > 0.0) ? sqrt(vsrly / (double)vdfrly / dlystatc.sxx()) : 0.0;
  return resok;                 // ロジット分析ができた
}

int varclass::init_drly(option& opt)
// 用量反応ロジット直線のパラメータ初期値を設定する
{
  vy0 = vy00 = (yrange.first > 0.0) ? yrange.first * 0.5 : 0.0; // 最小値
  vm = yrange.second - vy0;        // 最大値−最小値
  vc = vc0 = (sgn(vm) == 0) ? 0.0 : 8.0 * vb / vm; // 用量反応直線の回帰係数によって設定
  if (opt.ym() == -DBL_MAX)     // ymを微分回帰曲線から求める
  { vym = 0.0; }
  else if (opt.ym() == BLANKf)  // ym:データから自動設定
  {
    if (vym == BLANKf || sgn(vc0) < 0) // ym設定またはc<0
    { opt.ym(-DBL_MAX);  vym = 0.0; }
  }
  else                          // ym:指定値
  { vym = opt.ym(); }
  vm = yrange.second - vym;
  int np = 2;                   // 微分回帰曲線のパラメータ数
  if (opt.ym() == -DBL_MAX) np = 3; // ymを微分回帰曲線から求める
  int dim = np + 1, dyidx = np;
  mat<double> smat(dim);        // 積和行列
  int nd = cal_smat(smat, opt);
  if (nd - 2 < np) return nd;   // 微分回帰分析不可能、データから求めた初期値を返す
  double b0 = 0.0, b1 = 0.0, b2 = 0.0, ym0 = vym, m0 = vm;
  if (opt.ym() == -DBL_MAX)     // ymを微分回帰曲線から求める
  {
    b0 = smat(0, dyidx);  b1 = smat(1, dyidx);  b2 = smat(2, dyidx);
    if (sgn(b1 * b2) >= 0) return nd; // 不適解
    double tmp = b1 * b1 - 4.0 * b2 * b0;
    tmp = (sgn(tmp) > 0) ? sqrt(tmp) : 0.0;
    ym0 = (-b1 + tmp) / (2.0 * b2);
    double ymax = (-b1 - tmp) / (2.0 * b2);
    if (sgndiff(ym0, ymax) < 0)
    { m0 = ymax - ym0; }
    else
    {
      m0 = ym0 - ymax;
      ym0 = ymax;
    }
    vc0 = -b2 * m0;
  }
  else                          // ymをデータから求める／-ymで指定
  {
    b1 = smat(0, dyidx);  b2 = smat(1, dyidx);
    if (sgn(b1 * b2) >= 0) return nd; // 不適解
    m0 = -b1 / b2;  vc0 = b1;  ym0 = vym;
  }
  if (cal_drly0(vy00, ym0, vc0, m0, opt) != resok) return nd; // vy00が不適切
  vym = ym0;  vy0 = vy00;  vc = vc0;  vm = m0;
  return nd;                    // 有効な用量数返す
}

int varclass::cal_smat(mat<double>& smat, option& opt) const
// 2次回帰分析用に積和行列を計算する
{
  size_t dim = smat.r();        // 行列の次元
  int np = dim - 1;             // 微分回帰曲線のパラメータ数
  vecdbl fdatavec(dim);
  size_t offset = 0;
  if (dim == 4) { fdatavec[0] = 1.0;  offset = 1; } // 最小値も推定する時
  double md1 = 0.0, md2 = 0.0, y1 = 0.0, y2 = 0.0;
  int nd = 0;
  for (auto& dystat : vdysmap)
  {
    double dose = dystat.first, y = dystat.second.mean();
    if (opt.du() == dureal || dose > 0.0)
    {
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      if (nd < 1)
      { md1 = mdose;  y1 = y;  nd = 1; }
      else if (nd == 1)
      { md2 = mdose;  y2 = y;  nd = 2; }
      else
      {
        double mdd1 = md2 - md1, mdd2 = mdose - md2;
        double eps = DBL_EPSILON * fmax(5.0, fabs(mdd1)); // 計算誤差
        double w1 = (fabs(mdd1) < eps) ? 0.0 : 1.0 / mdd1;
        double w2 = (fabs(mdd2) < eps) ? 0.0 : 1.0 / mdd2;
        double b1 = (y2 - y1) * w1, b2 = (y - y2) * w2;
        w1 = fabs(w1);  w2 = fabs(w2);
        double dy = (fabs(w1 + w2) < eps) ? 0.0 : (b1 * w1 + b2 * w2) / (w1 + w2);
        double my = y2;
        if (opt.ym() != -DBL_MAX) my -= vym; // 仮最小値を引く
        double fd = my;               // 例数で重み付けしない
        for (size_t i = 0; i < 2; ++i)
        {
          fdatavec[offset + i] = fd;
          fd *= my;
        }
        fdatavec[np] = dy;
        for (size_t i = 0; i < dim; ++i)
        {
          fd = fdatavec[i];
          for (size_t j = i; j < dim; ++j) smat.add(i, j, fd * fdatavec[j]);
        }
        md1 = md2;  y1 = y2;
        md2 = mdose;  y2 = y;  ++nd;
      }
    }
  }
  for (int i = 0; i < np; ++i)
  {
    for (int j = i + 1; j <= np; ++j) smat.assign(j, i, smat(i, j));
  }
  for (int i = 0; i < np; ++i) smat.sweep(i);
  return nd;                    // 有効な用量数を返す
}

restype varclass::cal_drly0(double& y0, double ym, double c, double m, option& opt) const
// 用量反応ロジット直線のy0をロジット分析によって求める
{
  statcor<double> dlystatc0;
  for (auto& dystat : vdysmap)
  {
    double dose = dystat.first, my = dystat.second.mean() - ym;
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      double ly = 0.0;
      if (my <= 0.0) ly = log(EPSly / (1.0 - EPSly)); // 範囲外データを近似
      else if (m <= my) ly = log((1.0 - EPSly) / EPSly); // 範囲外データを近似
      else ly = log(my / (m - my));
      dlystatc0.sigma(mdose, ly); // 例数で重み付けしない
    }
  }
  if (dlystatc0.n() < 1) return reserr; // ロジットがない時
  double ly0 = dlystatc0.my() - c * dlystatc0.mx();
  y0 = ym + m / (1.0 + exp(-ly0));
  return resok;                     // y0が計算できた
}

restype varclass::cal_dslymap(option& opt, ldata datalst)
// ロジット分析を行って用量別反応ロジットマップをセットする
{
  dlystatc.clear();
  vnoutdly = 0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), yo = cdatai.fdata(1);
    if (dose != BLANKf && yo != BLANKf)
    {
      if (opt.du() == dureal || sgn(dose) > 0)
      {
        double mdose = (opt.du() == dulog) ? log10(dose) : dose;
        double y = yo - vym, ly = 0.0;
        if (y <= 0.0) { ly = log(EPSly / (1.0 - EPSly));  ++vnoutdly; }
        else if (vm <= y) { ly = log((1.0 - EPSly) / EPSly);  ++vnoutdly; }
        else ly = log(y / (vm - y));
        dlystatc.sigma(mdose, ly);
        mapds_itr dslyitr = vdlysmap.find(dose);
        if (dslyitr == vdlysmap.end()) // 新規用量
        {
          statv<double> statvi(ly);
          vdlysmap.insert(make_pair(dose, statvi));
        }
        else                      // すでにある用量
        { dslyitr->second.sigma(ly);  }
      }
    }
  }
  if (dlystatc.n() < 1) return reserr; // ロジットがない時
  vc = (sgn(dlystatc.sxx()) > 0) ? dlystatc.sxy() / dlystatc.sxx() : 0.0;
  double ly0 = dlystatc.my() - vc * dlystatc.mx();
  vy0 = vym + vm / (1.0 + exp(-ly0));
  return resok;                 // ロジット分析実行
}

restype varclass::cal_drc(option& opt, ldata datalst)
// 用量反応曲線を計算する
{
  vh0v.assign(0, vy0);  vh0v.assign(1, vc); // ロジット分析の結果をパラメータ初期値にする
  if (opt.nh0() > 0)            // 初期値が指定されている場合
  {
    for (size_t i = 0; i < opt.nh0(); ++i) vh0v.assign(i, opt.h0(i));
    if (sgndiff(vh0v(0), vym) < 0 || sgndiff(vh0v(0) - vym, vm) > 0) return reserr; // 不適切なy0
  }
  vhv = vh0v;  vnrdrc = 0;
  if (vndrl < vnpdrc) return reserr; // 用量数不足で計算できない
  restype resdrc = resnon;
  if (opt.meth() == methetc || opt.meth() == methsimplex)
  {  // 収束を良くするためSimplex法で1次近似計算をする
    resdrc = cal_simplex(vnrdrc, vhv, opt, datalst);
    if (resdrc == reserr) return resdrc; // 収束失敗
  }
  if (opt.meth() == methetc || opt.meth() == methnewton)
  {
    int nr = 0;
    resdrc = cal_newton(nr, vhv, opt, datalst);
    vnrdrc += nr;
    if (resdrc == reserr) return resdrc; // 収束失敗
  }
  cal_ssdrc(opt, datalst);
  return resdrc;
}

restype varclass::cal_simplex(int& nr, vec<double>& hv, option& opt, ldata& datalst) const
// Simplex法による非線形最小2乗計算
{
  int np = hv.dim();            // パラメーター数
  mat<double> simplexmat(np, np + 1); // シンプレックスを行列で表す
  vec<double> hlv(hv), hmv(hv), hhv(hv);
  double fl = 0.0, fs = 0.0, fh = 0.0; // 最小値、2番目に大きい値、最大値
  int lidx = 0, hidx = 0;
  restype ressimp = resnon;
  nr = 0;
  init_simplex(simplexmat, hv);
  while ((ressimp = check_simplex(simplexmat, hlv, hmv, hhv, fl, fs, fh, lidx, hidx, nr, opt, datalst)) == resnon)
  {
    vec<double> hrv(hmv);
    hrv *= 2.0;
    hrv -= hhv;
    double fr = fq(hrv, opt, datalst);
    if (fr < fl)                // 折り返し成功
    {
      vec<double> hev(hrv);
      hev *= 2.0;
      hev -= hmv;
      double fe = fq(hev, opt, datalst);
      if (fe < fr)              // 拡大成功
      { for (int i = 0; i < np; ++i) simplexmat.assign(i, hidx, hev(i)); }
      else                      // 拡大失敗(行き過ぎ)
      { for (int i = 0; i < np; ++i) simplexmat.assign(i, hidx, hrv(i)); }
    }
    else if (fr <= fs)          // 斜め方向の折り返し
    { for (int i = 0; i < np; ++i) simplexmat.assign(i, hidx, hrv(i)); }
    else                        // 縮小失敗、Simplex移動
    {
      if (fr < fh) { fh = fr;  hhv = hrv; } // 折り返し後のSimplex中に最小点
      vec hcv(hmv + hhv);
      hcv *= 0.5;
      double fc = fq(hcv, opt, datalst);
      if (fc < fh)              // 縮小成功
      { for (int i = 0; i < np; ++i) simplexmat.assign(i, hidx, hcv(i)); }
      else                      // 縮小失敗、Simplex移動
      {
        for (int j = 0; j <= np; ++j)
        {
          if (j != lidx)
          {
            for (int i = 0; i < np; ++i) simplexmat.assign(i, j, (simplexmat(i, j) + hlv(i)) * 0.5);
          }
        }
      }
    }
    ++nr;
  }
  if (ressimp == reserr || fabs(fl) > FDmax) return reserr; // 発散
  hv = hlv;
  return ressimp;               // 計算できた
}

void varclass::init_simplex(mat<double>& simplexmat, const vec<double>& hv) const
// シンプレックスの初期設定
{
  constexpr double Sqrt2 = sqrt(2.0); // コンパイル時定数
  constexpr double Rs = 0.1;          // シンプレックスの1辺の相対割合
  int np = hv.dim();
  double s = absmin(hv) * Rs;
  double v = (sqrt((double)(np + 1)) - 1.0 + (double)np) / (double)np / Sqrt2 * s;
  double w = (sqrt((double)(np + 1)) - 1.0) / (double)np / Sqrt2 * s;
  for (int i = 0; i < np; ++i) simplexmat.assign(i, 0, hv(i));
  for (int j = 1; j <= np; ++j)
  {
    for (int i = 0; i < np; ++i) simplexmat.assign(i, j, (i == j - 1) ? v + hv(i) : w + hv(i));
  }
}

restype varclass::check_simplex(mat<double>& simplexmat, vec<double>& hlv, vec<double>& hmv, vec<double>& hhv, double& fl, double& fs, double& fh, int& lidx, int& hidx, int nr, option& opt, ldata& datalst) const
// fiの計算と終了基準のチェック
{
  int np = hlv.dim();
  int eidx = np - 1;
  vecdbl fvec(np + 1, 0.0);
  fl = FDmax;  fh = -FDmax;
  for (int j = 0; j <= np; ++j)
  {
    vec<double> hjv(simplexmat.sub(0, j, eidx, j));
    double fj = fq(hjv, opt, datalst);
    fvec[j] = fj;
    if (fj < fl) { fl = fj;  lidx = j; }
    if (fj > fh) { fh = fj;  hidx = j; }
  }
  if (fl > FDmax || isnan(fl)) return reserr; // 発散
  hlv = simplexmat.sub(0, lidx, eidx, lidx); // 最小値の時のhv
  hhv = simplexmat.sub(0, hidx, eidx, hidx); // 最大値の時のhv
  if (check_hv(hlv) == reserr) return reserr; // 発散
  double ferr = fabs(fh - fl);
  if (ferr < fabs(fl * opt.eps())) return resok; // 収束
  if (opt.rmax() > 0 && nr >= opt.rmax()) return resrmax; // 反復回数最大値
  fs = fl;
  hmv = 0.0;
  for (int j = 0; j <= np; ++j)
  {
    if (j != hidx)
    {
      if (fvec[j] > fs) fs = fvec[j];
      hmv += simplexmat.sub(0, j, eidx, j);
    }
  }
  hmv *= 1.0 / (double)np;       // hhv以外のp個のhvの重心
  return resnon;                 // 収束せず
}

restype varclass::check_hv(const vec<double>& hv) const // パラメータの発散チェック
{
  restype reshv = resnon;
  for (size_t i = 0; reshv == resnon && i < hv.dim(); ++i)
  {
    if (fabs(hv(i)) > FDmax || isnan(hv(i))) reshv = reserr;
  }
  if (reshv == reserr) return reshv; // 発散
  if (sgndiff(hv(0), vym) < 0 || sgndiff(hv(0) - vym, vm) > 0) return reserr;
  return resok;                 // 発散せず
}

restype varclass::cal_newton(int& nr, vec<double>& hv, option& opt, ldata& datalst) const
// Newton-Raphson法
{
  int np = hv.dim();
  vec<double> gradv(np), htv(hv);
  mat<double> hessmat(np, np);
  double qnew = fq(htv, opt, datalst), qold = 0.0;
  restype resnew = resnon;
  nr = 0;
  do
  {
    cal_gh(gradv, hessmat, htv, opt, datalst);
    hessmat.inv();
    htv -= hessmat * gradv;
    qold = qnew;
    qnew = fq(htv, opt, datalst);  ++nr;
  } while ((resnew = check_q(qnew, qold, htv, nr, opt)) == resnon);
  if (resnew == reserr) return resnew; // 発散
  hv = htv;
  return resnew;                // 計算できた
}

void varclass::cal_gh(vec<double>& gradv, mat<double>& hessmat, const vec<double>& hv, option& opt, ldata& datalst) const
// 傾斜ベクトルとHessの行列の計算
{
  for (size_t i = 0; i < hv.dim(); ++i)
  {
    gradv.assign(i, fqd(i, hv, opt, datalst));
    for (size_t j = i; j < hv.dim(); ++j)
    { hessmat.assign(i, j, fqd2(i, j, hv, opt, datalst)); }
  }
  for (size_t i = 0; i < hv.dim() - 1; ++i)
  {
    for (size_t j = i + 1; j < hv.dim(); ++j) hessmat.assign(j, i, hessmat(i, j));
  }
}

restype varclass::check_q(double qnew, double qold, const vec<double>& hv, int nr, option& opt) const
// Q(h)の計算と終了基準のチェック
{
  if (check_hv(hv) == reserr) return reserr;      // 発散
  if (qnew > FDmax || isnan(qnew)) return reserr; // 発散
  double qerr = fabs(qold - qnew);
  if (qerr < fabs(qnew * opt.eps())) return resok; // 収束
  if (opt.rmax() > 0 && nr >= opt.rmax()) return resrmax; // 反復回数最大値
  return resnon;                // 収束せず
}

double varclass::fq(const vec<double>& hv, option& opt, ldata& datalst) const
// 誤差関数
{
  double q = 0.0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double x = (opt.du() == dulog) ? log10(dose) : dose;
      double ei = y - func(x, hv);
      q += ei * ei;
    }
  }
  return q;                     // 平方誤差合計を返す
}

double varclass::fqd(int hidx, const vec<double>& hv, option& opt, ldata& datalst) const
// 誤差の偏導関数
{
  double dq = 0.0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double x = (opt.du() == dulog) ? log10(dose) : dose;
      double dei = y * funcd(x, hidx, hv) - func(x, hv) * funcd(x, hidx, hv);
      dq -= 2.0 * dei;
    }
  }
  return dq;                    // 誤差導関数値を返す
}

double varclass::fqd2(int hiidx, int hjidx, const vec<double>& hv, option& opt, ldata& datalst) const
// 誤差の2階偏導関数
{
  double d2q = 0.0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double x = (opt.du() == dulog) ? log10(dose) : dose;
      double qd2 = funcd2(x, hiidx, hjidx, hv);
      double d2ei = y * qd2 - funcd(x, hiidx, hv) * funcd(x, hjidx, hv) - func(x, hv) * qd2;
      d2q -= 2.0 * d2ei;
    }
  }
  return d2q;                   // 誤差2階導関数値を返す
}

double varclass::func(double x, const vec<double>& hv) const
// 用量反応関数
{
  return vym + logdis(x, vm, hv(1), hv(0) - vym);
}

double varclass::funcd(double x, int hidx, const vec<double>& hv) const
// モデル偏導関数
{
  constexpr double Delta = EPSdef; // 微小変動の割合(差分法用)
  vec<double> hdv(hv);
  double dh = 0.0, grad = 0.0;
  dh = hv(hidx) * Delta;        // 差分法
  hdv.assign(hidx, hdv(hidx) + dh);
  grad = (sgn(dh) == 0) ? 0.0 : (func(x, hdv) - func(x, hv)) / dh;
  return grad;                  // 偏導関数値を返す
}

double varclass::funcd2(double x, int hiidx, int hjidx, const vec<double>& hv) const
// モデル2階偏導関数
{
  constexpr double Delta = EPSdef; // 微小変動の割合(差分法用)
  vec hijv(hv), hjv(hv);
  double fhi = 0.0, fhj = 0.0, dhi = 0.0, dhj = 0.0, grad2 = 0.0;
  dhi = hv(hiidx) * Delta;      // 差分法
  dhj = hv(hjidx) * Delta;
  hijv.assign(hiidx, hv(hiidx) + dhi);
  fhi = func(x, hijv);
  hjv.assign(hjidx, hv(hjidx) + dhj);
  fhj = func(x, hjv);
  hijv.assign(hjidx, hijv(hjidx) + dhj);
  grad2 = (sgn(dhi * dhj) == 0) ? 0.0 : (func(x, hijv) - fhi - fhj + func(x, hv)) / dhi / dhj;
  return grad2;                 // 2階偏導関数値を返す
}

void varclass::cal_ssdrc(option& opt, ldata& datalst)
// 用量反応曲線の各種の平方和を計算する
{
  vec<double> htv(vhv);
  htv.assign(0, vym + vm * 0.5);  htv.assign(1, 0.0);
  vstdrc = fq(htv, opt, datalst);  // y0=ym+M/2,c=0の時のqerrをstにする
  vqerr = fq(vhv, opt, datalst);
  vsbdrc = vstdrc - vqerr;
  if (vsbdrc < FDeps) vsbdrc = 0.0;
  double my = func(0.0, htv);   // ロジスティック関数の平均値
  vsddrc = 0.0;
  for (auto& dystat : vdysmap)
  {
    double dose = dystat.first, y = dystat.second.mean();
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double ei = y - my;
      vsddrc += (double)(dystat.second.n()) * ei * ei;
    }
  }
  if (vsddrc < FDeps) vsddrc = 0.0;
  vsldrc = vsddrc - vsbdrc;
  if (vsldrc < FDeps) vsldrc = 0.0;
  vsrdrc = vstdrc - vsddrc;
  if (vsrdrc < FDeps) vsrdrc = 0.0;
  vdftdrc = dystatc.n() - 1;  vdfddrc = vndrl - 1;  vdfbdrc = vhv.dim() - 1;
  vdfrdrc = vdftdrc - vdfddrc;  vdfldrc = vdfddrc - vdfbdrc;
  if (vdftdrc == vdfddrc)                  // 用量データが1例のみの時
  { vdfrdrc = vdfldrc;  vsrdrc = vsldrc; }
}

class doserclass                // doserメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_mstat(ostream& fout, option& opt, const varclass& var) const;
  // 用量別基礎統計量を出力する
  void put_maxcomp(ostream& fout, option& opt, const varclass& var) const;
  // 最大対比法結果を出力する
  void put_drlreg(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応直線の解析結果を出力する
  void put_drlestd(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応直線で用量を逆推定する
  void put_drltab(ostream& fout, const varclass& var) const;
  // 用量反応直線の分散分析表を出力する
  void put_drlyreg(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応ロジット直線の解析結果を出力する
  void put_drlyestd(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応ロジット直線の用量を逆推定する
  void put_drlytab(ostream& fout, const varclass& var) const;
  // 用量反応ロジット直線の分散分析表を出力する
  void put_drcreg(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応曲線の解析結果を出力する
  void put_drcestd(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応曲線の用量を逆推定する
  void put_drctab(ostream& fout, const varclass& var) const;
  // 用量反応曲線の分散分析表を出力する
  void put_dns(option& opt, idd& id, ldata& datalst, const varclass& var) const;
  // xとyとy推測値とlyとly推測値をデータファイル形式で出力する
  void put_dnsid(ofstream& fout, idd& id) const;
  // DANSデータのタイトルとデータ識別情報を出力する
   void plot_ngp(option& opt, idd& id, const varclass& var) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpmyse(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に反応の平均±SEを出力する
  void plot_ngpdrl(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に用量反応直線を出力する
  double set_xrange(double& xmin, double& xmax, option& opt, const varclass& var) const;
  // X軸の最小値と最大値を設定して範囲を返す
  void plot_ngpmlyse(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に反応ロジットの用量別平均値±SEを出力する
  void plot_ngpdrly(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に用量反応ロジット直線を出力する
  void plot_ngpdrc(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に用量反応曲線を出力する
 public:
  doserclass() : fname()        // コンストラクタ
  {}
  ~doserclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void doserclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "用量項目:" << id.item(0) << "\n";
  fout << "反応項目:" << id.item(1) << endl;
}

void doserclass::put_mstat(ostream& fout, option& opt, const varclass& var) const
// 用量別基礎統計量を出力する
{
  constexpr int LENline = 70;
  if (opt.res("staty"))
  {
    fout << "\n    用量     ";
    if (opt.du() == dulog) fout << "x=log(用量)";
    else fout << "     x=用量";
    fout << "      例数      平均値    標準偏差    標準誤差\n";
    putline(fout, LENline);
    for (auto& dystat : var.dysmap())
    {
      fout << setw(8) << dystat.first;
      if (opt.du() == dulog)
      {
        if (sgn(dystat.first) > 0) fout << setw(16) << log10(dystat.first);
        else fout << "  (計算から除外)";
      }
      else
      { fout << setw(16) << dystat.first; }
      fout << setw(10) << dystat.second.n() << setw(12) << dystat.second.mean();
      fout << setw(12) << dystat.second.sd() << setw(12) << dystat.second.se();
      fout << "\n";
    }
    putline(fout, LENline);
    string tstr("全体");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(8 + npad) << tstr << setw(16) << var.mx();
    fout << setw(10) << var.n() << setw(12) << var.my();
    fout << setw(12) << var.sdy() << setw(12) << var.sey() << "\n";
  }
  if (opt.res("statly"))
  {
    fout << "\n反応ロジット\n";
    fout << "    用量     ";
    if (opt.du() == dulog) fout << "x=log(用量)";
    else fout << "     x=用量";
    fout << "      例数      平均値    標準偏差    標準誤差\n";
    putline(fout, LENline);
    for (auto& dlystat : var.dlysmap())
    {
      fout << setw(8) << dlystat.first;
      if (opt.du() == dulog) fout << setw(16) << log10(dlystat.first);
      else fout << setw(16) << dlystat.first;
      fout << setw(10) << dlystat.second.n() << setw(12) << dlystat.second.mean();
      fout << setw(12) << dlystat.second.sd() << setw(12) << dlystat.second.se();
      fout << "\n";
    }
    putline(fout, LENline);
    string tstr("全体");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(8 + npad) << tstr << setw(16) << var.lymx();
    fout << setw(10) << var.lyn() << setw(12) << var.lymy();
    fout << setw(12) << var.lysdy() << setw(12) << var.lysey() << "\n";
  }
}

void doserclass::put_maxcomp(ostream& fout, option& opt, const varclass& var) const
// 最大対比法結果を出力する
{
  fout << "\n・最大対比法(maximum contrast method)\n";
  if ((int)(opt.csize()) != var.ndrl())
  {
    fout << "※最大対比法用係数の値が不適切！" << endl;
    return;
  }
  double sp = 0.0, div = 0.0;
  int nd = 0;
  fout << "  係数:";
  for (auto& dystat : var.dysmap())
  {
    double dose = dystat.first;
    if (opt.du() != dulog || sgn(dose) > 0)
    {
      fout << "用量" << dose << "=" << opt.c(nd) << " ";
      sp += dystat.second.n() * opt.c(nd) * dystat.second.mean();
      div += dystat.second.n() * opt.c(nd) * opt.c(nd);
      ++nd;
    }
  }
  fout << " 寄与率:r2乗=" << sp / div * sp / var.sd() << "\n";
  double f = (var.vr() > 0.0) ? sp / div * sp / var.vr() : 0.0;
  ldbl pr = ftop(f, 1, var.dfr());
  fout << "  Fisher型 :F=" << f << " 第1自由度=1 第2自由度=" << var.dfr() << " 有意確率p=" << pr << pind(pr) <<  "\n";
  f /= (double)(var.dfd());
  pr = ftop(f, var.dfd(), var.dfr());
  fout << "  Scheffe型:F=" << f << " 第1自由度=" << var.dfd() << " 第2自由度=" << var.dfr() << " 有意確率p=" << pr << pind(pr) << endl;
}

void doserclass::put_drlreg(ostream& fout, option& opt, const varclass& var) const
// 用量反応直線の解析結果を出力する
{
  double r2 = var.sb() / var.sd();
  cout << "\n・";
  if (opt.du() == dulog) fout << "対数";
  fout << "用量反応直線:y=" << var.a() << showpos << var.b() << noshowpos << "x\n";
  if (var.resdrl() == reserr)
  { fout << "  ※用量数不足で計算不可能！"; }
  else
  { fout << "  寄与率:r2乗=" << r2 << "    回帰係数の標準誤差=" << var.seb(); }
  fout << endl;
}

void doserclass::put_drlestd(ostream& fout, option& opt, const varclass& var) const
// 用量反応直線で用量を逆推定する
{
  double y = (opt.y() == BLANKf || opt.y() == -DBL_MAX) ? var.my() : opt.y();
  double cf = (sgn(var.b()) == 0) ? 0.0 : (y - var.my()) / var.b();
  double x = var.mx() + cf;
  fout << "\n・反応(y)=" << y << "の時の用量:";
  if (opt.du() == dureal) fout << "x=" << x << "\n";
  else fout << "D=" << pow(10.0, x) << "(x=log(D)=" << x << ")\n";
  double xl = x, xu = x;
  if (sgn(var.b()) != 0 && sgn(var.sxx()) > 0)
  {
    double t = ptot(1.0 - opt.cc(), var.dfr());
    double g = t / var.b() * t / var.b() * var.vr() / var.sxx();
    if (opt.g() == false || g >= 1.0) g = 0.0; // g>1だと計算できないため
    double gr = 1.0 - g;
    double v = var.vr() * (gr * (1.0 + 1.0 / (double)(var.n())) + cf * cf / var.sxx());
    double ci = t / var.b() * sqrt(v);
    if (ci < 0.0) ci = -ci;
    xl = (x - g * var.mx() - ci) / gr;  xu = (x - g * var.mx() + ci) / gr;
  }
  fout << "  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal) fout << "下限=" << xl << "  上限=" << xu << "\n";
  else fout << "下限=" << pow(10.0, xl) << "(x=" << xl << ")  上限=" << pow(10.0, xu) << "(x=" << xu << ")" << endl;
}

void doserclass::put_drltab(ostream& fout, const varclass& var) const
// 用量反応直線の分散分析表を出力する
{
  constexpr int LENline = 74;
  string tstr("用量反応直線の分散分析表(ANOVA table)");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr << "\n";
  fout << "要因          平方和    自由度    平均平方和         F値   有意確率p値\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  calvfp(v, f, pr, var.sb(), var.dfb(), var.sr(), var.dfr());
  fout << "回帰" << setw(16) << var.sb() << setw(10) << var.dfb();
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sl(), var.dfl(), var.sr(), var.dfr());
  fout << "ズレ" << setw(16) << var.sl() << setw(10) << var.dfl() << setw(14) << v;
  if (var.dft() > var.dfd()) fout << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  fout << "用量" << setw(16) << var.sd() << setw(10) << var.dfd();
  if (var.dft() > var.dfd())
  {
    fout << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    v = (var.dfr() > 0) ? var.sr() / (double)(var.dfr()) : 0.0;
    fout << "残差" << setw(16) << var.sr() << setw(10) << var.dfr();
    fout << setw(14) << v << "\n";
    putline(fout, LENline);
    fout << "全体" << setw(16) << var.st() << setw(10) << var.dft();
  }
  fout << endl;
}

void doserclass::put_drlyreg(ostream& fout, option& opt, const varclass& var) const
// 用量反応ロジット直線の解析結果を出力する
{
  double y = var.y00() - var.ym();
  double ly00 = (sgn(y) <= 0 || sgndiff(var.m(), y) <= 0) ? 0.0 : log(y / (var.m() - y));
  y = var.y0() - var.ym();
  double ly0 =  (sgn(y) <= 0 || sgndiff(var.m(), y) <= 0) ? 0.0 : log(y / (var.m() - y));
  double d50 = (sgn(var.c()) == 0) ? 0.0 : -ly0 / var.c();
  double r2 = (sgn(var.sbly()) > 0) ? var.sbly() / var.sdly() : 0.0;
  fout << "\n・";
  if (opt.du() == dulog) fout << "対数";
  fout << "用量反応ロジット直線:ly=ly0+c・x=" << ly0 << showpos << var.c() << noshowpos << "x\n";
  fout << "  logit(y)=ly=ly0+c・x=ln[(y0-ym)/{M-(y0-ym)}]+c・x=ln[(y-ym)/{M-(y-ym)}]\n";
  fout << "  反応:y=ym+M/{1+exp(-ly)}  範囲外データ近似数=" << var.noutdly() << "\n";
  fout << "  最小値:ym=" << var.ym() << "  最大値:ym+M=" << var.ym() << "+" << var.m() << "=" << var.ym() + var.m() << "\n";
  fout << "  パラメータ初期値:y0=" << var.y00() << "  ly0=" << ly00 << "  c=" << var.c0() << "\n";
  fout << "  パラメータ推定値:y0=" << var.y0() << "  ly0=" << ly0 << "  c=" << var.c() << "\n";
  if (var.resdrly() == reserr)
  { fout << "  ※エラー発生！"; }
  else
  {
    fout << "  y=ym+M/2=" << var.ym() + var.m() * 0.5 << "の時の用量:D50=";
    if (opt.du() == dureal) fout << "x=-ly0/c=" << d50 << "\n";
    else fout << pow(10.0, d50) << "(x=log(D50)=-ly0/c=" << d50 << ")\n";
    fout << "  重寄与率:R2乗=" << r2 << "  重相関係数:R=" << sqrt(r2);
  }
  fout << endl;
}

void doserclass::put_drlyestd(ostream& fout, option& opt, const varclass& var) const
// 用量反応ロジット直線で用量を逆推定する
{
  double ye = opt.y();
  if (ye == BLANKf) ye = var.ym() + var.m() * 0.5;
  else if (ye == -DBL_MAX) ye = var.my();
  double y = ye - var.ym();
  if (sgn(y) <= 0 || sgndiff(var.m(), y) <= 0)
  {
    fout << "\n・反応:y=" << ye << "の時の用量:逆推定不可能！\n";
    return;
  }
  double ly = log(y / (var.m() - y));
  y = var.y0() - var.ym();
  double ly0 = (sgn(y) <= 0 || sgndiff(var.m(), y) <= 0) ? 0.0 : log(y / (var.m() - y));
  double x = (sgn(var.c()) == 0) ? 0.0 : (ly - ly0) / var.c();
  fout << "\n・反応:y=" << ye << "(ly=" << "" << ly << ")の時の用量:";
  if (opt.du() == dureal) fout << "D=x=" << x << "\n";
  else fout << "D=" << pow(10.0, x) << "(x=log(D)=" << x << ")\n";
  double xlow = x, xup = x;
  if (sgn(var.c()) != 0 && sgn(var.lysxx()) > 0)
  {
    double cf = (ly - var.lymy()) / var.c();
    double t = ptot(1.0 - opt.cc(), var.dfrly());
    double g = t / var.c() * t / var.c() * var.vrly() / var.lysxx();
    if (opt.g() == false || g >= 1.0) g = 0.0; // g>1だと計算できないため
    double gr = 1.0 - g;
    double v = var.vrly() * (gr * (1.0 + 1.0 / (double)(var.lyn())) + cf * cf / var.lysxx());
    double ci = t / var.c() * sqrt(v);
    if (ci < 0.0) ci = -ci;
    xlow = (x - g * var.lymx() - ci) / gr;
    xup = (x - g * var.lymx() + ci) / gr;
  }
  fout << "  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal) fout << "下限=" << xlow << "  上限=" << xup;
  else fout << "下限=" <<  pow(10.0, xlow) << "(x=" << xlow << ")  上限=" << pow(10.0, xup) << "(x=" << xup << ")";
  fout << endl;
}

void doserclass::put_drlytab(ostream& fout, const varclass& var) const
// 用量反応ロジット直線の分散分析表を出力する
{
  constexpr int LENline = 74;
  string tstr("用量反応ロジット直線の分散分析表(ANOVA table)");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr << "\n";
  fout << "要因          平方和    自由度    平均平方和         F値   有意確率p値\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  calvfp(v, f, pr, var.sbly(), var.dfbly(), var.srly(), var.dfrly());
  fout << "回帰" << setw(16) << var.sbly() << setw(10) << var.dfbly();
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.slly(), var.dflly(), var.srly(), var.dfrly());
  fout << "ズレ" << setw(16) << var.slly() << setw(10) << var.dflly() << setw(14) << v;
  if (var.dftly() > var.dfdly()) fout << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sdly(), var.dfdly(), var.srly(), var.dfrly());
  fout << "用量" << setw(16) << var.sdly() << setw(10) << var.dfdly();
  if (var.dftly() > var.dfdly())
  {
    fout << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    v = (var.dfrly() > 0) ? var.srly() / (double)(var.dfrly()) : 0.0;
    fout << "残差" << setw(16) << var.srly() << setw(10) << var.dfrly();
    fout << setw(14) << v << "\n";
    putline(fout, LENline);
    fout << "全体" << setw(16) << var.stly() << setw(10) << var.dftly();
  }
  fout << endl;
}

void doserclass::put_drcreg(ostream& fout, option& opt, const varclass& var) const
// 用量反応曲線の解析結果を出力する
{
  double y = var.y00drc() - var.ymdrc();
  double ly00 = (sgn(y) > 0 && sgndiff(var.mdrc(), y) > 0) ? log(y / (var.mdrc() - y)) : 0.0;
  y = var.y0drc() - var.ymdrc();
  double ly0 = (sgn(y) > 0 && sgndiff(var.mdrc(), y) > 0) ? log(y / (var.mdrc() - y)) : 0.0;
  double d50 = (sgn(var.cdrc()) == 0) ? 0.0: -ly0 / var.cdrc();
  fout << "\n・";
  int w = 16;
  if (opt.du() == dulog) { fout << "対数";  w += 4; }
  fout << "用量反応曲線:y=ym+M/[1+[{M-(y0-ym)}/(y0-ym)]exp(-c・x)]\n";
  fout << setw(w) << " " << "=ym+M/{1+exp(-ly0-c・x)}=ym+M/{1+exp(-ly)}\n";
  fout << "  logit(y)=ly=ly0+c・x=ln[(y0-ym)/{M-(y0-ym)}]+c・x=ln[(y-ym)/{M-(y-ym)}]\n";
  fout << "  計算方法:" << opt.methname() << "  反復回数=" << var.nrdrc();
  if (var.resdrc() == resrmax) fout << "(最大反復回数)";
  fout << "\n  最小値(固定):ym=" << var.ymdrc() << "  最大値(固定):ym+M=" << var.ymdrc() << "+" << var.mdrc() << "=" << var.ymdrc() + var.mdrc() << "\n";
  w = 21;
  fout << "  パラメーター初期値:y0=" << var.y00drc() << "  ly0=" << ly00 << "  c=" << var.c0drc() << "\n";
  fout << "  パラメーター収束値:y0=" << var.y0drc() << "  ly0=" << ly0 << "  c=" << var.cdrc() << "\n";
  if (var.resdrc() == reserr)
  { fout << "  ※エラー発生！"; }
  else
  {
    fout << "  y=ym+M/2=" << var.ymdrc() + var.mdrc() * 0.5 << "の時の用量:D50=";
    if (opt.du() == dureal) fout << "x=-ly0/c=" << d50;
    else fout << pow(10.0, d50) << "(x=log(D50)=-ly0/c=" << d50 << ")";
    double aic = (double)(var.n()) * log(var.qerr()) + 2.0 * (double)(var.npdrc());
    fout << "\n  残差平方和(Q)=" << var.qerr() << "  赤池の情報量基準(AIC)=" << aic << "\n";
    double r2 = (var.sddrc() > 0.0) ? var.sbdrc() / var.sddrc() : 0.0;
    fout << "  重寄与率:R2乗=" << r2 << "  重相関係数:R=" << sqrt(r2);
  }
  fout << endl;
}

void doserclass::put_drcestd(ostream& fout, option& opt, const varclass& var) const
// 用量反応曲線で用量を逆推定する
{
  double ye = opt.y();
  if (ye == BLANKf) ye = var.ymdrc() + var.mdrc() * 0.5; // ym+M/2
  else if (ye == -DBL_MAX) ye = var.my();                // y総平均値
  double y = ye - var.ymdrc();
  if (sgn(y) <= 0 || sgndiff(var.mdrc(), y) <= 0)
  {
    fout << "\n・反応:y=" << ye << "の時の用量:逆推定不可能！\n";
    return;
  }
  double ly = (opt.y() == BLANKf) ? 0.0 : log(y / (var.mdrc() - y));
  y = var.y0drc() - var.ymdrc();
  double ly0 = (sgn(y) > 0) ? log(y / (var.mdrc() - y)) : 0.0;
  double x = (sgn(var.cdrc()) == 0 || sgndiff(ly, ly0) == 0) ? 0.0 : (ly - ly0) / var.cdrc();
  fout << "\n・反応:y=" << ye << "の時の用量:";
  if (opt.du() == dureal) fout << "D=x=" << x << "\n";
  else fout << "D=" << pow(10.0, x) << "(x=log(D)=" << x << ")\n";
  double xlow = x, xup = x;
  if (sgn(var.lysxx()) > 0)     // ロジット分析の結果を利用して近似解を求める
  {
    double xdef = (ly - var.lymy()) / var.cdrc();
    double v = var.vrly() * (1.0 + 1.0 / (double)(var.lyn()) + xdef * xdef / var.lysxx());
    double t = ptot(1.0 - opt.cc(), var.dfrly());
    double ci = (sgn(var.cdrc()) == 0) ? 0.0 : t / var.cdrc() * sqrt(v);
    if (ci < 0.0) ci = -ci;
    xlow = x - ci;  xup = x + ci;
  }
  fout << "  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal) fout << "下限=" << xlow << "  上限=" << xup << "\n";
  else fout << "下限=" << pow(10.0, xlow) << "(x=" << xlow << ")  上限=" << pow(10.0, xup) << "(x=" << xup << ")" << endl;
}

void doserclass::put_drctab(ostream& fout, const varclass& var) const
// 用量反応曲線の分散分析表出力
{
  constexpr int LENline = 74;
  string tstr("用量反応曲線の分散分析表(ANOVA table)");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr << "\n";
  fout << "要因          平方和    自由度    平均平方和         F値   有意確率p値\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  calvfp(v, f, pr, var.sbdrc(), var.dfbdrc(), var.srdrc(), var.dfrdrc());
  fout << "回帰" << setw(16) << var.sbdrc() << setw(10) << var.dfbdrc();
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sldrc(), var.dfldrc(), var.srdrc(), var.dfrdrc());
  fout << "ズレ" << setw(16) << var.sldrc() << setw(10) << var.dfldrc() << setw(14) << v;
  if (var.dftdrc() > var.dfddrc()) fout << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sddrc(), var.dfddrc(), var.srdrc(), var.dfrdrc());
  fout << "用量" << setw(16) << var.sddrc() << setw(10) << var.dfddrc();
  if (var.dftdrc() > var.dfddrc())
  {
    fout << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    v = (var.dfrdrc() > 0) ? var.srdrc() / (double)(var.dfrdrc()) : 0.0;
    fout << "残差" << setw(16) << var.srdrc() << setw(10) << var.dfrdrc();
    fout << setw(14) << v << "\n";
    putline(fout, LENline);
    fout << "全体" << setw(16) << var.stdrc() << setw(10) << var.dftdrc();
  }
  fout << endl;
}

void doserclass::put_dns(option& opt, idd& id, ldata& datalst, const varclass& var) const
// xとyとy推測値とlyとly推測値をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout, id);
  datac tdata(7);
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), yo = cdatai.fdata(1);
    tdata.assign(0, dose);  tdata.assign(1, yo);
    tdata.assign(2, BLANKs);  tdata.assign(3, BLANKs);
    tdata.assign(4, BLANKs);  tdata.assign(5, BLANKs);
    tdata.assign(6, BLANKs);
    if (dose == BLANKf || yo == BLANKf || (opt.du() == dulog && sgn(dose) <= 0))
    {
      if (dose != BLANKf && sgn(dose) > 0) tdata.assign(2, log(dose));
    }
    else
    {
      if (sgn(dose) > 0) tdata.assign(2, log(dose));
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      if (opt.res("drl") && var.resdrl() == resok)
      { tdata.assign(3, var.a() + var.b() * mdose); }
      if (opt.res("drly") && var.resdrly() == resok)
      {
        double y = yo - var.ym();
        if (sgn(y) > 0 && sgndiff(var.m(), y) > 0) tdata.assign(4, log(y / (var.m() - y)));
        y = var.y0() - var.ym();
        double ly0 = log(y / (var.m() - y));
        tdata.assign(5, ly0 + var.c() * mdose);
      }
      if (opt.res("drc") && (var.resdrc() == resok || var.resdrc() == resrmax))
      {
        double y = var.y0drc() - var.ymdrc();
        double ly0 = log(y / (var.mdrc() - y));
        double ly = ly0 + var.cdrc() * mdose;
        tdata.assign(6, var.ymdrc() + (var.mdrc() / (1.0 + exp(-ly))));
      }
    }
    tdata.put(fout);
  }
}

void doserclass::put_dnsid(ofstream& fout, idd& id) const
// DANSデータのタイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), 7);
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, id.item(1));
  ostringstream ostr;
  ostr << "log(" << id.item(0) << ")";
  tid.assign_item(2, ostr.str());
  ostr.str("");
  ostr << id.item(1) << "の用量反応直線による推測値";
  tid.assign_item(3, ostr.str());
  ostr.str("");
  ostr << id.item(1) << "のロジット";
  tid.assign_item(4, ostr.str());
  ostr.str("");
  ostr << id.item(1) << "ロジットの用量反応ロジット直線による推測値";
  tid.assign_item(5, ostr.str());
  ostr.str("");
  ostr << id.item(1) << "の用量反応曲線による推測値";
  tid.assign_item(6, ostr.str());
  tid.put_item(fout, formtag);
}

void doserclass::plot_ngp(option& opt, idd& id, const varclass& var) const
// Ngraph用にデータを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("myse"))          // 反応の用量別平均値±SE出力
  {
    fout << "-s1 -dmark,00 -o300";
    if (opt.du() == dulog) fout << " -exlog";
    fout << "\n";
    plot_ngpid(fout, id);
    plot_ngpmyse(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("drl"))           // 用量反応直線出力
  {
    if (delml)                  // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    plot_ngpid(fout, id);
    if (var.resdrl() == resok) plot_ngpdrl(fout, opt, var);
    else fout << "%計算不可能\n";
  }
  if (opt.ngp("mlyse"))         // 反応ロジットの用量別平均値±SE出力
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -o300";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true; 
    }
    plot_ngpid(fout, id);
    plot_ngpmlyse(fout, opt, var);
  }
  if (opt.ngp("drly"))          // 用量反応ロジット直線出力
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    plot_ngpid(fout, id);
    if (var.resdrl() == resok) plot_ngpdrly(fout, opt, var);
    else fout << "%計算不可能\n";
  }
  if (opt.ngp("drc"))           // 用量反応曲線出力
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    if (var.resdrc() == reserr) fout << "%計算不可能\n";
    else plot_ngpdrc(fout, opt, var);
  }
}

void doserclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%用量項目:" << id.item(0) << "\n";
  fout << "%反応項目:" << id.item(1) << endl;
}

void doserclass::plot_ngpmyse(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に反応の平均±SEを出力する
{
  fout << "%平均値±SE:SEを描く時は-derrorbar_y -w40 -o500\n";
  for (auto& dystat : var.dysmap())
  {
    if (opt.du() != dulog || sgn(dystat.first) > 0)
    {
      double se = dystat.second.se();
      fout << dystat.first << " " << dystat.second.mean() << " " << -se << " " << se << endl;
    }
  }
}

void doserclass::plot_ngpdrl(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に用量反応直線を出力する
{
  fout << "%用量反応直線\n";
  double xmin = 0.0, xmax = 0.0;
  double xrange = set_xrange(xmin, xmax, opt, var);
  double y = var.a() + var.b() * xmin;
  double dose = (opt.du() == dulog) ? pow(10.0, xmin) : xmin;
  fout << dose << " " << y << "\n";
  y = var.a() + var.b() * xmax;
  dose = (opt.du() == dulog) ? pow(10.0, xmax) : xmax;
  fout << dose << " " << y << endl;
}

double doserclass::set_xrange(double& xmin, double& xmax, option& opt, const varclass& var) const
// X軸の最小値と最大値を設定して範囲を返す
{
  xmin = (opt.du() == dureal) ? var.dmin() : log10(var.dmin());
  xmax = (opt.du() == dureal) ? var.dmax() : log10(var.dmax());
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf) // X軸の範囲指定あり
  { xmin = opt.xmin();  xmax = opt.xmax(); }
  else
  {
    double xrange = xmax - xmin, xmg = 0.1;
    xmin -= xrange * xmg;  xmax += xrange * xmg;
  }
  return xmax - xmin;           // 範囲を返す
}

void doserclass::plot_ngpmlyse(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に反応ロジットの用量別平均値±SEを出力する
{
  fout << "%反応ロジットの平均値±SE:SEを描く時は-derrorbar_y -w40 -o500\n";
  for (auto& dlystat : var.dlysmap())
  {
    double se = dlystat.second.se();
    fout << dlystat.first << " " << dlystat.second.mean() << " " << -se << " " << se << endl;
  }
}

void doserclass::plot_ngpdrly(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に用量反応ロジット直線を出力する
{
  fout << "%用量反応ロジット直線\n";
  double xmin = 0.0, xmax = 0.0;
  double xrange = set_xrange(xmin, xmax, opt, var);
  double y = var.y0() - var.ym();
  double ly0 = log(y / (var.m() - y));
  double ly = ly0 + var.c() * xmin;
  double dose = (opt.du() == dulog) ? pow(10.0, xmin) : xmin;
  fout << dose << " " << ly << "\n";
  ly = ly0 + var.c() * xmax;
  dose = (opt.du() == dulog) ? pow(10.0, xmax) : xmax;
  fout << dose << " " << ly << endl;
}

void doserclass::plot_ngpdrc(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に用量反応曲線を出力する
{
  fout << "%用量反応曲線(ロジスティック曲線)\n";
  constexpr int RDdef = 50;     // グラフの描画範囲分割数省略時値
  double xmin = 0.0, xmax = 0.0;
  double xrange = set_xrange(xmin, xmax, opt, var);
  double dx = xrange / (double)RDdef;
  double x = xmin;
  double y = var.y0drc() - var.ymdrc();
  double ly0 = log(y / (var.mdrc() - y));
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    double dose = (opt.du() == dureal) ? x : pow(10.0, x);
    fout << dose << " " << var.ymdrc() + var.mdrc() / (1.0 + exp(-ly0 - var.cdrc() * x)) << endl;
  }
}

int doserclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void doserclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  varclass var;                 // 変数オブジェクト
  var.cal(opt, datalst);
  if (opt.res("res"))           // 解析結果出力
  {
    put_resid(cout, id);
    if (opt.res("staty") || opt.res("statly")) put_mstat(cout, opt, var);
    if (var.resvar() == reserr)
    {
      cout << "※用量反応解析不可能！\n\n" << endl;
      return;
    }
    if (opt.csize() > 0 && var.ndrl() > 2) put_maxcomp(cout, opt, var);
    if (opt.res("drl"))       // 用量反応直線分析結果出力
    {
      put_drlreg(cout, opt, var);
      if (var.resdrl() != reserr)
      {
        if (opt.res("y")) put_drlestd(cout, opt, var);
        put_drltab(cout, var);
      }
    }
    if (opt.res("drly"))      // 用量反応ロジット分析結果出力
    {
      put_drlyreg(cout, opt, var);
      if (var.resdrly() != reserr)
      {
        if (opt.res("y")) put_drlyestd(cout, opt, var);
        put_drlytab(cout, var);
      }
    }
    if (opt.res("drc"))       // 用量反応曲線分析結果出力
    {
      put_drcreg(cout, opt, var);
      if (var.resdrc() != reserr)
      {
        if (opt.res("y")) put_drcestd(cout, opt, var);
        put_drctab(cout, var);
      }
    }
    cout << "\n" << endl;
  }
  if (opt.dnsname() != "") put_dns(opt, id, datalst, var);
  if (opt.ngpname() != "") plot_ngp(opt, id, var);
}

int main(int argc, char *argv[]) // メインルーチン
{
  doserclass doser;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) doser.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { doser.get(argv[i], opt, id, datalst);  fok = false; }
    }
    doser.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
