// doserl.C Version 7.43  Copyleft (c) 杉本典夫
// Time-stamp: <2025-03-05 08:52:00 tac>
//
// 最初の項目を用量項目、次の項目を反応項目として
// ロジスティック関数を用いた用量反応解析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W doserl doserl.C
//               -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "doserl  Version 7.43         Copyleft(Vo\\oV) 2025 杉本典夫",
  "書式:doserl [<オプション> …]  [<データファイル> …]",
  "機能:最初の項目を用量、次の項目を反応としてロジスティック関数を用いた",
  "     用量反応解析をするのだ。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するのだ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -du={l|r}:用量の単位を指定するのだ。",
  "    l:用量を対数変換して計算(オプション省略時)  r:実数のまま計算",
  "  -eps[=<収束誤差>]:非線形最小二乗法の収束誤差を指定するのだ。(0<収束誤差)",
  "  -f=<ファイル>:オプションを指定したファイルから読み込むのだ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するのだ。(-h:-h=man)",
  "  -h0[=<初期値1>,<初期値2>,…]:パラメーターの初期値を指定するのだ。",
  "  -meth={newton|non|simplex|etc}:計算法を指定するのだ。",
  "  -ngp={drc|mse}[,…]:Ngraph用出力データを指定するのだ。",
  "    drc:用量反応曲線  mse:平均±SE  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データの出力ファイル名を指定するんよ。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=doserl_ngp.dat",
  "  -res=[-][{def|all|reg|stat|y[{Mh|my|<反応値>}]}][,…]",
  "      :解析結果の出力項目を指定するのだ。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    reg:回帰分析  stat:基礎統計量",
  "    y[{Mh|my|<反応値>}]:指定された反応値の用量逆推定値(y=yMh)",
  "      Mh:ym+M/2(D50)  my:反応の総平均値  <反応値>:反応値指定",
  "  -rmax=<最大計算回数>:最大計算回数を指定するのだ。(0≦最大計算回数)",
  "  -ym=<a|d|最小値>:反応の最小値を指定するのだ。(-ym:-ym=a)",
  "    a:自動設定(オプション省略時)  d:微分回帰分析によって設定",
  "  -xrange=<最小値>-<最小値>:X軸の最小値と最大値を指定するのだ。"
};

const string Title = "=== ロジスティック関数による用量反応解析 ==="; // 表題
const string NGPNAMEdef = "doserl_ngp.dat"; // 省略時Ngraphファイル名
constexpr double EPSdef = 1.0e-4; // 収束誤差省略時値
constexpr int RMAXdef = 200;      // 最大計算回数省略時値
constexpr int NPdef = 3;          // 非線形最小2乗法のパラメータ数

using mapds = map<double, statv<double>>; // 用量別統計量マップの別名定義
using mapds_itr = mapds::iterator;        // 反復子の別名定義
using mapds_citr = mapds::const_iterator; // const反復子の別名定義

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

enum methtype                   // 非線形最小2乗法の種類
{
  methnon,                      // 指定なし
  methnewton,                   // Newton-Raphson法
  methsimplex,                  // Simplex法
  methetc                       // その他
};

double absmin(const vec<double>& hv) // 絶対値最小のパラメターの値を返す
{
  double hmin = FDmax;
  for (size_t i = 0;  i < hv.dim();  ++i)
  {
    double hi = fabs(hv(i));
    if (FDmin < hi && hi < hmin) hmin = hi;
  }
  return (hmin == FDmax) ? 1.0 : hmin;
}

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  dutype vdu;                   // 用量の単位
  double veps;                  // 非線形最小2乗法の収束誤差
  lstr h0lst;                   // 初期値リスト
  vecdbl vh0vec;                // 初期値ベクトル
  methtype vmeth;               // 非線形最小2乗法の種類
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  int vrmax;                    // 最大計算回数
  double vy;                    // 用量を推定する反応値
  double vym;                   // yの定数項
  pairdbl vxrange;              // X軸の最小値と最大値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  double eps() const            // 非線形最小2乗法の収束誤差を返す
  { return veps; }
  size_t nh0() const            // 指定された初期値数を返す
  { return vh0vec.size(); }
  double h0(size_t idx) const   // 初期値を返す
  { return (idx < vh0vec.size()) ? vh0vec[idx] : 0.0; }
  methtype meth() const         // 計算法の種類を返す
  { return vmeth; }
  const string methname() const; // 計算法の名称を返す
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大計算回数を返す
  { return vrmax; }
  double xmin() const           // X軸の最小値を返す
  { return vxrange.first; }
  double xmax() const           // X軸の最大値を返す
  { return vxrange.second; }
  double y() const              // 用量を推定する反応値を返す
  { return vy; }
  double ym() const             // ロジスティック関数の最小値を返す
  { return vym; }
  double ym(double ym0)         // ロジスティック関数の最小値を設定する
  { return (vym = ym0); }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdu(dulog), veps(EPSdef), h0lst(), vh0vec(), vmeth(methetc), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vrmax(RMAXdef), vxrange(BLANKf, BLANKf), vy(BLANKf), vym(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_citr ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "du")      // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 非線形最小2乗法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < FDeps) err = badopt(vpname, optstr, USAGE, FDeps, FDmax);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "h0")      // 初期値の指定
  {
    if (strsplit(h0lst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "meth")    // 計算法の指定
  {
    if (opt[1].empty() || opt[1] == "etc") vmeth = methetc;
    else if (opt[1] == "newton") vmeth = methnewton;
    else if (opt[1] == "non") vmeth = methnon;
    else if (opt[1] == "simplex") vmeth = methsimplex;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大計算回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "xrange")  // X軸の最小値と最大値の指定
  {
    if (getrange(vxrange.first, vxrange.second, opt[1]) == 0) // 範囲指定無
    { err = badopt(vpname, optstr, USAGE, -FDmax, FDmax); }
    else if (vxrange.first >= vxrange.second) // 範囲指定不適切
    { err = badopt(vpname, optstr, USAGE, -FDmax, FDmax); }
  }
  else if (opt[0] == "ym")      // 定数項指定
  {
    double fdata = 0.0;
    if (opt[1].empty() || opt[1] == "a") vym = BLANKf; // 自動設定(省略時と同じ)
    else if (opt[1] == "d") vym = -DBL_MAX; // 微分回帰分析によって設定
    else vym = fdata = strtodf(opt[1]);     // 定数指定
    if (fdata == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni != 2) return disperr(vpname, "項目は2個なのだ！");
  if (set_ngp() < 0) return -2; // -resオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (h0lst.size() < 1) return 0; // 初期値指定なし
  if ((int)(h0lst.size()) > NPdef) return disperr(vpname, "初期値の値が多すぎるのだ！");
  vh0vec.resize(h0lst.size());
  vh0vec.assign(h0lst.size(), 0.0); // 初期値を0にする
  size_t i = 0;
  for (const string& hstr : h0lst)
  {
    vh0vec[i] = strtodf(hstr);
    if (vh0vec[i] == BLANKf) return disperr(vpname, "初期値指定=" + hstr + " が不適切なのだ！");
    ++i;
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["drc"] = vngpmap["mse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["drc"] = vngpmap["mse"] = false;
    for (const string& ngpstr : vngplst)
    {
      mapsb_citr ngpitr = vngpmap.find(ngpstr);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngpstr + "オプション指定が不適切！"); }
      else
      { vngpmap[ngpstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["reg"] = vresmap["stat"] = vresmap["y"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["reg"] = vresmap["stat"] = vresmap["y"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr resitr = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(resstr.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def" || resstr == "all")
      {
        vresmap["def"] = vresmap["reg"] = vresmap["stat"] = vresmap["y"] = true;
      }
      else if (resstr == "yMh")
      { vresmap["y"] = true;  vy = BLANKf; }
      else if (resstr == "ymy")
      { vresmap["y"] = true;  vy = -DBL_MAX; }
      else if (resstr[0] == 'y' && resstr.size() > 1)
      {
        vy = strtodf(resstr.substr(1)); // 反応値
        if (vy == BLANKf) return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！");
        vresmap["y"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

const string option::methname() const // 計算法の名称を返す
{
  switch (vmeth)
  {
  case methnewton : return "Newton法";
  case methnon : return "なし";
  case methsimplex : return "滑降Simplex法";
  default : return "滑降Simplex+Newton法";
  }
}

class varclass                  // 変数クラス
{
private:
  int vnp;                      // パラメータ数
  int vnd;                      // 用量数
  mapds vdsmap;                 // 用量別反応統計量マップ
  pairdbl drange;               // 用量の最小値と最大値
  pairdbl yrange;               // 反応の最小値と最大値
  statcor<double> statdxy;      // メタ用量と反応の相関基礎統計量
  vec<double> vh0v;             // パラメーターの初期値ベクトル
  vec<double> vhv;              // パラメーター推測値ベクトル
  double vym;                   // ロジスティック関数の最小値
  int vnr;                      // 繰り返し計算数
  int vdfb;                     // 回帰自由度
  double vsb;                   // 回帰平方和
  int vdfl;                     // ズレ自由度
  double vsl;                   // ズレ平方和
  int vdfd;                     // 用量自由度
  double vsd;                   // 用量平方和
  int vdfr;                     // 残差自由度
  double vsr;                   // 残差平方和
  int vdft;                     // 全自由度
  double vst;                   // 全平方和
  double vqerr;                 // 回帰誤差
  restype vresvar;              // 処理結果のフラッグ
  int init_h(option& opt);   // パラメータの初期値を設定する
  int init_log( option& opt); // 微分回帰分析による初期値の設定
  int cal_smat(mat<double>& smat, option& opt);
  // 微分回帰分析用に積和行列を計算する
  restype cal_drc(option& opt, ldata& datalst);
  // 非線形最小2乗法によって用量反応曲線を計算する
  restype cal_simplex(int& nr, vec<double>& hv, option& opt, ldata& datalst) const;
  // Simplex法による非線形最小2乗計算
  void init_simplex(mat<double>& simplexmat, const vec<double>& hv) const;
  // シンプレックスの初期設定
  restype check_simplex(mat<double>& simplexmat, vec<double>& hlv, vec<double>& hmv, vec<double>& hhv, double& fl, double& fs, double& fh, int& lidx, int& hidx, int nr, option& opt, ldata& datalst) const;
  // fiの計算と終了基準のチェック
  restype check_hv(const vec<double>& hv) const; // パラメータの発散チェック
  restype cal_newton(int& nr, vec<double>& hv, option& opt, ldata& datalst) const;
  // Newton-Raphson法
  void cal_gh(vec<double>& gradv, mat<double>& hessmat, const vec<double>& hv, option& opt, ldata& datalst) const;
  // 傾斜ベクトルとHessの行列の計算
  restype check_q(double qnew, double qold, const vec<double>& hv, int nr, option& opt) const;
  // Q(h)の計算と終了基準のチェック
  double fq(const vec<double>& hv, option& opt, ldata& datalst) const;
  // 誤差関数
  double fqd(int hidx, const vec<double>& hv, option& opt, ldata& datalst) const;
  // 誤差の偏導関数
  double fqd2(int hiidx, int hjidx, const vec<double>& hv, option& opt, ldata& datalst) const;
  // 誤差の2階偏導関数
  double func(double x, const vec<double>& hv) const;
  // 用量反応関数
  double funcd(double x, int hidx, const vec<double>& hv) const;
  // モデル偏導関数
  double funcd2(double x, int hiidx, int hjidx, const vec<double>& hv) const;
  // モデル2階偏導関数
  void cal_ss(option& opt, ldata& datalst);
  // 各種の平方和を計算する
public:
  varclass(int np0) : vnp(np0), vnd(0), vdsmap(), drange(FDmax, -FDmax), yrange(FDmax, -FDmax), statdxy(), vh0v(np0), vhv(np0), vym(0.0), vnr(0), vdfb(0), vdfl(0), vsl(0.0), vdfd(0), vsd(0.0), vdfr(0), vsr(0.0), vdft(0), vst(0.0), vqerr(0.0), vresvar(resnon)
  // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int np() const                // パラメータ数を返す
  { return vnp; }
  int nd() const                // 用量数を返す
  { return vnd; }
  const mapds& dsmap() const    // 用量別統計量マップを返す
  { return vdsmap; }
  int n() const                 // 用量反応の例数を返す
  { return statdxy.n(); }
  double mx() const             // 用量反応の用量平均値を返す
  { return statdxy.mx(); }
  double my() const             // 用量反応の反応平均値を返す
  { return statdxy.my(); }
  double sxx() const            // 用量反応の用量平方和を返す
  { return statdxy.sxx(); }
  double syy() const            // 用量反応の反応平方和を返す
  { return statdxy.syy(); }
  double sdy() const            // 用量反応の反応標準偏差を返す
  { return statdxy.sdy(); }
  double sey() const            // 用量反応の反応標準誤差を返す
  { return statdxy.sey(); }
  double sxy() const            // 用量反応の用量と反応積和を返す
  { return statdxy.sxy(); }
  double dmin() const           // 用量の最小値を返す
  { return drange.first; }
  double dmax() const           // 用量の最大値を返す
  { return drange.second; }
  double ymin() const           // 反応の最小値を返す
  { return yrange.first; }
  double ymax() const           // 反応の最大値を返す
  { return yrange.second; }
  double h0(size_t idx) const   // パラメータ初期値を返す
  { return (idx < vh0v.dim()) ? vh0v(idx) : 0.0; }
  double h(size_t idx) const    // パラメータを返す
  { return (idx < vhv.dim()) ? vhv(idx) : 0.0; }
  double ym() const             // ロジスティック関数の最小値を返す
  { return vym; }
  int nr() const                // 繰り返し計算数を返す
  { return vnr; }
   int dfb() const              // 回帰自由度を返す
  { return vdfb; }
  double sb() const             // 回帰平方和を返す
  { return vsb; }
   int dfl() const              // ズレ自由度を返す
  { return vdfl; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  int dfd() const               // 用量自由度を返す
  { return vdfd; }
  double sd() const             // 用量平方和を返す
  { return vsd; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dft() const               // 全自由度を返す
  { return vdft; }
  double st() const             // 全平方和を返す
  { return vst; }
  double qerr() const           // 回帰誤差を返す
  { return vqerr; }
  restype resvar() const        // 処理結果のフラッグを返す
  { return vresvar; }
  void cal(option& opt, ldata datalst); // データを読み込みロジット分析を行う
  double fdy(double x) const    // 用量反応関数(vhvを使ったもの)
  { return func(x, vhv); }
  double fddy(double x, int hidx) const // 用量反応偏導関数(vhvを使ったもの)
  { return funcd(x, hidx, vhv); }
};

void varclass::cal(option& opt, ldata datalst)
// データを読み込みロジット分析を行う
{
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (dose != BLANKf && y != BLANKf) // dose<=0も集計する
    {
      mapds_itr dsitr = vdsmap.find(dose);
      if (dsitr == vdsmap.end()) // 新規用量
      {
        statv<double> statvi(y);
        vdsmap.insert(make_pair(dose, statvi));
      }
      else                      // すでにある用量
      { dsitr->second.sigma(y);  }
      if (opt.du() != dulog || sgn(dose) > 0)
      {
        if (dose < drange.first) drange.first = dose;
        if (dose > drange.second) drange.second = dose;
        if (y < yrange.first) yrange.first = y;
        if (y > yrange.second) yrange.second = y;
        double x = (opt.du() == dulog) ? log10(dose) : dose; // メタ用量
        statdxy.sigma(x, y);
      }
    }
  }
  vnd = init_h(opt);            // パラメータの初期値を設定する
  if (vnd < vnp)                // 計算不可能
  {
    vresvar = reserr;
    return;
  }
  vresvar = cal_drc(opt, datalst);
  if (vresvar == reserr) return; // 非線形最小2乗法発散
  cal_ss(opt, datalst);         // 平方和と自由度を計算する
}

int varclass::init_h(option& opt) // パラメータの初期値を設定する
{
  double y0 = (yrange.first > 0.0) ? yrange.first * 0.5 : 0.0;
  double m = yrange.second - y0;
  double b = (statdxy.sxx() > 0.0) ? statdxy.sxy() / statdxy.sxx() : 0.0;
  double c = (sgn(m) == 0) ? 0.0 : 8.0 * b / m; // ロジスティック関数の回帰係数
  if (opt.ym() == -DBL_MAX)     // ymを微分回帰曲線から求める
  { vym = 0.0; }
  else if (opt.ym() == BLANKf)  // ym:データから自動設定
  {
    vym = BLANKf;
    for (mapds_itr ditr = vdsmap.begin(); vym == BLANKf && ditr != vdsmap.end(); ++ditr)
    {
      double dose = ditr->first, y = ditr->second.mean();
      if (sgn(dose) == 0 && vym == BLANKf) vym = y;
    }
    if (vym == BLANKf || sgn(c) < 0) // ym設定またはc<0
    { opt.ym(-DBL_MAX);  vym = 0.0; }
  }
  else                          // ym:指定値
  { vym = opt.ym(); }
  m = yrange.second - vym;      // Mの再設定
  vh0v.assign(0, y0);  vh0v.assign(1, c);  vh0v.assign(2, m); // 仮初期値
  int nd = init_log(opt);       // 2次回帰分析による初期値設定
  if (opt.nh0() > 0)            // 初期値が指定されている場合
  {
    for (size_t i = 0; i < opt.nh0(); ++i) vh0v.assign(i, opt.h0(i));
  }
  vhv = vh0v;                   // パラメータ初期値を代入する
  return nd;                    // 有効な用量数を返す
}

int varclass::init_log(option& opt)
// 微分回帰分析による初期値の設定
{
  int p = 2;                    // 2次回帰式のパラメータ数
  if (opt.ym() == -DBL_MAX) p = 3; // ymを微分回帰分析から求める
  int dim = p + 1;
  size_t dyidx = p;             // dyのインデックス
  mat<double> smat(dim);        // 積和行列
  int nd = cal_smat(smat, opt);
  if (nd - 2 < p) return nd;    // 2次回帰分析不可能
  double b0 = 0.0, b1 = 0.0, b2 = 0.0, c = 0.0, m = 0.0;
  if (opt.ym() == -DBL_MAX)     // ymを微分回帰分析から求める
  {
    b0 = smat(0, dyidx);  b1 = smat(1, dyidx);  b2 = smat(2, dyidx);
    if (sgn(b1 * b2) >= 0) return nd; // 不適解
    double tmp = b1 * b1 - 4.0 * b2 * b0;
    tmp = (sgn(tmp) > 0) ? sqrt(tmp) : 0.0;
    vym = (-b1 + tmp) / (2.0 * b2);
    double ymax = (-b1 - tmp) / (2.0 * b2);
    if (sgndiff(vym, ymax) < 0)
    { m = ymax - vym; }
    else
    {
      m = vym - ymax;
      vym = ymax;
    }
    c = -b2 * m;
  }
  else                          // ymをデータから求める／-ymで指定
  {
    b1 = smat(0, dyidx);  b2 = smat(1, dyidx);
    if (sgn(b1 * b2) >= 0) return nd; // 不適解
    c = b1;  m = -b1 / b2;
  }
  statcor<double> statcdly0;
  for (auto& dstat : vdsmap)
  {
    double dose = dstat.first, my = dstat.second.mean() - vym;
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      if (0.0 < my && my < m)   // myが範囲内の時だけ集計
      {
        double x = (opt.du() == dulog) ? log10(dose) : dose;
        double ly = log(my / (m - my));
        statcdly0.sigma(x, ly);
      }
    }
  }
  if (statcdly0.n() < 1) return nd; // ロジットがない時
  double ly0 = statcdly0.my() - c * statcdly0.mx();
  double y0 = vym + m / (1.0 + exp(-ly0));
  vh0v.assign(0, y0);  vh0v.assign(1, c);  vh0v.assign(2, m);
  return nd;                     // 有効な用量数返す
}

int varclass::cal_smat(mat<double>& smat, option& opt)
// 微分回帰分析用に積和行列を計算する
{
  size_t dim = smat.r();        // 行列の次元
  size_t dyidx = dim - 1;       // dyのインデックス
  vecdbl fdatavec(dim);
  size_t offset = 0;
  if (dim == 4) { fdatavec[0] = 1.0;  offset = 1; } // 定数項がある時
  double md1 = 0.0, md2 = 0.0, y1 = 0.0, y2 = 0.0;
  int nd = 0;
  for (auto& dstat : vdsmap)
  {
    double dose = dstat.first, y = dstat.second.mean();
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      if (nd < 1)
      { md1 = mdose;  y1 = y;  nd = 1; }
      else if (nd == 1)
      { md2 = mdose;  y2 = y;  nd = 2; }
      else
      {
        double mdd1 = md2 - md1, mdd2 = mdose - md2;
        double eps = DBL_EPSILON * fmax(5.0, fabs(mdd1)); // 計算誤差
        double w1 = (fabs(mdd1) < eps) ? 0.0 : 1.0 / mdd1;
        double w2 = (fabs(mdd2) < eps) ? 0.0 : 1.0 / mdd2;
        double b1 = (y2 - y1) * w1, b2 = (y - y2) * w2;
        w1 = fabs(w1);  w2 = fabs(w2);
        double dy = (fabs(w1 + w2) < eps) ? 0.0 : (b1 * w1 + b2 * w2) / (w1 + w2);
        double my = y2;
        if (vym != -DBL_MAX) my -= vym; // 定数項を引く
        double fd = my;               // 例数で重み付けしない
        for (size_t i = 0; i < 2; ++i)
        {
          fdatavec[offset + i] = fd;
          fd *= my;
        }
        fdatavec[dyidx] = dy;
        for (size_t i = 0; i < dim; ++i)
        {
          fd = fdatavec[i];
          for (size_t j = i; j < dim; ++j) smat.add(i, j, fd * fdatavec[j]);
        }
        md1 = md2;  y1 = y2;
        md2 = mdose;  y2 = y;  ++nd;
      }
    }
  }
  for (size_t i = 0; i < dyidx; ++i)
  {
    for (size_t j = i + 1; j <= dyidx; ++j) smat.assign(j, i, smat(i, j));
  }
  for (size_t i = 0; i < dyidx; ++i) smat.sweep(i);
  return nd;                    // 有効な用量数を返す
}

restype varclass::cal_drc(option& opt, ldata& datalst)
// 非線形最小2乗法によって用量反応曲線を計算する
{
//  int nr1 = 0, nr2 = 0;         // 1次と2次の反復計算数
  restype resdrc = resnon;
  if (opt.meth() == methetc || opt.meth() == methsimplex)
  {  // 収束を良くするためSimplex法で1次近似計算をする
    resdrc = cal_simplex(vnr, vhv, opt, datalst);
    if (resdrc == reserr) return resdrc; // 発散
  }
  if (opt.meth() == methetc || opt.meth() == methnewton)
  {
    int nr = 0;
    resdrc = cal_newton(nr, vhv, opt, datalst);
    vnr += nr;
  }
  return resdrc;                // 処理の結果フラッグを返す
}

restype varclass::cal_simplex(int& nr, vec<double>& hv, option& opt, ldata& datalst) const
// Simplex法による非線形最小2乗計算
{
  int p = hv.dim();             // パラメーター数
  mat<double> simplexmat(p, p + 1); // シンプレックスを行列で表す
  vec<double> hlv(hv), hmv(hv), hhv(hv);
  double fl = 0.0, fs = 0.0, fh = 0.0; // 最小値、2番目に大きい値、最大値
  int lidx = 0, hidx = 0;
  restype ressimp = resnon;
  nr = 0;
  init_simplex(simplexmat, hv);
  while ((ressimp = check_simplex(simplexmat, hlv, hmv, hhv, fl, fs, fh, lidx, hidx, nr, opt, datalst)) == resnon)
  {
    vec<double> hrv(hmv);
    hrv *= 2.0;
    hrv -= hhv;
    double fr = fq(hrv, opt, datalst);
    if (fr < fl)                // 折り返し成功
    {
      vec<double> hev(hrv);
      hev *= 2.0;
      hev -= hmv;
      double fe = fq(hev, opt, datalst);
      if (fe < fr)              // 拡大成功
      { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hev(i)); }
      else                      // 拡大失敗(行き過ぎ)
      { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hrv(i)); }
    }
    else if (fr <= fs)          // 斜め方向の折り返し
    { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hrv(i)); }
    else                        // 縮小失敗、Simplex移動
    {
      if (fr < fh) { fh = fr;  hhv = hrv; } // 折り返し後のSimplex中に最小点
      vec hcv(hmv + hhv);
      hcv *= 0.5;
      double fc = fq(hcv, opt, datalst);
      if (fc < fh)              // 縮小成功
      { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hcv(i)); }
      else                      // 縮小失敗、Simplex移動
      {
        for (int j = 0; j <= p; ++j)
        {
          if (j != lidx)
          {
            for (int i = 0; i < p; ++i) simplexmat.assign(i, j, (simplexmat(i, j) + hlv(i)) * 0.5);
          }
        }
      }
    }
    ++nr;
  }
  if (ressimp == reserr || fabs(fl) > FDmax) return reserr; // 発散
  hv = hlv;
  return ressimp;               // 計算できた
}

void varclass::init_simplex(mat<double>& simplexmat, const vec<double>& hv) const
// シンプレックスの初期設定
{
  constexpr double Sqrt2 = sqrt(2.0); // コンパイル時定数
  constexpr double Rs = 0.1;          // シンプレックスの1辺の相対割合
  int p = hv.dim();
  double s = absmin(hv) * Rs;
  double v = (sqrt((double)(p + 1)) - 1.0 + (double)p) / (double)p / Sqrt2 * s;
  double w = (sqrt((double)(p + 1)) - 1.0) / (double)p / Sqrt2 * s;
  for (int i = 0; i < p; ++i) simplexmat.assign(i, 0, hv(i));
  for (int j = 1; j <= p; ++j)
  {
    for (int i = 0; i < p; ++i) simplexmat.assign(i, j, (i == j - 1) ? v + hv(i) : w + hv(i));
  }
}

restype varclass::check_simplex(mat<double>& simplexmat, vec<double>& hlv, vec<double>& hmv, vec<double>& hhv, double& fl, double& fs, double& fh, int& lidx, int& hidx, int nr, option& opt, ldata& datalst) const
// fiの計算と終了基準のチェック
{
  int p = hlv.dim();
  int eidx = p - 1;
  vecdbl fvec(p + 1, 0.0);
  fl = FDmax;  fh = -FDmax;
  for (int j = 0; j <= p; ++j)
  {
    vec<double> hjv(simplexmat.sub(0, j, eidx, j));
    double fj = fq(hjv, opt, datalst);
    fvec[j] = fj;
    if (fj < fl) { fl = fj;  lidx = j; }
    if (fj > fh) { fh = fj;  hidx = j; }
  }
  if (fl > FDmax || isnan(fl)) return reserr; // 発散
  hlv = simplexmat.sub(0, lidx, eidx, lidx); // 最小値の時のhv
  hhv = simplexmat.sub(0, hidx, eidx, hidx); // 最大値の時のhv
  if (check_hv(hlv) == reserr) return reserr; // 発散
  double ferr = fabs(fh - fl);
  if (ferr < fabs(fl * opt.eps())) return resok; // 収束
  if (opt.rmax() > 0 && nr >= opt.rmax()) return resrmax; // 反復回数最大値
  fs = fl;
  hmv = 0.0;
  for (int j = 0; j <= p; ++j)
  {
    if (j != hidx)
    {
      if (fvec[j] > fs) fs = fvec[j];
      hmv += simplexmat.sub(0, j, eidx, j);
    }
  }
  hmv *= 1.0 / (double)p;       // hhv以外のp個のhvの重心
  return resnon;                // 収束せず
}

restype varclass::check_hv(const vec<double>& hv) const
// パラメータの発散チェック
{
  restype reshv = resnon;
  for (size_t i = 0; reshv == resnon && i < hv.dim(); ++i)
  {
    if (fabs(hv(i)) > FDmax || isnan(hv(i))) reshv = reserr;
  }
  if (reshv == reserr) return reshv; // 発散
  if (sgndiff(hv(0), vym) < 0 || sgndiff(hv(0) - vym, hv(2)) > 0) return reserr;
  return resok;                 // 発散せず
}

restype varclass::cal_newton(int& nr, vec<double>& hv, option& opt, ldata& datalst) const
// Newton-Raphson法
{
  int p = hv.dim();
  vec<double> gradv(p), htv(hv);
  mat<double> hessmat(p, p);
  double qnew = fq(htv, opt, datalst), qold = 0.0;
  restype resnew = resnon;
  nr = 0;
  do
  {
    cal_gh(gradv, hessmat, htv, opt, datalst);
    hessmat.inv();
    htv -= hessmat * gradv;
    qold = qnew;
    qnew = fq(htv, opt, datalst);  ++nr;
  } while ((resnew = check_q(qnew, qold, htv, nr, opt)) == resnon);
  if (resnew == reserr) return resnew; // 発散
  hv = htv;
  return resnew;                // 計算できた
}

void varclass::cal_gh(vec<double>& gradv, mat<double>& hessmat, const vec<double>& hv, option& opt, ldata& datalst) const
// 傾斜ベクトルとHessの行列の計算
{
  for (size_t i = 0; i < hv.dim(); ++i)
  {
    gradv.assign(i, fqd(i, hv, opt, datalst));
    for (size_t j = i; j < hv.dim(); ++j)
    { hessmat.assign(i, j, fqd2(i, j, hv, opt, datalst)); }
  }
  for (size_t i = 0; i < hv.dim() - 1; ++i)
  {
    for (size_t j = i + 1; j < hv.dim(); ++j) hessmat.assign(j, i, hessmat(i, j));
  }
}

restype varclass::check_q(double qnew, double qold, const vec<double>& hv, int nr, option& opt) const
// Q(h)の計算と終了基準のチェック
{
  if (check_hv(hv) == reserr) return reserr;    // 発散
  if (qnew > FDmax || isnan(qnew)) return reserr; // 発散
  double qerr = fabs(qold - qnew);
  if (qerr < fabs(qnew * opt.eps())) return resok; // 収束
  if (opt.rmax() > 0 && nr >= opt.rmax()) return resrmax; // 反復回数最大値
  return resnon;                                          // 収束せず
}

double varclass::fq(const vec<double>& hv, option& opt, ldata& datalst) const
// 誤差関数
{
  double q = 0.0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double x = (opt.du() == dulog) ? log10(dose) : dose;
      double ei = y - func(x, hv);
      q += ei * ei;
    }
  }
  return q;                     // 平方誤差合計を返す
}

double varclass::fqd(int hidx, const vec<double>& hv, option& opt, ldata& datalst) const
// 誤差の偏導関数
{
  double dq = 0.0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double x = (opt.du() == dulog) ? log10(dose) : dose;
      double dei = y * funcd(x, hidx, hv) - func(x, hv) * funcd(x, hidx, hv);
      dq -= 2.0 * dei;
    }
  }
  return dq;                    // 誤差導関数値を返す
}

double varclass::fqd2(int hiidx, int hjidx, const vec<double>& hv, option& opt, ldata& datalst) const
// 誤差の2階偏導関数
{
  double d2q = 0.0;
  for (auto& cdatai : datalst)
  {
    double dose = cdatai.fdata(0), y = cdatai.fdata(1);
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double x = (opt.du() == dulog) ? log10(dose) : dose;
      double qd2 = funcd2(x, hiidx, hjidx, hv);
      double d2ei = y * qd2 - funcd(x, hiidx, hv) * funcd(x, hjidx, hv) - func(x, hv) * qd2;
      d2q -= 2.0 * d2ei;
    }
  }
  return d2q;                   // 誤差2階導関数値を返す
}

double varclass::func(double x, const vec<double>& hv) const
// 用量反応関数
{
  return vym + logdis(x, hv(2), hv(1), hv(0) - vym);
}

double varclass::funcd(double x, int hidx, const vec<double>& hv) const
// モデル偏導関数
{
  constexpr double Delta = EPSdef; // 微小変動の割合(差分法用)
  vec<double> hdv(hv);
  double dh = 0.0, grad = 0.0;
  dh = hv(hidx) * Delta;        // 差分法
  hdv.assign(hidx, hdv(hidx) + dh);
  grad = (sgn(dh) == 0) ? 0.0 : (func(x, hdv) - func(x, hv)) / dh;
  return grad;                  // 偏導関数値を返す
}

double varclass::funcd2(double x, int hiidx, int hjidx, const vec<double>& hv) const
// モデル2階偏導関数
{
  constexpr double Delta = EPSdef; // 微小変動の割合(差分法用)
  vec hijv(hv), hjv(hv);
  double fhi = 0.0, fhj = 0.0, dhi = 0.0, dhj = 0.0, grad2 = 0.0;
  dhi = hv(hiidx) * Delta;      // 差分法
  dhj = hv(hjidx) * Delta;
  hijv.assign(hiidx, hv(hiidx) + dhi);
  fhi = func(x, hijv);
  hjv.assign(hjidx, hv(hjidx) + dhj);
  fhj = func(x, hjv);
  hijv.assign(hjidx, hijv(hjidx) + dhj);
  grad2 = (sgn(dhi) == 0 || sgn(dhj) == 0) ? 0.0 : (func(x, hijv) - fhi - fhj + func(x, hv)) / dhi / dhj;
  return grad2;                 // 2階偏導関数値を返す
}

void varclass::cal_ss(option& opt, ldata& datalst)
// 各種の平方和を計算する
{
  int n = statdxy.n();
  vec<double> htv(vhv);
  htv.assign(0, vym + vhv(2) * 0.5);  htv.assign(1, 0.0);
  vst = fq(htv, opt, datalst);  // y0=M/2,c=0の時のqerrをstにする
  vqerr = fq(vhv, opt, datalst);
  vsb = vst - vqerr;
  if (vsb < FDeps) vsb = 0.0;
  double my = func(0.0, htv);   // ロジスティック関数の平均値
  vsd = 0.0;
  for (auto& dstat : vdsmap)
  {
    double dose = dstat.first, y = dstat.second.mean();
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double ei = y - my;
      vsd += (double)(dstat.second.n()) * ei * ei;
    }
  }
  if (vsd < FDeps) vsd = 0.0;
  vsl = vsd - vsb;
  if (vsl < FDeps) vsl = 0.0;
  vsr = vst - vsd;
  if (vsr < FDeps) vsr = 0.0;
  vdft = n - 1;  vdfd = vnd - 1;  vdfb = 2;
  vdfr = vdft - vdfd;  vdfl = vdfd - vdfb;
  if (vdft == vdfd) { vdfr = vdfl;  vsr = vsl; } // 用量データが1例のみの時
}

class doserlclass               // doserlメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_mstat(ostream& fout, option& opt, const varclass& var) const;
  // 用量別基礎統計量を出力する
  void put_drc(ostream& fout, option& opt, const varclass& var) const;
  // 用量反応解析結果を出力する
  void put_estd(ostream& fout, option& opt, const varclass& var) const;
  // 用量を逆推定する
  void put_dratab(ostream& fout, const varclass& var) const;
  // 用量反応解析の分散分析表出力
  void plot_ngp(option& opt, idd& id, const varclass& var) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpmse(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に平均±SEを出力する
  void plot_ngpdrc(ofstream& fout, option& opt, const varclass& var) const;
  // Ngraph用に用量反応曲線を出力する
public:
  doserlclass() : fname()       // コンストラクタ
  {}
  ~doserlclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void doserlclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "用量項目:" << id.item(0) << "\n";
  fout << "反応項目:" << id.item(1) << "\n";
}

void doserlclass::put_mstat(ostream& fout, option& opt, const varclass& var) const
// 用量別基礎統計量を出力する
{
  fout << "\n    用量      ";
  if (opt.du() == dulog) fout << "x=log(用量)";
  else fout << "     x=用量";
  fout << "     例数      平均値    標準偏差    標準誤差\n";
  constexpr int LENline = 70;
  putline(fout, LENline);
  for (auto& dstat : var.dsmap())
  {
    fout << setw(8) << dstat.first;
    if (opt.du() == dulog)
    {
      if (sgn(dstat.first) > 0) fout << setw(16) << log10(dstat.first);
      else fout << "  (計算から除外)";
    }
    else
    { fout << setw(16) << dstat.first; }
    fout << setw(10) << dstat.second.n() << setw(12) << dstat.second.mean();
    fout << setw(12) << dstat.second.sd() << setw(12) << dstat.second.se();
    fout << "\n";
  }
  putline(fout, LENline);
  string tstr("平均");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(8 + npad) << tstr << setw(16) << var.mx();
  fout << setw(10) << var.n() << setw(12) << var.my();
  fout << setw(12) << var.sdy() << setw(12) << var.sey();
  fout << endl;
}

void doserlclass::put_drc(ostream& fout, option& opt, const varclass& var) const
// 用量反応解析結果を出力する
{
  double y00 = var.h0(0), c0 = var.h0(1), m0 = var.h0(2);
  double y0 = var.h(0), c = var.h(1), m = var.h(2);
  double y = y00 - var.ym();
  double ly00 = (sgn(y) > 0) ? log(y / (m0 - y)) : 0.0;
  y = y0 - var.ym();
  double ly0 = (sgn(y) > 0) ? log(y / (m - y)) : 0.0;
  double d50 = (sgn(c) == 0) ? 0.0 : -ly0 / c;
  fout << "\n・";
  int w = 16;
  if (opt.du() == dulog) { fout << "対数";  w += 4; }
  fout << "用量反応曲線:y=ym+M/[1+[{M-(y0-ym)}/(y0-ym)]exp(-c・x)]\n";
  fout << setw(w) << " " << "=ym+M/{1+exp(-ly0-c・x)}\n";
  w = 21;
  fout << "  パラメーター初期値:y0=" << y00 << "  c=" << c0 << "  M=" << m0 << "\n";
  fout << setw(w) << " " << "ym=" << var.ym() << "  ly0=" << ly00 << "\n";
  fout << "  パラメーター推定値:y0=" << y0 << "  c=" << c << "  M=" << m << "\n";
  fout << setw(w) << " " << "ym=" << var.ym() << "  ly0=" << ly0 << "\n";
  fout << "  最小値:ym=" << var.ym() << "  最大値:ym+M=" << var.ym() << "+" << m << "=" << var.ym() + m << "\n";
  fout << "  計算方法:" << opt.methname() << "  反復回数=" << var.nr();
  if (var.resvar() == resrmax)
  { fout << "(最大反復回数)"; }
  else if (var.resvar() == reserr)
  {
    fout << "\n  ※非線形最小2乗法発散！" << endl;
    return;
  }
  fout << "\n  y=ym+M/2=" << var.ym() + m * 0.5 << "の時の用量:D50=";
  if (opt.du() == dureal) fout << "-ly0/c=" << d50;
  else fout << "10^(-ly0/c)=10^(" << d50 << ")=" << pow(10.0, d50);
  double aic = (double)(var.n()) * log(var.qerr()) + 2.0 * (double)(var.np());
  fout << "\n  残差平方和(Q)=" << var.qerr() << "  赤池の情報量基準(AIC)=" << aic << "\n";
  double r2 = (var.sd() > 0.0) ? var.sb() / var.sd() : 0.0;
  fout << "  重寄与率:R2乗=" << r2 << "  重相関係数:R=" << sqrt(r2) << endl;
}

void doserlclass::put_estd(ostream& fout, option& opt, const varclass& var) const
// 用量を逆推定する
{
  double y0 = var.h(0), c = var.h(1), m = var.h(2);
  double ye = opt.y();
  if (ye == BLANKf) ye = var.ym() + m * 0.5;
  else if (ye == -DBL_MAX) ye = var.my();
  double y = ye - var.ym();
  fout << "\n・反応:y=" << ye << "の時の用量:";
  if (y <= 0.0 || m <= y)
  {
    fout << "逆推定不可能！\n";
    return;
  }
  double ly = (opt.y() == BLANKf) ? 0.0 : log(y / (m - y));
  double y00 = y0 - var.ym();
  double ly0 = log(y00 / (m - y00));
  double x = (sgn(c) == 0) ? 0.0 : (ly - ly0) / c;
  if (opt.du() == dureal) fout << "D=x=" << x;
  else fout << "D=10^(" << x << ")=" << pow(10.0, x);
  double vr = (var.dfr() > 0) ? var.sr() / (double)(var.dfr()) : 0.0;
  double xdef = x - var.mx();
  double v = vr * ((1.0 + 1.0 / (double)(var.n())) + xdef * xdef / var.sxx());
  double t = ptot(1.0 - opt.cc(), var.dfr());
  double cx = var.fddy(x, 1);   // x=xの時の用量反応曲線の傾き
  double ci = (sgn(cx) == 0) ? 0.0 : t / cx * sqrt(v); // 直線近似
  double xl = x - ci, xu = x + ci;
  fout << "\n  " << opt.cc() * 100.0 << "%信頼区間:";
  if (opt.du() == dureal) fout << "下限=" << xl << "  上限=" << xu << "\n";
  else fout << "下限=" << pow(10.0, xl) << "(x=" << xl << ")  上限=" << pow(10.0, xu) << "(x=" << xu << ")" << endl;
}

void doserlclass::put_dratab(ostream& fout, const varclass& var) const
// 用量反応解析の分散分析表出力
{
  string tstr("分散分析表(ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(51 + npad) << tstr;
  fout << "\n";
  fout << "要因          平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 74;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  calvfp(v, f, pr, var.sb(), var.dfb(), var.sr(), var.dfr());
  fout << "回帰" << setw(16) << var.sb() << setw(10) << var.dfb();
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sl(), var.dfl(), var.sr(), var.dfr());
  fout << "ズレ" << setw(16) << var.sl() << setw(10) << var.dfl() << setw(14) << v;
  if (var.dft() > var.dfd()) fout << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  fout << "用量" << setw(16) << var.sd() << setw(10) << var.dfd();
  if (var.dft() > var.dfd())
  {
    fout << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    v = (var.dfr() > 0) ? var.sr() / (double)(var.dfr()) : 0.0;
    fout << "残差" << setw(16) << var.sr() << setw(10) << var.dfr();
    fout << setw(14) << v << "\n";
    putline(fout, LENline);
    fout << "全体" << setw(16) << var.st() << setw(10) << var.dft();
  }
  fout << endl;
}

void doserlclass::plot_ngp(option& opt, idd& id, const varclass& var) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("drc"))           // 用量反応曲線
  {
    fout << "-s1 -dline -w80";
    if (opt.du() == dulog) fout << " -exlog";
    fout << "\n";
    plot_ngpid(fout, id);
    plot_ngpdrc(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("mse"))           // 平均値±SE
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -o300";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    plot_ngpid(fout, id);
    plot_ngpmse(fout, opt, var);
  }
}

void doserlclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%用量項目:" << id.item(0) << "\n";
  fout << "%反応項目:" << id.item(1) << "\n";
}

void doserlclass::plot_ngpmse(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に平均±SEを出力する
{
  fout << "%平均値±SE：SEを描く時は-derrorbar_y -w40 -o500\n";
  for (auto& dstat : var.dsmap())
  {
    if (opt.du() != dulog || dstat.first > 0.0)
    {
      double se = dstat.second.se();
      fout << dstat.first << " " << dstat.second.mean() << " " << -se << " " << se << "\n";
    }
  }
}

void doserlclass::plot_ngpdrc(ofstream& fout, option& opt, const varclass& var) const
// Ngraph用に用量反応曲線を出力する
{
  constexpr int RDdef = 50;     // グラフの描画範囲分割数省略時値
  fout << "%用量反応曲線:計算方法=" << opt.methname() << "  反復回数=" << var.nr();
  if (var.resvar() == resrmax) fout << "(最大反復回数)";
  fout << "\n";
  double dmin = var.dmin(), dmax = var.dmax();
  double xmin = (opt.du() == dulog) ? log(dmin) : dmin;
  double xmax = (opt.du() == dulog) ? log(dmax) : dmax;
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf) // X軸の範囲指定あり
  { xmin = opt.xmin();  xmax = opt.xmax(); }
  else
  {
    double xrange = xmax - xmin, xmg = 0.05;
    xmin -= xrange * xmg;  xmax += xrange * xmg;
  }
  double dx = (xmax - xmin) / (double)RDdef;
  double x = xmin;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
    fout << dose << " " << var.fdy(x) << endl;
  }
}

int doserlclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void doserlclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  varclass var(NPdef);          // 変数オブジェクト
  var.cal(opt, datalst);
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout, id);
    if (opt.res("stat")) put_mstat(cout, opt, var);
    if (var.resvar() == reserr)
    {
      cout << "※用量反応解析不可能！\n\n" << endl;
      return;
    }
    if (opt.res("reg"))
    {
      put_drc(cout, opt, var);
      if (opt.res("y")) put_estd(cout, opt, var);
      put_dratab(cout, var);
    }
    cout << "\n" << endl;
  }
  if (var.resvar() != reserr && opt.ngpname() != "") plot_ngp(opt, id, var);
}

int main(int argc, char *argv[]) // メインルーチン
{
  doserlclass doserl;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) doserl.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { doserl.get(argv[i], opt, id, datalst);  fok = false; }
    }
    doserl.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
