// parapl.C Version 7.43  Copyleft (c) 杉本典夫
// Time-stamp: <2025-03-05 08:55:45 tac>
//
// 層別クロス集計表を入力し、プロビット法による平行線検定を行う。
// 層別クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W parapl parapl.C
//               -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "parapl  Version 7.43         Copyleft(Vo\\oV) 2025 杉本典夫",
  "書式:parapl [<オプション> …]  [<データファイル> …]",
  "機能:層別項目を群、縦項目を用量、横項目を反応としたクロス集計表に対して",
  "     プロビット法による平行線検定をするよ。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するよ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -du={l|r}:用量の単位を指定するよ。",
  "    l:用量を対数変換して計算(オプション省略時)  r:実数のまま計算",
  "  -eps[=<収束誤差>]:最尤法の収束誤差を指定するよ。(0＜収束誤差)",
  "  -f=<ファイル>:オプションを指定したファイルから読み込むよ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するよ。(-h:-h=man)",
  "  -rmax=<最大反復回数>:最大反復回数を指定するよ。(0≦最大反復回数)",
  "  -res=[-][{def|all|at|er|g|ma[{b|f|s|t}]|p[<反応率>]|stat",
  "    |tab[{fr|pc|pr|pt}]}|v][,…]:解析結果の出力項目を指定するよ。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    at:分散分析表  er:効力比  g:Fiellerの式を使う",
  "    ma[{b|f|s|t}]:修正切片差の多重比較",
  "      b:Bonferroni型  f:Fisher型  s:Scheffe型  t:Tukey型",
  "    p[<反応率>]:用量の逆推定値(0＜反応率＜1  p:p0.5)",
  "    stat:基礎統計量  v:計算過程",
  "    オプション省略時:データに合わせて自動判定"
};

const string Title = "=== プロビット法による平行線検定 ==="; // 表題
constexpr double EPSdef = 0.00001; // 省略時収束誤差
constexpr double Pdef = 0.5;      // 省略時用量推定反応率
constexpr int RMAXdef = 200;      // 最大計算回数省略時値

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  dutype vdu;                   // 用量の単位
  double veps;                  // 最尤法の収束誤差
  double vp;                    // 用量を逆推定する反応率
  mctype vma;                   // 修正平均値の多重比較法
  int vrmax;                    // 最大反復回数
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  mctype ma() const             // 修正群差の多重比較法を返す
  { return vma; }
  double p() const              // 用量を逆推定する反応率を返す
  { return vp; }
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nftab);         // オプションのチェック
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdu(dulog), veps(EPSdef), vma(mcbon), vp(Pdef), vrmax(RMAXdef), vtab(tabfprr), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "du")      // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < FDeps) err = badopt(vpname, optstr, USAGE, FDeps, FDmax);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int nftab)    // オプションのチェック
{
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["at"] = vresmap["er"] = vresmap["g"] = vresmap["p"] = vresmap["stat"] = true;
  vresmap["ma"] = vresmap["tab"] = vresmap["v"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["at"] = vresmap["er"] = vresmap["g"] = vresmap["p"] = vresmap["stat"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      {
        vresmap["def"] = vresmap["at"] = vresmap["er"] = vresmap["g"] = vresmap["p"] = vresmap["stat"] = true;
      }
      else if (resstr == "all")
      {
        vresmap["res"] = vresmap["def"] = vresmap["at"] = vresmap["er"] = vresmap["g"] = vresmap["p"] = vresmap["ma"] = vresmap["stat"] = vresmap["tab"] = vresmap["v"] = true;
      }
      else if (resstr[0] == 'p' && resstr.size() > 1)
      {
        vp = strtodf(resstr.substr(1)); // 用量を推定する反応率
        if (vp <= 0.0 || 1.0 <= vp) return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！");
        vresmap["p"] = true;
      }
      else if (resstr == "mab")
      { vresmap["ma"] = true;  vma = mcbon; }
      else if (resstr == "maf")
      { vresmap["ma"] = true;  vma = mcfis; }
      else if (resstr == "mas")
      { vresmap["ma"] = true;  vma = mcsch; }
      else if (resstr == "mat")
      { vresmap["ma"] = true;  vma = mctuk; }
      else if (resstr == "tabfr") // 度数のみ
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpc") // 列合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpr") // 行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt") // 全合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class npro                      // 用量別プロビットクラス
{
private:
  double vd;                    // 用量
  double vx;                    // 計算用用量
  int vtn;                      // 総例数
  int vrn;                      // 反応例数
  double vpc;                   // 補正反応率
  double vyo;                   // 観測プロビット
  double vy;                    // 期待プロビット
public:
  npro() : vd(0.0), vx(0.0), vtn(0), vrn(0), vpc(0.0), vyo(0.0), vy(0.0)
  // デフォールトコンストラクタ
  {}
  npro(double d0, double x0, int tn0, int rn0) : vd(d0), vx(x0), vtn(tn0), vrn(rn0), vpc(0.0), vyo(0.0), vy(0.0)
  // コンストラクタ
  {}
  npro(const npro& npro0) : vd(npro0.vd), vx(npro0.vx), vtn(npro0.vtn), vrn(npro0.vrn), vpc(npro0.vpc), vyo(npro0.vyo), vy(npro0.vy)
  // コピーコンストラクタ
  {}
  virtual ~npro()               // デストラクタ
  {}
  const npro& operator=(const npro& npro0); // コピー代入演算子
  double d() const                          // 用量を返す
  { return vd; }
  double d(double d0)           // 用量を設定する
  { return (vd = d0); }
  double x() const              // 計算用用量を返す
  { return vx; }
  double x(double x0)           // 計算用用量を設定する
  { return (vx = x0); }
  int tn() const                // 総例数を返す
  { return vtn; }
  int tn(int tn0)               // 総例数を設定する
  { return (vtn = tn0); }
  int rn() const                // 反応例数を返す
  { return vrn; }
  int rn(int rn0)               // 反応例数を設定する
  { return (vrn = rn0); }
  double p() const              // 反応率を返す
  { return (vtn > 0) ? (double)vrn / (double)vtn : 0.0; }
  double pc() const             // 補正反応率を返す
  { return vpc; }
  double yo() const             // 観測プロビットを返す
  { return vyo; }
  double yo(double yo0)         // 観測プロビットを設定する
  { return (vyo = yo0); }
  double y() const              // 期待プロビットを返す
  { return vy; }
  double y(double y0)           // 期待プロビットを設定する
  { return (vy = y0); }
  void sigma(int tn0, int rn0)  // データを累積する
  { vtn += tn0;  vrn += rn0; }
  double calpc(double p0);      // 補正反応率を計算する
};

const npro& npro::operator=(const npro& npro0) // コピー代入演算子
{
  if (&npro0 == this) return *this; // 自分自身への代入を防ぐ
  vd = npro0.vd;  vx = npro0.vx;  vtn = npro0.vtn;  vrn = npro0.vrn;
  vpc = npro0.vpc;  vyo= npro0.vyo;  vy = npro0.vy;
  return *this;
}

double npro::calpc(double p0)   // 補正反応率を計算する
{
  vpc = (vtn > 0) ? (double)vrn / (double)vtn : 0.0;
  if (p0 > 0.0) vpc = (vpc - p0) / (1.0 - p0);
  if (vpc < 0.0) vpc = 0.0;
  return vpc;                   // 補正反応率を返す
}

class gnpro                     // 群別プロビットクラス
{
private:
  string vgcode;                // 群コード
  vector<npro> nprovec;         // 用量別プロビットベクトル
  statcor<double> dstatc;       // 相関用基礎統計量
  double vsnw;                  // プロビット計算用例数
  double vmxw;                  // プロビット計算用X平均
  double vmyw;                  // プロビット計算用Y平均
  double vsxxw;                 // プロビット計算用Sxx
  double vsyyw;                 // プロビット計算用Syy
  double vsxyw;                 // プロビット計算用Sxy
  double vsb;                   // 用量平方和
  double va;                    // 群別定数
  double vb;                    // 群別回帰係数
  double vac;                   // 共通定数
  double vbc;                   // 共通回帰係数
public:
  gnpro() : vgcode(), nprovec(), dstatc(), vsnw(0.0), vmxw(0.0), vmyw(0.0), vsxxw(0.0), vsyyw(0.0), vsxyw(0.0), vsb(0.0), va(0.0), vb(0.0), vac(0.0), vbc(0.0)
  // デフォールトコンストラクタ
  {}
  gnpro(dutype du, const string& gcode0, const ftable& ftab);
  // コンストラクタ
  gnpro(const gnpro& gnpro0) : vgcode(gnpro0.vgcode), nprovec(gnpro0.nprovec), dstatc(gnpro0.dstatc), vsnw(gnpro0.vsnw), vmxw(gnpro0.vmxw), vmyw(gnpro0.vmyw), vsxxw(gnpro0.vsxxw), vsyyw(gnpro0.vsyyw), vsxyw(gnpro0.vsxyw), vsb(gnpro0.vsb), va(gnpro0.va), vb(gnpro0.vb), vac(gnpro0.vac), vbc(gnpro0.vbc)
  // コピーコンストラクタ
  {}
  virtual ~gnpro()              // デストラクタ
  {}
  const gnpro& operator=(const gnpro& gnpro0); // コピー代入演算子
  const string& gcode() const                  // 群コードを返す
  { return vgcode; }
  double snw() const            // プロビット計算用例数を返す
  { return vsnw; }
  double mxw() const            // プロビット計算用用量平均を返す
  { return vmxw; }
  double myw() const            // プロビット計算用反応平均を返す
  { return vmyw; }
  double sxxw() const           // プロビット計算用用量積和を返す
  { return vsxxw; }
  double syyw() const           // プロビット計算用反応積和を返す
  { return vsyyw; }
  double sxyw() const        // プロビット計算用用量と反応の積和を返す
  { return vsxyw; }
  int nd() const                // 用量数を返す
  { return nprovec.size(); }
  int n() const                 // プロビットが計算できた用量数を返す
  { return dstatc.n(); }
  double mx() const             // 用量平均を返す
  { return dstatc.mx(); }
  double my() const             // 反応平均を返す
  { return dstatc.my(); }
  double sxx() const            // 用量積和を返す
  { return dstatc.sxx(); }
  double syy() const            // 反応積和を返す
  { return dstatc.syy(); }
  double sxy() const            // 用量と反応の積和を返す
  { return dstatc.sxy(); }
  double sb() const             // 用量平方和を返す
  { return vsb; }
  double a() const              // 群別定数を返す
  { return va; }
  double b() const              // 群別回帰係数を返す
  { return vb; }
  double ac() const             // 共通定数を返す
  { return vac; }
  double ac(double ac0)         // 共通定数を設定する
  { return (vac = ac0); }
  double bc() const             // 共通回帰係数を返す
  { return vbc; }
  double bc(double bc0)         // 共通回帰係数を設定する
  { return (vbc = bc0); }
  double dose(int idx) const    // 用量を返す
  { return nprovec[idx].d(); }
  double x(int idx) const       // 計算用用量を返す
  { return nprovec[idx].x(); }
  int tn(int idx) const         // 総例数を返す
  { return nprovec[idx].tn(); }
  int rn(int idx) const         // 反応例数を返す
  { return nprovec[idx].rn(); }
  double p(int idx) const       // 反応率を返す
  { return nprovec[idx].p(); }
  double pc(int idx) const      // 補正反応率を返す
  { return nprovec[idx].pc(); }
  double yo(int idx) const      // 観測プロビットを返す
  { return nprovec[idx].yo(); }
  double yo(int idx, double yo0) // 観測プロビットをセットする
  { return nprovec[idx].yo(yo0); }
  double y(int idx) const       // 期待プロビットを返す
  { return nprovec[idx].y(); }
  double y(int idx, double y0)  // 期待プロビットをセットする
  { return nprovec[idx].y(y0); }
  int calprobit(dutype du);     // 反応率とプロビットを計算する
  int caldpl(dutype du);        // 近似プロビット直線を計算する
};

gnpro::gnpro(dutype du, const string& gcode0, const ftable& ftab) : vgcode(gcode0), nprovec(ftab.nr()), dstatc(), vsnw(0.0), vmxw(0.0), vmyw(0.0), vsxxw(0.0), vsyyw(0.0), vsxyw(0.0), vsb(0.0), va(0.0), vb(0.0), vac(0.0), vbc(0.0)
// コンストラクタ
{
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double dose = strtodf(ftab.rcode(i));
    double x = dose;
    if (du == dulog) x = (dose > 0.0) ? log10(dose) : BLANKf;
    npro nproi(dose, x, ftab.frtotal(i), ftab.f(i, 2));
    nprovec[i - 1] = nproi;
  }
}

const gnpro& gnpro::operator=(const gnpro& gnpro0) // コピー代入演算子
{
  if (&gnpro0 == this) return *this; // 自分自身への代入を防ぐ
  vgcode = gnpro0.vgcode;  nprovec = gnpro0.nprovec;
  dstatc = gnpro0.dstatc;  vsnw = gnpro0.vsnw;
  vmxw = gnpro0.vmxw;  vmyw = gnpro0.vmyw;  vsxxw = gnpro0.vsxxw;
  vsyyw = gnpro0.vsyyw;  vsxyw = gnpro0.vsxyw;  vsb = gnpro0.vsb;
  va = gnpro0.va;  vb = gnpro0.vb;  vac = gnpro0.vac;  vbc = gnpro0.vbc;
  return *this;
}

int gnpro::calprobit(dutype du) // 反応率とプロビットを計算する
{
  double p0 = 0.0;
  dstatc.clear();
  for (size_t i = 0; i < nprovec.size(); ++i)
  {
    double dose = nprovec[i].d();
    if (i == 0 && du == dulog && dose <= 0.0) // 用量0のデータがある時
    {
      p0 = nprovec[i].p();      // 反応率補正用
      nprovec[i].calpc(p0);     // 補正反応率計算(0になる)
      nprovec[i].yo(BLANKf);    // 用量0の観測プロビットは欠損にする
      nprovec[i].y(BLANKf);     // 用量0の期待プロビットも欠損にする
    }
    else
    {
      double x = nprovec[i].x();
      double yo = probit(nprovec[i].calpc(p0)); // 観測プロビット計算
      nprovec[i].yo(yo);
      if (x != BLANKf && yo != BLANKf) dstatc.sigma(x, yo);
    }
  }
  return dstatc.n();            // プロビットが計算できた用量数を返す
}

int gnpro::caldpl(dutype du)    // 近似プロビット直線を計算する
{
  static const double Sqrpi = 1.0 / sqrt(2.0 * M_PI);
  vsnw = vmxw = vsxxw = vmyw = vsyyw = vsxyw = 0.0;
  vsb = va = vb = 0.0;
  double p0 = 0.0;
  size_t jstart = 0;
  if (du == dulog && sgn(nprovec[0].d()) <= 0) // 用量0のデータがある時
  { p0 = nprovec[0].p();  jstart = 1; }
  double pc = p0 / (1.0 - p0);
  for (size_t j = jstart; j < nprovec.size(); ++j)
  {
    double x = nprovec[j].x(), y = nprovec[j].y();
    if (x != BLANKf && y != BLANKf)
    {
      double ned = (y - 5.0) * (y - 5.0);
      if (ned > 50.0) ned = 50.0;
      double z = Sqrpi * exp(-ned * 0.5);
      if ((ned = y - 5.0) < -10.0) ned = -10.0;
      double p = nedp(ned);
      y += (nprovec[j].pc() - p) / z;
      double nw = nprovec[j].tn() * z * z / (p + pc) / (1.0 - p);
      vsnw += nw;  vmxw += nw * x;  vsxxw += nw * x * x;
      vmyw += nw * y;  vsyyw += nw * y * y;  vsxyw += nw * x * y;
    }
  }
  if (vsnw < FDeps)
  {
    vsnw = 0.0;
    return 0;
  }
  vsxxw -= vmxw * vmxw / vsnw;
  if (vsxxw < FDeps) vsxxw = 0.0;
  vsyyw -= vmyw * vmyw / vsnw;
  if (vsyyw < FDeps) vsyyw = 0.0;
  vsxyw -= vmxw * vmyw / vsnw;
  if (fabs(vsxyw) < FDeps) vsxyw = 0.0;
  vmxw /= vsnw;
  vmyw /= vsnw;
  vb = (sgn(vsxxw) > 0) ? vsxyw / vsxxw : 0.0;
  va = vmyw - vb * vmxw;
  vsb = vb * vsxyw;
  return 1;                     // 計算実行
}

class varclass                  // 変数クラス
{
private:
  int vng;                      // 群数
  int vnr;                      // 繰り返し計算数
  vector<gnpro> gnprovec;       // 群別統計量ベクトル
  gnpro gnprodum;               // 群別統計量ダミー
  double vsbc;                  // 共通回帰平方和
  double vsid;                  // 非平行性平方和
  double vsxxwc;                // 共通用量の平方和
  double vsyywc;                // 共通プロビットの平方和
  double vsxywc;                // 共通用量プロビットの積和
  restype vresvar;              // 処理結果のフラッグ
  restype cal_ppline(int& nr, option& opt);
  // プロビット法による平行線検定用計算
  restype cal_ldpline(int& nr, option& opt);
  // Newton法による対数用量-プロビット直線の反復計算
public:
  varclass(size_t nf0) : vng(0), vnr(0), gnprovec(nf0), gnprodum(), vsbc(0.0), vsid(0.0), vsxxwc(0.0), vsyywc(0.0), vresvar(resnon)
  // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ng() const                // 群数を返す
  { return vng; }
  int nr() const                // 繰り返し計算数を返す
  { return vnr; }
  const gnpro& gnproi(size_t idx) const // 群別統計量を返す
  { return (idx < gnprovec.size()) ? gnprovec[idx] : gnprodum; }
  double sbc() const            // 共通回帰平方和を返す
  { return vsbc; }
  double sid() const            // 非平行性平方和を返す
  { return vsid; }
  double sxxwc() const          // 共通用量の平方和を返す
  { return vsxxwc; }
  double syywc() const          // 共通プロビットの平方和を返す
  { return vsyywc; }
  double sxywc() const          // 共通用量プロビットの積和を返す
  { return vsxywc; }
  restype resvar() const        // 処理結果のフラッグを返す
  { return vresvar; }
  void cal(option& opt, listidd& idlst, listftab& ftablst);
  // 集計表を読み込み平行線検定を行う
};

void varclass::cal(option& opt, listidd& idlst, listftab& ftablst)
// 集計表を読み込み平行線検定を行う
{
  vng = 0;
  listidd_itr iditr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    ftabi.trm();                // 行・列の合計が0のものを削除する
    if (ftabi.nr() > 1 && ftabi.nc() == 2)
    {
      gnpro gnproi(opt.du(), iditr->estra(0).pcdata(), ftabi);
      if (gnproi.calprobit(opt.du()) > 1) // 有効群だけ保存する
      {
        gnprovec[vng] = gnproi;
        ++vng;
      }
    }
    ++iditr;
  }
  vresvar = cal_ppline(vnr, opt);
}

restype varclass::cal_ppline(int& nr, option& opt)
// プロビット法による平行線検定用計算
{
  double sxxc = 0.0, syyc = 0.0, sxyc = 0.0;
  for (int i = 0; i < vng; ++i)
  {
    sxxc += gnprovec[i].sxx();  syyc += gnprovec[i].syy();
    sxyc += gnprovec[i].sxy();
  }
  double bc = (sxxc > 0.0) ? sxyc / sxxc : 0.0;
  for (int i = 0; i < vng; ++i)
  {
    double ac = gnprovec[i].my() - bc * gnprovec[i].mx();
    for (int j = 0; j < gnprovec[i].nd(); ++j) // 期待プロビットをセットする
    {
      double x = gnprovec[i].x(j);
      gnprovec[i].y(j, (x == BLANKf) ? BLANKf : ac + bc * x);
    }
  }
  restype resldp = cal_ldpline(nr, opt);
  vsid = 0.0;
  for (int i = 0; i < vng; ++i) vsid += gnprovec[i].sb();
  vsbc = (sgn(vsxxwc) > 0) ? vsxywc * vsxywc / vsxxwc : 0.0;
  vsid -= vsbc;
  if (vsid < FDeps) vsid = 0.0;
  return resldp;                // 処理結果フラッグを返す
}

restype varclass::cal_ldpline(int& nr, option& opt)
// Newton法による対数用量-プロビット直線の反復計算
{
  nr = 0;
  restype resldp = resnon;
  do {
    vsxxwc = vsyywc = vsxywc = 0.0;
    for (int i = 0; i < vng; ++i)
    {
      if (gnprovec[i].caldpl(opt.du())) // 近似プロビット直線計算
      {
        vsxxwc += gnprovec[i].sxxw();  vsyywc += gnprovec[i].syyw();
        vsxywc += gnprovec[i].sxyw();
      }
    }
    double emax = 0.0;
    double bc = (sgn(vsxxwc) > 0) ? vsxywc / vsxxwc : 0.0;
    for (int i = 0; i < vng; ++i)
    {
      gnprovec[i].bc(bc);
      gnprovec[i].ac(gnprovec[i].myw() - bc * gnprovec[i].mxw());
      for (int j = 0; j < gnprovec[i].nd(); ++j)
      {
        double x = gnprovec[i].x(j);
        if (x != BLANKf)
        {
          double y = gnprovec[i].ac() + bc * x;
          double e = fabs(y - gnprovec[i].y(j));
          if (e > emax) emax = e;
          gnprovec[i].y(j, y);
        }
      }
    }
    ++nr;
    if (emax < opt.eps()) resldp = resok;
    else if (opt.rmax() > 0 && nr >= opt.rmax()) resldp = resrmax;
  } while (resldp == resnon);
  return resldp;                // 処理結果フラッグを返す
}

class paraplclass               // paraplメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_dstat(ostream& fout, option& opt, listftab& ftablst, const varclass& var) const;
  // 群ごとの基礎統計量と対数用量-反応直線を出力する
  void put_nstat(ostream& fout, option& opt, const gnpro& gnproi) const;
  // 反応率・プロビット等を出力する
  void put_parapltab(ostream& fout, option& opt, const varclass& var) const;
  // プロビット法による平行線検定用分散分析表を出力する
  void put_erest(ostream& fout, option& opt, const varclass& var) const;
  // 効力比推定値を出力する
  void put_macmp(ostream& fout, option& opt, const varclass& var) const;
  // 修正定数項の多重比較を出力する
  void put_maci(ostream& fout, option& opt, const varclass& var) const;
  // 修正定数項の信頼区間を出力する
public:
  paraplclass() : fname()       // コンストラクタ
  {}
  ~paraplclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // クロス集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 集計結果を出力する
};

void paraplclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "用量項目(縦):" << id.item(i) << "\n";
      break;
    default : fout << "反応項目(横):" << id.item(i) << "\n";
    }
  }
}

void paraplclass::put_dstat(ostream& fout, option& opt, listftab& ftablst, const varclass& var) const
// 群ごとの基礎統計量と対数用量-反応直線を出力する
{
  listftab::const_iterator fitr = ftablst.begin();
  for (int i = 0; i < var.ng(); ++i, ++fitr)
  {
    fout << "\n・群" << setw(2) << i + 1;
    fout << ":" << (var.gnproi(i)).gcode() << "\n";
    if (opt.res("tab")) putftable(fout, *fitr, opt.tab());
    put_nstat(fout, opt, var.gnproi(i));
    fout << "群別";
    if (opt.du() == dulog) fout << "対数";
    fout << "用量-プロビット直線:y=" << (var.gnproi(i)).a();
    fout << showpos << (var.gnproi(i)).b() << noshowpos << "x\n";
    double y = probit(opt.p()), pr = opt.p() * 100.0;
    if (opt.res("p") && (var.gnproi(i)).b() != 0.0)
    {
      double m = (y - (var.gnproi(i)).a()) / (var.gnproi(i)).b();
      fout << "  反応率=" << pr << "%(y=" << y << ")の時の用量:D" << pr << "=";
      if (opt.du() == dureal) fout << m << "\n";
      else fout << pow(10.0, m) << "(log(D" << pr << ")=" << m << ")\n";
    }
    fout << "共通";
    if (opt.du() == dulog) fout << "対数";
    fout << "用量-プロビット直線:y=" << (var.gnproi(i)).ac();
    fout << showpos << (var.gnproi(i)).bc() << noshowpos << "x\n";
    if (opt.res("p") && (var.gnproi(i)).bc() != 0.0)
    {
      double m = (y - (var.gnproi(i)).ac()) / (var.gnproi(i)).bc();
      fout << "  反応率=" << pr << "%(y=" << y << ")の時の用量:D" << pr << "=";
      if (opt.du() == dureal) fout << m << "\n";
      else fout << pow(10.0, m) << "(log(D" << pr << ")=" << m << ")\n";
    }
  }
  if (opt.res("v") == false) return;
  fout << "※反復計算回数=" << var.nr();
  if (var.resvar() == resrmax) fout << "(最大反復計算回数)";
  fout << endl;
}

void paraplclass::put_nstat(ostream& fout, option& opt, const gnpro& gnproi) const
// 反応率・プロビット等を出力する
{
  const int LENline = (opt.du() == dureal) ? 68 : 78;
  string tstr("観測      期待プロ");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(LENline - 1 + npad) << tstr;
  fout << "\n";
  fout << "実数用量";
  if (opt.du() == dulog) fout << "  対数用量";
  fout << "     反応率(反応数/総数)  補正反応率  プロビット   ビット(y)\n";
  putline(fout, LENline);
  int js = 0;
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  if (opt.du() == dulog && gnproi.dose(0) <= 0.0) // 用量0のデータがある時
  {
    fout << setw(8) << setprecision(4) << gnproi.dose(0) << setprecision(prec0);
    fout << setw(10) << "-";
    fout << setw(11) << gnproi.p(0);
    fout << "(" << setw(5) << gnproi.rn(0);
    fout << "/" << left << setw(5) << gnproi.tn(0) << right;
    fout << ")" << setw(12) << gnproi.pc(0);
    fout << "\n";
    js = 1;
  }
  tstr = "↓";
  npad = tstr.size() - strclen(tstr, CHARwide);
  string tup("↑");
  int npadu = tup.size() - strclen(tup, CHARwide);
  for (int j = js; j < gnproi.nd(); ++j)
  {
    fout << setw(8) << setprecision(4) << gnproi.dose(j);
    if (opt.du() == dulog)
    { fout << setw(10) << gnproi.x(j) << setprecision(prec0); }
    fout << setw(11) << gnproi.p(j);
    fout << "(" << setw(5) << gnproi.rn(j);
    fout << "/" << left << setw(5) << gnproi.tn(j) << right;
    fout << ")" << setw(12) << gnproi.pc(j);
    if (gnproi.pc(j) <= 0.0) fout << setw(12 + npad) << tstr;
    else if (gnproi.pc(j) >= 1.0) fout << setw(12 + npadu) << tup;
    else fout << setw(12) << gnproi.yo(j);
    fout << setw(12) << gnproi.y(j);
    fout << "\n";
  }
  putline(fout, LENline);
}

void paraplclass::put_parapltab(ostream& fout, option& opt, const varclass& var) const
// プロビット法による平行線検定用分散分析表を出力する
{
  string tstr("平行線検定表(分散分析表)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(47 + npad) << tstr;
  fout << "\n";
  fout << "要因                      平方和    自由度        χ2乗値   有意確率p値\n";
  constexpr int LENline = 74;
  putline(fout, LENline);
  int dfb = 1, dfla = 0, dft = 0, dfd = 0;
  double sla = 0.0;
  ldbl pr = 0.0L;
  for (int i = 0; i < var.ng(); ++i)
  {
    int df = (var.gnproi(i)).nd() - 1;
    if (opt.du() == dulog && (var.gnproi(i)).dose(0) <= 0.0) --df;
    double sl = (var.gnproi(i)).syyw() - (var.gnproi(i)).sb();
    int dfl = df - dfb;
    if (dfl < 1 || sl < FDeps) sl = 0.0;
    pr = ctop((var.gnproi(i)).sb(), dfb);
    fout << "群" << setw(2) << i + 1;
    fout << "の直線性" << setw(20) << (var.gnproi(i)).sb() << setw(10) << dfb;
    fout << setw(14) << (var.gnproi(i)).sb() << setw(14) << pr << pind(pr) << "\n";
    pr = ctop(sl, dfl);
    dfla += dfl;  sla += sl;  dft += df;  dfd += dfb;
    fout << "群" << setw(2) << i + 1;
    fout << "の異質性(ズレ)" << setw(14) << sl << setw(10) << dfl;
    fout << setw(14) << sl << setw(14) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
  dfd -= dfb;
  pr = ctop(var.sbc(), dfb);
  fout << "共通直線性" << setw(22) << var.sbc() << setw(10) << dfb;
  fout << setw(14) << var.sbc() << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(sla, dfla);
  fout << "異質性(ズレ)合計" << setw(16) << sla << setw(10) << dfla;
  fout << setw(14) << sla << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(var.sid(), dfd);
  fout << "非平行性" << setw(24) << var.sid() << setw(10) << dfd;
  fout << setw(14) << var.sid() << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "全体" << setw(28) << var.syywc() << setw(10) << dft << endl;
}

void paraplclass::put_erest(ostream& fout, option& opt, const varclass& var) const
// 効力比推定値を出力する
{
  double bc = (var.gnproi(0)).bc();
  if (bc == 0.0) return; // 効力比推定できず
  int ng = var.ng();
  int dc = ng * (ng - 1) / 2;
  if (opt.du() == dureal) fout << "\n○同一反応時の用量差d";
  else fout << "\n○効力比R";
  fout << "とその" << opt.cc() * 100.0 << "%";
  if (dc > 1) fout << "Bonferroni型同時";
  fout << "信頼区間";
  if (opt.res("g") == false) fout << "(近似計算)";
  fout << "\n";
  double bcx = bc * bc * var.sxxwc();
  double t = ptoz((1.0 - opt.cc()) / (double)dc);
  double g = (opt.res("g")) ? t * t / bcx : 0.0;
  if (g >= 1.0) g = 0.0;        // g>1だと計算できないため
  double gr = 1.0 - g;
  for (int i = 0; i < ng - 1; ++i)
  {
    double wi = 1.0 / (var.gnproi(i)).snw();
    double mxwi = (var.gnproi(i)).mxw(), mywi = (var.gnproi(i)).myw();
    double aci = (var.gnproi(i)).ac();
    for (int j = i + 1; j < ng; ++j)
    {
      double w = wi + 1.0 / (var.gnproi(j)).snw();
      double mdx = mxwi - (var.gnproi(j)).mxw();
      double mdy = mywi - (var.gnproi(j)).myw();
      double dm = (var.gnproi(j)).ac() - aci;
      double m = dm / bc;
      if (opt.du() == dureal) fout << "・群" << i + 1 << "-群" << j + 1 << ":d=" << m;
      else fout << "・群" << j + 1 << "/群" << i + 1 << ":R=" << pow(10.0, m) << " (log(R)=" << m << ")\n";
      double vm = gr * w + mdy * mdy / bcx;
      double ci = t / bc * sqrt(vm);
      double ml = (m - g * mdx - ci) / gr, mu = (m - g * mdx + ci) / gr;
      fout << "  信頼区間:下限=";
      if (opt.du() == dureal) fout << ml << "  上限=" << mu << "\n";
      else fout << pow(10.0, ml) << "(log(R)=" << ml << ")  上限=" << pow(10.0, mu) << "(log(R)=" << mu << ")\n";
    }
  }
}

void paraplclass::put_macmp(ostream& fout, option& opt, const varclass& var) const
// 修正定数項の多重比較を出力する
{
  int ng = var.ng(), df1 = 1, dc = 1;
  string method = "Tukey";
  switch (opt.ma())
  {
  case mcbon : method = "Bonferroni";  dc = ng * (ng - 1) / 2;
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";  df1 = ng - 1;
  }
  fout << "\n・修正切片差の" << method << "型多重比較(" << method << " type multiple comparison)\n";
  int llen = 65;                // 区切り線の長さ
  fout << "群         - 群        ";
  if (opt.ma() == mctuk) fout << "       q値  群数";
  else { fout << "      χ2乗値";  llen -= 4; }
  fout << "  自由度   有意確率p値\n";
  putline(fout, llen);
  for (int i = 0; i < ng - 1; ++i)
  {
    double wi = 1.0 / (var.gnproi(i)).snw(), mxwi = (var.gnproi(i)).mxw();
    double aci = (var.gnproi(i)).ac();
    for (int j = i + 1; j < ng; ++j)
    {
      double w = wi + 1.0 / (var.gnproi(j)).snw(), dmxw = mxwi - (var.gnproi(j)).mxw();
      double dm = (var.gnproi(j)).ac() - aci;
      double vd = w + dmxw * dmxw / var.sxxwc();
      double csqr = dm * dm / vd;
      double pr = calmcpr(opt.ma(), csqr, ng, df1, 0, dc);
      fout << left << setw(11) << j + 1;
      fout << "- " << setw(10) << i + 1;
      if (opt.ma() == mctuk) fout << right << setw(10) << csqr << setw(6) << ng << setw(8) << "inf";
      else fout << right << setw(12) << csqr << setw(8) << df1;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, llen);
}

void paraplclass::put_maci(ostream& fout, option& opt, const varclass& var) const
// 修正定数項の信頼区間を出力する
{
  int ng = var.ng(), df1 = 1, dc = 1;
  string method = "Tukey";
  switch (opt.ma())
  {
  case mcbon : method = "Bonferroni";  dc = ng * (ng - 1) / 2;
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";  df1 = ng - 1;
  }
  fout << "\n・修正切片差の" << method << "型" << opt.cc() * 100.0 << "%同時信頼区間(simultaneous confidence interval)\n";
  fout << "群      - 群        修正切片差    絶対精度        下限        上限\n";
  constexpr int LENline = 66;   // 区切り線の長さ
  putline(fout, LENline);
  for (int i = 0; i < ng - 1; ++i)
  {
    double wi = 1.0 / (var.gnproi(i)).snw(), mxwi = (var.gnproi(i)).mxw();
    double aci = (var.gnproi(i)).ac();
    for (int j = i + 1; j < ng; ++j)
    {
      double w = wi + 1.0 / (var.gnproi(j)).snw(), dmxw = mxwi - (var.gnproi(j)).mxw();
      double dm = (var.gnproi(j)).ac() - aci;
      double vd = w + dmxw * dmxw / var.sxxwc();
      double ci = calmcci(opt.ma(), opt.cc(), ng, df1, 0, dc, vd);
      fout << left << setw(8) << j + 1;
      fout << "- ";
      fout << setw(8) << i + 1 << right << setw(12) << dm;
      fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci;
      fout << "\n";
    }
  }
  putline(fout, LENline);
}

int paraplclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// クロス集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check(nftab) != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void paraplclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 集計結果を出力する
{
  if (opt.res("res") == false || ftablst.size() < 1) return;
  listidd::iterator iditr = idlst.begin();
  put_id(cout, *iditr);
  varclass var(ftablst.size()); // 変数オブジェクト
  var.cal(opt, idlst, ftablst);
  if (var.resvar() == reserr)
  {
    cout << "※平行線検定不可能！\n\n" << endl;
    return;
  }
  if (opt.res("stat")) put_dstat(cout, opt, ftablst, var);
  if (opt.res("at")) put_parapltab(cout, opt, var);
  if (var.ng() > 1)
  {
    if (opt.res("er")) put_erest(cout, opt, var);
    if (opt.res("ma"))
    {
      put_macmp(cout, opt, var);
      put_maci(cout, opt, var);
    }
  }
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  paraplclass parapl;           // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          parapl.get(nullptr, opt, idlst, ftablst);
          parapl.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        parapl.get(argv[i], opt, idlst, ftablst);  fok = false;
        parapl.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
